/*
 *
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file map_mech.c
 * @brief The implementation of maping mechanism
 */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

#include "xos_dbaux.h"
#include "xos_utils.h"
#include "xos_policy_rule.h"
#include "xos_policy_mgmt.h"
#include "xos_policy_db.h"
#include "map_mech.h"


#if DEBUG_INFO
#define AMS_DEBUG(args...)      DEBUG(args)
#else
#define AMS_DEBUG(args...)      ((void)(0))
#endif

static int _init = 0;

/** allocate a new MAPTOKEN struct
 */
MAPTOKEN *mtoken_new()
{
	MAPTOKEN *mtk = NULL;
	mtk = malloc(sizeof(*mtk));

	if (mtk == NULL) {
		goto err;
	}
	memset(mtk, 0, sizeof(MAPTOKEN));
      err:
	return mtk;
}

/** allocate a new PLYTOKEN struct
 */
PLYTOKEN *ptoken_new()
{
	PLYTOKEN *ptk = NULL;
	ptk = malloc(sizeof(*ptk));

	if (ptk == NULL) {
		goto err;
	}
	memset(ptk, 0, sizeof(PLYTOKEN));
      err:
	return ptk;
}



/** find and allocate a free uid */
uid_t _xos_get_local_freeuid(int uid_max, int uid_min)
{
	static unsigned int seed;
	int t_uid, tt_uid;
	char *gpwd_hdl = NULL;
	GPASSWD *gpwd = gpwd_new();
	GIDTOKEN oldest_gtk;
	long oldest_time = -1;
	uid_t oldest_uid = -1;

	if (_init == 0) {

		seed = (unsigned int) time((time_t *) 0);
		_init = 1;
	}


	tt_uid = t_uid = (uid_t) (uid_min +
				  (int) (((float) (uid_max - uid_min)) *
					 rand_r(&seed) / (RAND_MAX +
							  1.0)));

	/* test the id slot */

	/* forward search: t_uid --> uid_max  */
	do {
		gpwd_hdl = bdb_getdbuid_pwd(t_uid, gpwd);
		if ((gpwd->g_time < oldest_time || oldest_time == -1)
		    && gpwd_hdl) {
			memcpy(&oldest_gtk, &(gpwd->g_idtoken),
			       sizeof(GIDTOKEN));
			oldest_uid = t_uid;
			oldest_time = gpwd->g_time;
		}
		t_uid++;
	} while ((t_uid <= uid_max) && gpwd_hdl);

	if (gpwd_hdl == NULL)
		return t_uid - 1;

	/* backward search: uid_min <-- t_uid  */
	gpwd_hdl = NULL;
	t_uid = tt_uid;
	do {
		gpwd_hdl = bdb_getdbuid_pwd(t_uid, gpwd);
		if ((gpwd->g_time < oldest_time || oldest_time == -1)
		    && gpwd_hdl) {
			memcpy(&oldest_gtk, &(gpwd->g_idtoken),
			       sizeof(GIDTOKEN));
			oldest_uid = t_uid;
			oldest_time = gpwd->g_time;
		}
		t_uid--;
	} while ((t_uid >= uid_min) && gpwd_hdl);

	if (gpwd_hdl == NULL)
		return t_uid + 1;

	/* if id space is full, choose the oldest one for allocation */
	if (bdb_deldbgtk_pwd(&oldest_gtk))
		AMS_DEBUG("Can not delete GPASSWD by oldest uid");

	return oldest_uid;

}

/** find and allocate a free gid */
gid_t _xos_get_local_freegid(int gid_max, int gid_min)
{
	static unsigned int seed;
	int t_gid, tt_gid;
	char *ggrp_hdl = NULL;
	GGROUPS *ggrp = ggrp_new();
	GGRPTOKEN oldest_gtk;
	long oldest_time = -1;
	gid_t oldest_gid = -1;

	if (_init == 0) {

		seed = (unsigned int) time((time_t *) 0);
		_init = 1;
	}


	tt_gid = t_gid = (gid_t) (gid_min +
				  (int) (((float) (gid_max - gid_min)) *
					 rand_r(&seed) / (RAND_MAX +
							  1.0)));

	/* test the gid slot */

	/* forward search: t_gid --> gid_max  */
	do {
		ggrp_hdl = bdb_getdbgid_grp(t_gid, ggrp);
		if ((ggrp->g_time < oldest_time || oldest_time == -1)
		    && ggrp_hdl) {
			memcpy(&oldest_gtk, &(ggrp->g_grptoken),
			       sizeof(GGRPTOKEN));
			oldest_gid = t_gid;
			oldest_time = ggrp->g_time;
		}
		t_gid++;
	} while ((t_gid <= gid_max) && ggrp_hdl);

	if (ggrp_hdl == NULL)
		return t_gid - 1;

	/* backward search: gid_min <-- t_gid  */
	ggrp_hdl = NULL;
	t_gid = tt_gid;
	do {
		ggrp_hdl = bdb_getdbgid_grp(t_gid, ggrp);
		if ((ggrp->g_time < oldest_time || oldest_time == -1)
		    && ggrp_hdl) {
			memcpy(&oldest_gtk, &(ggrp->g_grptoken),
			       sizeof(GGRPTOKEN));
			oldest_gid = t_gid;
			oldest_time = ggrp->g_time;
		}
		t_gid--;
	} while ((t_gid >= gid_min) && ggrp_hdl);

	if (ggrp_hdl == NULL)
		return t_gid + 1;

	/* if id space is full, choose the oldest one for allocation */
	if (bdb_deldbgtk_grp(&oldest_gtk))
		AMS_DEBUG("Can not delete GPASSWD by oldest gid");

	return oldest_gid;

}

int find_matchrule_mapping(int rtype, char *sub_1, char *sub_2,
			   MAPTOKEN * mtk)
{
	int ret = -1;
	struct subject_t map_sub;
	struct rule_t *map_rule = NULL;

	if (!sub_1 || !sub_2 || !mtk)
		return ret;

	if (rtype < XPLY_ST_NUL || rtype > XPLY_ST_GRP)
		return ret;

	map_sub.sub_type = rtype;
	strncpy(map_sub.sub_content_1, sub_1, XPLY_MAXLEN);
	strncpy(map_sub.sub_content_2, sub_2, XPLY_MAXLEN);

	if (xos_policy_findrule(XPLY_RT_MA, &map_sub, &map_rule, NULL)) {
		AMS_DEBUG("Unfound the match rule!!");
		return ret;
	}

	sprintf(mtk->mapped_from, "%s:%s",
		map_rule->r_subject.sub_content_1,
		map_rule->r_subject.sub_content_2);
	sprintf(mtk->mapped_to, "%s",
		map_rule->r_object.obj_content.objc_object);

	if (!strncmp
	    (sub_1, map_rule->r_subject.sub_content_1, strlen(sub_1))
	    && !strncmp(sub_2, map_rule->r_subject.sub_content_2,
			strlen(sub_2)))
		mtk->mapped_mode = M_EXACT;
	else
		mtk->mapped_mode = M_BULK;
	if (mtk->mapped_to[0] == '*')
		mtk->mapped_type = T_ARBITRA;
	else
		mtk->mapped_type = T_USER;

	return 0;

}

int find_matchrule_policy(PLYTOKEN * ptk)
{
	int ret = -1;
	struct rule_t *xrule = NULL;

	if (!ptk)
		return ret;

	if (ruleinfo_db_getbykey(-1, &xrule, NULL)) {
		AMS_DEBUG("Can not get setting from databases");
		return ret;
	}

	if (!xrule)
		return -1;

	ptk->uid_max = atoi(xrule->r_subject.sub_content_1);
	ptk->uid_min = atoi(xrule->r_subject.sub_content_2);
	ptk->gid_max = atoi(xrule->r_content.rc_driver.driver_name);
	ptk->gid_min = atoi(xrule->r_content.rc_driver.driver_params);

	return 0;
}

/** Account mapping: mapping to existed account
 * The function is called by accountmapping()
 * @param mtk The \a mtk holds the mapping type and mapping mode of current mapping.
 * @param gpwd The mapping information to be stored.
 * @return 0 if success, or non-zero
 */
int accountmapping_to_exist(MAPTOKEN * mtk, GPASSWD * gpwd)
{

	struct passwd pw, *pw_ptr = NULL;
	char pw_buf[MAXBUF];


	if (getpwnam_r
	    (mtk->mapped_to, &pw, pw_buf, sizeof(pw_buf), &pw_ptr) != 0
	    || !pw_ptr)
		return -E_NOEXIST;

	gpwd->l_idtoken.g_mappeduid = pw.pw_uid;
	gpwd->l_idtoken.g_mappedgid = pw.pw_gid;

	strncpy(gpwd->l_idtoken.g_mappedname, pw.pw_name,
		strlen(pw.pw_name) + 1);
	strncpy(gpwd->g_shell, pw.pw_shell, strlen(pw.pw_shell) + 1);
	strncpy(gpwd->g_home, pw.pw_dir, strlen(pw.pw_dir) + 1);

	return 0;
}





/** Mapping to unexisted account
 * The function do maps the account to a virtual account which is not
 * stroed in /etc/passwd. But kernel can recognize the account.
 * @param mtk The mapping information of current mapping rule
 * @param gpwd The GPASSWD struct
 */
int accountmapping_to_noexist(MAPTOKEN * mtk, GPASSWD * gpwd)
{
	char *policy_file;
	int size, ret;
	PLYTOKEN *ptk = NULL;
	//static unsigned int seed;

	size = strlen(DEFAULT_MAPFILEDIR) + strlen(POLICYFILE) + 1;
	policy_file = malloc(size);
	snprintf(policy_file, size, "%s%s", DEFAULT_MAPFILEDIR,
		 POLICYFILE);

	ptk = ptoken_new();
	if (find_matchrule_policy(ptk)) {
		AMS_DEBUG("%s: find_matchrule_policy fail", __FUNCTION__);
		ret = E_NOEXIST;
		goto err;
	}
#if 0
	if (_init == 0) {

		seed = (unsigned int) time((time_t *) 0);
		_init = 1;
	}


	gpwd->l_idtoken.g_mappeduid =
	    (uid_t) (ptk->uid_min +
		     (int) (((float) (ptk->uid_max - ptk->uid_min)) *
			    rand_r(&seed) / (RAND_MAX + 1.0)));
	gpwd->l_idtoken.g_mappedgid =
	    (gid_t) (ptk->gid_min +
		     (int) (((float) (ptk->gid_max - ptk->gid_min)) *
			    rand_r(&seed) / (RAND_MAX + 1.0)));;

#else
        gpwd->l_idtoken.g_mappeduid =
	            _xos_get_local_freeuid(ptk->uid_max, ptk->uid_min);
        gpwd->l_idtoken.g_mappedgid =
	            _xos_get_local_freegid(ptk->gid_max, ptk->gid_min);
#endif
	sprintf(gpwd->l_idtoken.g_mappedname, "xosuser_u%d",
		gpwd->l_idtoken.g_mappeduid);
	strncpy(gpwd->g_shell, "/bin/bash", MAXFIELD);
	strncpy(gpwd->g_home, "/tmp", MAXFIELD);

	ret = 0;

      err:
	return ret;
}

/** Check whether the mapping token is right
 * @param mtk the mapping token
 * @param is_group whether the mapping token is typed by group mapping
 * @return zero if success, or non-zero
 */
int check_mappedtoken(MAPTOKEN * mtk, int is_group)
{
	int ret = 0;

	if (!mtk)
		return 1;

	switch (mtk->mapped_type) {
	case T_USER:
		if (mtk->mapped_mode != M_EXACT && !is_group) {
			AMS_DEBUG
			    ("account mapping: one-by-one for mapping to exist user");
			ret = 1;
		}
		break;
#if 0
	case T_POOL:

		if (is_group)	// group is not allowed to mapping pool
		{
			AMS_DEBUG
			    ("group mapping: error in mapping group to pool");
			ret = 1;
		}

		break;
#endif
	case T_ARBITRA:
		break;
	default:
		AMS_DEBUG("Not unkown type");
		ret = 1;
	}

	return ret;
}


/** Mapping user's global identity to local account.
 *
 * The function maps user's identity token [DN,VO,ROLE] to local account. Exactly,
 * map user to a local user which has a uid and a primary gid.
 *
 * @param mapped_dn user's DN
 * @param mapped_vo VO user belonged to
 * @param mapped_role user's role
 *
 * @return non-zero if errors occur, or zero
 */
int accountmapping(char *mapped_vo, char *mapped_dn, char *mapped_role,
		   GPASSWD * gpwd)
{
	char *mapped_string = NULL, *mapped_file = NULL;
	char *skey = NULL;
	int size, ret = -E_ERRORS;
	GPASSWD *tgpwd = NULL;
	MAPTOKEN *mtk = NULL;
	GIDTOKEN gtk;


	tgpwd = gpwd_new();
	if (!gpwd || !tgpwd) {
		AMS_DEBUG("%s: gpwd_new() fail", __FUNCTION__);
		goto err;
	}
	/* searching exist mapped GPASSWD */
	strncpy(gtk.g_dn, mapped_dn, strlen(mapped_dn) + 1);
	strncpy(gtk.g_vo, mapped_vo, strlen(mapped_vo) + 1);
	if (mapped_role)
		strncpy(gtk.g_role, mapped_role, strlen(mapped_role) + 1);
	else
		strncpy(gtk.g_role, "OTHER", MAXFIELD);


	skey = bdb_getdbgtk_pwd(&gtk, gpwd);
	if (skey) {
		/* found the mapped GPASSWD */
                gpwd->g_time = time(NULL);
                if (bdb_deldbgtk_pwd(&gtk) || bdb_storedben_pwd(gpwd))
                        AMS_DEBUG("fail to update timestamp in PASSWD\n");

		ret = 0;
		goto err;
	}

	strncpy(gpwd->g_idtoken.g_dn, mapped_dn, MAXFIELD);
	strncpy(gpwd->g_idtoken.g_vo, mapped_vo, MAXFIELD);
	if (mapped_role)
		strncpy(gpwd->g_idtoken.g_role, mapped_role,
			strlen(mapped_role));
	else
		strncpy(gpwd->g_idtoken.g_role, "OTHER", MAXFIELD);


	size = strlen(mapped_vo) + strlen(mapped_dn) + 2;
	mapped_string = malloc(size);
	snprintf(mapped_string, size, "%s:%s", mapped_vo, mapped_dn);

	size = strlen(DEFAULT_MAPFILEDIR) + strlen(AMAPPEDFILE) + 1;
	mapped_file = malloc(size);
	snprintf(mapped_file, size, "%s%s", DEFAULT_MAPFILEDIR,
		 AMAPPEDFILE);

	mtk = mtoken_new();
	if (find_matchrule_mapping(1, mapped_dn, mapped_vo, mtk)) {
		AMS_DEBUG("%s: find_matchrule_mapping fail\n", __FUNCTION__);
		if (mtk)
			free(mtk);
		goto err;
	}


	ret = check_mappedtoken(mtk, 0);
	if (ret) {
		AMS_DEBUG("%s:Mapping rule is wrong", __FUNCTION__);
		goto err;
	}

	gpwd->g_time = time(NULL);

	do {
		switch (mtk->mapped_type) {
		case T_USER:	/* mapping to exact account */
			ret = accountmapping_to_exist(mtk, gpwd);
			if (ret) {
				// TID free(gpwd);
				//goto err;
			}
			break;
		case T_ARBITRA:	/* bulk mapping  */
			ret = accountmapping_to_noexist(mtk, gpwd);
			if (ret) {
				// TID free(gpwd);
				//goto err;
			}
			break;
		default:
			AMS_DEBUG("unknown type");
			free(gpwd);
		}
	} while (ret == -E_NEXTRULE);

	if (ret) {
		goto err;
	}


      err:
	if (mapped_file)
		free(mapped_file);
	if (mapped_string)
		free(mapped_string);
	if (tgpwd)
		free(tgpwd);
	return ret;
}

/** Group mapping: mapping to exist group
 * @param mtk the mapping token
 * @param ggrp the GGROUPS struct
 * @return zero if success, or non-zero
 */
int groupmapping_to_exist(MAPTOKEN * mtk, GGROUPS * ggrp)
{
	char cmd[MAXFIELD];
	char *account = NULL;

	char *policy_file;
	int size, ret, assign_gid = -1;
	PLYTOKEN *ptk = NULL;
	static unsigned int seed;


	size = strlen(DEFAULT_MAPFILEDIR) + strlen(POLICYFILE) + 1;
	policy_file = malloc(size);
	snprintf(policy_file, size, "%s%s", DEFAULT_MAPFILEDIR,
		 POLICYFILE);

	ptk = ptoken_new();
	if (find_matchrule_policy(ptk)) {
		AMS_DEBUG("find_matchrule_policy fail");
		ret = E_NOEXIST;
		goto err;
	}
	if (_init == 0) {

		seed = (unsigned int) time((time_t *) 0);
		_init = 1;
	}


	account = strdup(mtk->mapped_to);

	struct group gp, *gp_ptr = NULL;
	char gp_buf[MAXBUF];

	if ((getgrnam_r
	     (account, &gp, gp_buf, sizeof(gp_buf),
	      &gp_ptr) == 0) && gp_ptr)
		goto exist_group;

	if (getuid() != 0) {
		AMS_DEBUG("root to run AMS");
		return -E_NOEXIST;
	}

	assign_gid = (ptk->gid_min +
		      (int) (((float) (ptk->gid_max - ptk->gid_min)) *
			     rand_r(&seed) / (RAND_MAX + 1.0)));

	sprintf(cmd, "/usr/sbin/groupadd  -g %d %s", assign_gid, account);
	system(cmd);

	/* getgrnam() should be replaced by getgrnam_r */
	if ((getgrnam_r
	     (account, &gp, gp_buf, sizeof(gp_buf),
	      &gp_ptr) == 0) && !gp_ptr) {
		AMS_DEBUG("Cann't add group");
		ret = -E_NEXTRULE;
		goto err;
	}

      exist_group:
	DEBUG_ASSERT(gp_ptr != NULL);
	DEBUG_ASSERT(ggrp != NULL);
	/** \todo add current account in group nember list */
	strncpy(ggrp->l_grptoken.g_mappedname, account,
		strlen(account) + 1);

	ggrp->l_grptoken.g_mappedgid = gp_ptr->gr_gid;
	ret = 0;
      err:
	return ret;
}


/** Group mapping: mapping to virtual group
 *
 * @param mtk the mapping token
 * @param ggrp the GGROUPS struct
 * @return zero if success, or non-zero 
 */
int groupmapping_to_noexist(MAPTOKEN * mtk, GGROUPS * ggrp)
{
	char *policy_file;
	int size, ret;
	PLYTOKEN *ptk = NULL;
//	static unsigned int seed;


	size = strlen(DEFAULT_MAPFILEDIR) + strlen(POLICYFILE) + 1;
	policy_file = malloc(size);
	snprintf(policy_file, size, "%s%s", DEFAULT_MAPFILEDIR,
		 POLICYFILE);

	ptk = ptoken_new();
	if (find_matchrule_policy(ptk)) {
		AMS_DEBUG("find_matchrule_policy fail");
		ret = E_NOEXIST;
		goto err;
	}
#if 0
	if (_init == 0) {

		seed = (unsigned int) time((time_t *) 0);
		_init = 1;
	}


	ggrp->l_grptoken.g_mappedgid =
	    (gid_t) (ptk->gid_min +
		     (int) (((float) (ptk->gid_max - ptk->gid_min)) *
			    rand_r(&seed) / (RAND_MAX + 1.0)));

#else
	ggrp->l_grptoken.g_mappedgid =
	      _xos_get_local_freegid(ptk->gid_max, ptk->gid_min);
#endif	
	sprintf(ggrp->l_grptoken.g_mappedname, "xosuser_g%d",
		ggrp->l_grptoken.g_mappedgid);

	ret = 0;
      err:
	if (policy_file)
		free(policy_file);
	return ret;
}



/** Mapping user's global identity to local group.
 *
 * The function focuses the mapping from user's global identity information to
 * local group. The information used in group mapping is VO identity, subgroup
 * user's belonged to.
 *
 * @param mapped_vo VO identity user belonged to
 * @param mapped_subgroup Subgroup user belonged to
 *
 * @return NULL if not success, or a pointer of GGROUPS struct.
 */

int groupmapping(char *mapped_vo, char *mapped_subgroup, GGROUPS * ggrp)
{
	char *mapped_string = NULL, *mapped_file = NULL;
	char *skey = NULL;
	int size, ret = -E_ERRORS;
	GGROUPS *tggrp = NULL;
	MAPTOKEN *mtk = NULL;
	GGRPTOKEN tgtk;

	DEBUG_ASSERT(mapped_subgroup != NULL);
	DEBUG_ASSERT(mapped_vo != NULL);

	if (!ggrp) {
		AMS_DEBUG("%s: ggrp is null\n", __FUNCTION__);
		goto err;
	}

	tggrp = ggrp_new();
	if (!tggrp) {
		AMS_DEBUG("%s: ggrp_new() fail\n", __FUNCTION__);
		goto err;
	}


	strncpy(tgtk.g_subgroup, mapped_subgroup,
		strlen(mapped_subgroup) + 1);
	strncpy(tgtk.g_vo, mapped_vo, strlen(mapped_vo) + 1);
	skey = bdb_getdbgtk_grp(&tgtk, ggrp);
	if (skey) {
		/*AMS_DEBUG("%s: found existed mapped GROUPS", __FUNCTION__); */
                ggrp->g_time = time(NULL);
                if (bdb_deldbgtk_grp(&tgtk) || bdb_storedben_grp(ggrp))
                        AMS_DEBUG("fail to update timestamp in GGROUPS\n");
		ret = 0;
		goto err;
	}

	strncpy(ggrp->g_grptoken.g_subgroup, mapped_subgroup,
		strlen(mapped_subgroup));
	strncpy(ggrp->g_grptoken.g_vo, mapped_vo, strlen(mapped_vo));

	size = strlen(mapped_vo) + strlen(mapped_subgroup) + 2;
	/*
	mapped_string = malloc(size);
	snprintf(mapped_string, size, "%s:%s", mapped_vo, mapped_subgroup);

	size = strlen(DEFAULT_MAPFILEDIR) + strlen(GMAPPEDFILE) + 1;
	mapped_file = malloc(size);
	snprintf(mapped_file, size, "%s%s", DEFAULT_MAPFILEDIR,
		 GMAPPEDFILE);
	*/


	do {

		mtk = mtoken_new();
		if (find_matchrule_mapping
		    (2, mapped_vo, mapped_subgroup, mtk)) {
			AMS_DEBUG("%s: find_matchrule_mapping fail\n",
				  __FUNCTION__);
			goto err;
		}

		ret = check_mappedtoken(mtk, 1);
		if (ret)
			break;

		switch (mtk->mapped_type) {
		case T_USER:	/* mapping to exact group */
			ret = groupmapping_to_exist(mtk, ggrp);
			break;
#if 0
		case T_POOL:
			ret = groupmapping_to_pool(mtk, ggrp);
			break;
#endif
		case T_ARBITRA:	/* bulk mapping */
			ret = groupmapping_to_noexist(mtk, ggrp);
			break;
		default:
			AMS_DEBUG("unknown type");
		}

		if (!ret)
			break;
		if (ret == -E_NEXTRULE) {	/* refind the mapping rule */
		  /*size = strlen(mtk->mapped_from) + 1;
			mapped_string = malloc(size);
			snprintf(mapped_string, size, "%s",
				 mtk->mapped_from);
				 mapped_string[size - 2] = '*';*/
		  free(mtk);
		  mtk = NULL;
		}

	} while (ret == -E_NEXTRULE);

	if (ret) {
		goto err;
	}

      err:
	if(tggrp)
	  free(tggrp);
	if (skey)
	  free(skey);
	if (mtk)
	  free(mtk);
	return ret;
}

int clear_group_membership(GGROUPS * old_ggrp, GGROUPS * new_ggrp,
			   GGROUPS * null_ggrp)
{
	/* nothing to do */
	if (!old_ggrp || !new_ggrp)
		return 0;

	char group_mb[63], group_mblst[MAXFIELD], *p1 = NULL, *p2 = NULL;
	int ret = -1;

	sprintf(group_mb, "%s", new_ggrp->l_grptoken.g_members);
	sprintf(group_mblst, "%s", old_ggrp->l_grptoken.g_members);
	p1 = strstr(group_mblst, group_mb);
	if (p1) {
		p2 = p1 + strlen(group_mb);
		if (p1 != group_mblst) {
			if (*(p1 - 1) == ','
			    && (*p2 == ',' || *p2 == '\0')) {
				*(p1 - 1) = '\0';
				if (*p2 != '\0')
					sprintf(group_mblst, "%s%s",
						group_mblst, p2);
			}

		} else {
			if (strlen(group_mblst) == strlen(group_mb))
				group_mblst[0] = '\0';
			else if (*p2 == ',')
				strncpy(group_mblst, p2 + 1, MAXFIELD);
		}
		if (strlen(group_mblst) <
		    strlen(old_ggrp->l_grptoken.g_members)) {
			sprintf(old_ggrp->l_grptoken.g_members, "%s",
				group_mblst);
			bdb_deldbgtk_grp(&(old_ggrp->g_grptoken));
			if (bdb_storedben_grp(old_ggrp)) {
				AMS_DEBUG
				    ("Can not change the group member");
				goto err;
			}
		}
	}

	ret = 0;
      err:
	return ret;
}
/** Clear mapped account in database
 * @param gpwd The  GPASSWD struct mapped from user's global identity information
 * @retval 0 if clear successfully
 * @retval !0 if error in clearing.
 */
int clear_account(GPASSWD * gpwd)
{


	int ret = 0;
	char *skey = NULL;
	GPASSWD *tgpwd = NULL;

	if (!gpwd)
		return 0;
	/* find the record in GPASSWD */
	tgpwd = gpwd_new();
	skey = bdb_getdbgtk_pwd(&(gpwd->g_idtoken), tgpwd);
	if (!skey) {
		ret = -E_NOEXIST;
		goto err;
	}
	/* remove the record in GPASSWD */
	bdb_deldbgtk_pwd(&(gpwd->g_idtoken));

	GGROUPS *tggrp = NULL;

	tggrp = ggrp_new();
	sprintf(tggrp->l_grptoken.g_members, "%d",
		tgpwd->l_idtoken.g_mappeduid);
	if (bdb_operatedb_grp
	    (tggrp, NULL, (OPTGRPFUNC) clear_group_membership)) {
		AMS_DEBUG("Fail to operate database");
		goto err;
	}

	ret = 0;

      err:
	if (tgpwd)
		free(tgpwd);
	return ret;
}

/**  Clear mapped group in database.
 * @param ggrp The GGROUP struct mapped by user's global identity.
 * @retval 0 clear successfully
 * @retval !0 error
 */
int clear_group(GGROUPS * ggrp)
{
	int ret;
	char *skey, *account = NULL;
	struct group *gp = NULL;
	char cmd[MAXFIELD];
	GGROUPS *tggrp = NULL;

	if (!ggrp)
		return 0;

	tggrp = ggrp_new();
	skey = bdb_getdbgtk_grp(&(ggrp->g_grptoken), tggrp);
	if (!skey) {
		ret = -E_NOEXIST;
		goto err;
	}
	/* remove the record in group database */
	bdb_deldbgtk_grp(&(tggrp->g_grptoken));

	if (!strncmp
	    (tggrp->l_grptoken.g_mappedname, "xosuser_g",
	     strlen("xosuser_g")))
		return 0;
	else
		account = strdup(tggrp->l_grptoken.g_mappedname);

	char *policy_file;
	int size;
	PLYTOKEN *ptk = NULL;

	size = strlen(DEFAULT_MAPFILEDIR) + strlen(POLICYFILE) + 1;
	policy_file = malloc(size);
	snprintf(policy_file, size, "%s%s", DEFAULT_MAPFILEDIR,
		 POLICYFILE);

	ptk = ptoken_new();
	if (find_matchrule_policy(ptk)) {
		AMS_DEBUG("find_matchrule_policy fail");
		ret = E_NOEXIST;
		goto err;
	}

	/* remove the group from /etc/group */
	gp = getgrnam(account);
	if (gp) {
		if (getuid() != 0) {
			AMS_DEBUG("root to run AMS");
			ret = -E_NOEXIST;
			goto err;
		}
		/* if the group has its members, it should not be removed */
		if (gp->gr_mem && gp->gr_mem[0] != NULL) {
		} else if (gp->gr_gid >= ptk->gid_min
			   && gp->gr_gid <= ptk->gid_max) {
			sprintf(cmd, "/usr/sbin/groupdel  %s", account);
			system(cmd);
		}
	}

	ret = 0;
      err:
	if (tggrp)
		free(tggrp);
	return ret;

}

/** Clear all the item in database.
 *
 * @return 0 if success, or !0
 */
int clear_account_t(GPASSWD * gpwd1, GPASSWD * r_gpwd, GPASSWD * gpwd2)
{
	return clear_account(gpwd1);
}

int clear_group_t(GGROUPS * ggrp1, GGROUPS * r_grp, GGROUPS * ggrp2)
{
	return clear_group(ggrp1);
}

int clear_all()
{
	int exit_ret, ret;

	exit_ret = ret = 0;
	ret = bdb_operatedb_pwd(NULL, NULL, (OPTPWDFUNC) clear_account_t);
	if (ret)
		exit_ret = ret;


	ret = bdb_operatedb_grp(NULL, NULL, (OPTGRPFUNC) clear_group_t);
	if (!exit_ret)
		exit_ret = ret;
	return exit_ret;
}


char * _get_GUID_from_DN(char * temp_dn);

/** Mapping user's global identity to local one. This function is the main function
 *  to finish the mapping.
 *
 * @param mapped_dn User's DN for mapping
 * @param mapped_vo VO identity user belonged to
 * @param mapped_role user's role
 * @param mapped_subgroup the subgroup in VO, which can be multiple group name as
 *        the input format: vo-subgroup-1/vo-subgroup-2/...
 *
 * @return NULL if error occurred, or a GPASSWD struct holding the mapped user's
 * informatoin
 */

int xos_mapping(char *mapped_dn, char *mapped_vo, char *mapped_role,
		char *mapped_subgroup, GPASSWD * gpwd,
		GGROUPS * primarygrp)
{
	GPASSWD *tgpwd = NULL;
	GGROUPS *primary_grp = NULL, *second_grp = NULL, *tgrp = NULL;
	char *p = NULL, *buf = NULL, *skey = NULL;

	int ret = -E_ERRORS;
	char mapped_dn_t[MAXFIELD], mapped_vo_t[MAXFIELD],
	    mapped_role_t[MAXFIELD], mapped_id[MAXFIELD];

	tgpwd = gpwd_new();
	tgrp = ggrp_new();
	primary_grp = ggrp_new();
	if (!mapped_vo) {
		AMS_DEBUG("Undefine login VO");
		goto err;
	} else
		strcpy(mapped_vo_t, mapped_vo);

	if (mapped_dn)
		strcpy(mapped_dn_t, mapped_dn);

	int ret_grp = -1;
	if (mapped_role)
		strcpy(mapped_role_t, mapped_role);
	else
		strcpy(mapped_role_t, "OTHER");
	
	/* XtreemFS support */
	if ((ret_grp = groupmapping(mapped_vo, mapped_role, primary_grp))) {
		AMS_DEBUG
		    ("No rule for mapping '%s' to primary group",
		     mapped_role);
		goto err;
	}
	
	/* mapping to local account */
	if (mapped_dn) {

		ret =
		    accountmapping(mapped_vo_t, mapped_dn_t,
				   mapped_role_t, gpwd);
		if (ret < 0) {
			AMS_DEBUG("No rule for mapping account");
			goto err;
		}

		if (!ret_grp) {
			gpwd->l_idtoken.g_mappedgid =
			    primary_grp->l_grptoken.g_mappedgid;
			strncpy(gpwd->g_idtoken.g_role, mapped_role,
				MAXFIELD);
			bdb_deldbgtk_pwd(&(gpwd->g_idtoken));


		} else
			strncpy(gpwd->g_idtoken.g_role, "OTHER", MAXFIELD);

		/* set the GUID in gpwd*/
		strncpy(gpwd->g_idtoken.g_guid, _get_GUID_from_DN(mapped_dn_t), MAXFIELD);
		
		if (!bdb_getdbgtk_pwd(&(gpwd->g_idtoken), tgpwd)) {
			gpwd->g_time = time(NULL);
			if (bdb_storedben_pwd(gpwd)) {
				AMS_DEBUG("%s: fail in store gpwd",
					  __FUNCTION__);
				free(gpwd);
				gpwd = NULL;
				goto err;
			}
		}

	}


	sprintf(mapped_id, "%d", gpwd->l_idtoken.g_mappeduid);

	if (!ret_grp) {
		DEBUG_ASSERT(primary_grp->l_grptoken.g_mappedname[0] !=
			     '\0');
		DEBUG_ASSERT(primary_grp->l_grptoken.g_mappedgid != 0);

		if (!ret) {
			if (primary_grp->l_grptoken.g_members[0] == '\0')
				sprintf(primary_grp->l_grptoken.g_members,
					"%d", gpwd->l_idtoken.g_mappeduid);
			else if(!strstr((primary_grp->l_grptoken.g_members), mapped_id))
				sprintf(primary_grp->l_grptoken.g_members,
					"%s,%d",
					primary_grp->l_grptoken.g_members,
					gpwd->l_idtoken.g_mappeduid);

			bdb_deldbgtk_grp(&(primary_grp->g_grptoken));

		}
		skey = bdb_getdbgtk_grp(&(primary_grp->g_grptoken), tgrp);

		if (!skey) {
			primary_grp->g_time = time(NULL);
			if (bdb_storedben_grp(primary_grp)) {
				AMS_DEBUG("%s: fail in store ggrp",
					  __FUNCTION__);

			}
		}

		if (primarygrp) {
			strcpy(primarygrp->g_grptoken.g_vo,
			       primary_grp->g_grptoken.g_vo);
			strcpy(primarygrp->g_grptoken.g_subgroup,
			       primary_grp->g_grptoken.g_subgroup);
			strcpy(primarygrp->l_grptoken.g_mappedname,
			       primary_grp->l_grptoken.g_mappedname);
			primarygrp->l_grptoken.g_mappedgid =
			    primary_grp->l_grptoken.g_mappedgid;
			strcpy(primarygrp->l_grptoken.g_members,
			       primary_grp->l_grptoken.g_members);
			primarygrp->g_time = primary_grp->g_time;

		}
	}

	/* mapping to second group */
	if (mapped_subgroup) {

		buf = (char *) malloc(strlen(mapped_subgroup) + 1);
		if (!buf)
			goto err;
		sscanf(mapped_subgroup, "%s", buf);

		p = strtok(buf, "/");
		/*AMS_DEBUG("%s: subgroup -> %s,%d", __FUNCTION__, p,
		   strlen(p)); */
		second_grp = ggrp_new();
		ret_grp = groupmapping(mapped_vo, p, second_grp);
		if (ret_grp) {
			AMS_DEBUG
			    ("No rule for mapping '%s' to second group",
			     p);
		} else {
			if (!ret) {
				if (second_grp->l_grptoken.g_members[0] ==
				    '\0')
					sprintf(second_grp->l_grptoken.
						g_members, "%d",
						gpwd->l_idtoken.
						g_mappeduid);
				else if (!strstr(second_grp->l_grptoken.g_members,mapped_id))
					sprintf(second_grp->l_grptoken.
						g_members, "%s,%d",
						second_grp->l_grptoken.
						g_members,
						gpwd->l_idtoken.
						g_mappeduid);

				bdb_deldbgtk_grp(&
						 (second_grp->g_grptoken));
			}
			skey = NULL;
			skey =
			    bdb_getdbgtk_grp(&(second_grp->g_grptoken),
					     tgrp);
			if (!skey) {
				second_grp->g_time = time(NULL);
				if (bdb_storedben_grp(second_grp)) {
					AMS_DEBUG("%s: fail in store ggrp",
						  __FUNCTION__);
					free(second_grp);
				}
			}
		}

		while ((p = strtok(NULL, "/"))) {
			/*AMS_DEBUG("%s: subgroup -> %s,%d", __FUNCTION__, p,
			   strlen(p)); */
			second_grp = ggrp_new();
			if (groupmapping(mapped_vo, p, second_grp)) {
				AMS_DEBUG
				    ("No rule for mapping '%s' to second group",
				     p);
				continue;
			} else {
				if (!ret) {
					if (second_grp->l_grptoken.
					    g_members[0] == '\0')
						sprintf(second_grp->
							l_grptoken.
							g_members, "%d",
							gpwd->l_idtoken.
							g_mappeduid);
					else if (!strstr(second_grp->l_grptoken.g_members,mapped_id))
						sprintf(second_grp->
							l_grptoken.
							g_members, "%s,%d",
							second_grp->
							l_grptoken.
							g_members,
							gpwd->l_idtoken.
							g_mappeduid);
					bdb_deldbgtk_grp(&
							 (second_grp->
							  g_grptoken));
				}
				skey = NULL;
				skey =
				    bdb_getdbgtk_grp(&
						     (second_grp->
						      g_grptoken), tgrp);
				if (!skey) {
					second_grp->g_time = time(NULL);
					if (bdb_storedben_grp(second_grp)) {
						AMS_DEBUG
						    ("%s: fail in store ggrp",
						     __FUNCTION__);
						free(second_grp);
					}
				}
			}
		}

	}

	ret = 0;

      err:
	return ret;
}

char* _get_GUID_from_DN(char * temp_dn)
{
	char *p = NULL, *pp = NULL;
	char guid_t[512];
	int i = 0;

	if(!temp_dn) return NULL;

	memset(guid_t,0,512);

	p = strstr(temp_dn,"/CN=");
	if(!p) p = strstr(temp_dn, "/cn=");

	if(p) {
		pp = p + strlen("/CN=");
		i = 0;
		while(pp) {
			if(isdigit(*pp) || isalpha(*pp) || *pp == '-')
				i++;
			else if(*pp == '/') break;
			pp++;
		}
		strncpy(guid_t,p+strlen("/CN="),i);
	}

	return strdup(guid_t);
}
