/** @file pam_app_conv.c
 * @brief Simple PAM-aware application - pam_app_conv.c
 *
 * To compile, you should install xtreemos-nss-pam-0.04.tar.gz.
 * Then in console,
 * 
 *  $ gcc -o pam_app_conv pam_app_conv.c -lpam -lpam_misc 
 *					-lxos_common -lxos_security
 *
 * To run the execuable program, you first have to add a configuration
 * file in /etc/pam.d/. The file name should be the same as program name
 * (for example, pam_app_conv).
 * Add the following lines in configuration file.
 *
 * auth 	sufficient	/path/to/pam_xos.so
 * account	sufficient	/path/to/pam_xos.so 
 * session	sufficient	/path/to/pam_xos.so
 *
 * And, you neet root privilege, since you
 * will change the uid/gid.
 *
 *  $ su
 *  # ./pam_app_conv  -pem /path/to/certificate
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include<sys/stat.h>
#include<unistd.h>

#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <security/pam_appl.h>	/* header for building PAM-aware application */

#include "xos_interfaces.h"	/* header of our interfaces */

#define EXT_ATTR_VO	"1.34.5.0.14.1"
#define EXT_ATTR_ROLE	"1.34.5.0.14.6"

/* data structure for exchanging data */
struct credentials {
	const char *uname;
	const char *passwd;
};

/* PAM conversation function */
static int pam_conv_func(int num_msg,
			 const struct pam_message **msg,
			 struct pam_response **resp, void *appdata_ptr)
{
	int i = 0;
	/* exchanged data of conversation function */
	struct credentials *user_creds =
	    (struct credentials *) appdata_ptr;
	struct pam_response *response = 0;

	/* check args */
	if (!resp || !msg || !user_creds)
		return PAM_CONV_ERR;
	if (!
	    (response =
	     (struct pam_response *) malloc(num_msg *
					    sizeof(struct pam_response))))
		return PAM_CONV_ERR;

	/* build each response */
	for (i = 0; i < num_msg; i++) {
		response[i].resp_retcode = 0;
		response[i].resp = 0;
		switch (msg[i]->msg_style) {
		case PAM_PROMPT_ECHO_ON:
			/* assume it's asking for uname */
			response[i].resp = strdup(user_creds->uname);
			break;
		case PAM_PROMPT_ECHO_OFF:
			/* assume it's asking for passwd */
			response[i].resp = strdup(user_creds->passwd);
			break;
		default:
			if (response)
				free(response);
			return PAM_CONV_ERR;
		}
	}
	*resp = response;
	return PAM_SUCCESS;
}


//////////////////////////////// main function //////////////////////////////////////// 
void usage(char *grpnam);
void self_solution();

int main(int argc, char *argv[])
{

	pam_handle_t *pamh;
	struct passwd *pw = NULL;
	uid_t uid;
	int res,non_root = 0;
	unsigned char *proxy_buffer = NULL;
	struct stat file_stat;

	int proxy_len;

	struct pam_conv my_conv;
	struct credentials my_creds;
	XOS_CREDS *creds = NULL;
	char dn_buffer[1024];


	char *proxy_path = NULL;
	int i, status;
	uid_t pid, pid_1;


	/* parse parameters */
	if(argc < 2) {
		usage(argv[0]);
		return -1;
	}

	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-pem")) {
			proxy_path = argv[++i];
		} else if (!strcmp(argv[i], "-u")) {
			non_root = 1;
		} else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
			usage(argv[0]);
			return 0;
		} else if (!strcmp(argv[i], "-solution")) {
			self_solution();
			return 0;
		} else { 
			printf("Unknown parameters '%s'! Please, check Usage\n",argv[i]);
		}
	}

	if (stat(proxy_path, &file_stat) != 0){
		fprintf(stderr,"Invalid -pem path '%s'\n",proxy_path);
		fprintf(stderr,"run with '-h' for usage\n");
		return 0;
	}

	if(getuid() != 0 && !non_root) {
		printf("Please run with root!\n"
			"Or, run with -u for non-privilege user.\n");
		return 0;
	}

	/* get proxy to buffer */
	if (proxy_path
	    && xos_cert_buffromfile(proxy_path, &proxy_buffer,
				    &proxy_len) != XOS_SUCCESS) {
		printf("Can not load certificate to buffer\n");
		exit(1);
	}
	/* prepare conversation with PAM */
	if (proxy_buffer) {
		/* assign data pointer to our customized data structure */
		my_creds.uname = "";
		my_creds.passwd = (char *) proxy_buffer;

		/* bind conversation function and data */
		my_conv.conv = &pam_conv_func;
		my_conv.appdata_ptr = (void *) &my_creds;
	}


	/* load proxy from file */
	creds = xos_creds_new();
	if (xos_cert_loadfromfile(creds, proxy_path) != XOS_SUCCESS) {
		printf("can not load proxy from file\n");
		exit(1);
	}

	/* load dn from certificate */
	if (xos_cert_subject_get(creds->certificate, dn_buffer, 1024) !=
	    XOS_SUCCESS) {
		printf("Can not get DN\n");
		exit(1);
	}

	/* load attributes from proxy */
	char *vop = NULL, *rolep = NULL;
	char ext_buf[1024] = { '\0' };
	char attr_vo_buf[1024] = { '\0' };
	char attr_role_buf[1024] = { '\0' };
	if (xos_cert_attributes_fetch_ext(creds->certificate, ext_buf) !=
	    XOS_SUCCESS) {
		DEBUG("PAM:Can not get extension from CDA certificate\n");
		/* Have not attributes ? */
	} else
	    if (xos_cert_attributes_fetch_attr
		(ext_buf, EXT_ATTR_VO, attr_vo_buf) != XOS_SUCCESS) {
		DEBUG("PAM:Can not get VO name from extensions\n");
		/* Have not vo name ? */
		strcpy(attr_vo_buf, "xtreemos");
	} else
	    if (xos_cert_attributes_fetch_attr
		(ext_buf, EXT_ATTR_ROLE, attr_role_buf) != XOS_SUCCESS) {
		DEBUG("PAM:Can not get ROLE from extensions\n");
		/* Have not vo name ? */
		strcpy(attr_role_buf, "admin");
	}

	vop = strdup(attr_vo_buf);
	rolep = strdup(attr_role_buf);
	printf("dn = [%s], vo = [%s], role = [%s]\n", dn_buffer, vop, rolep);

	uid = geteuid();
	pw = getpwuid(uid);

	/* PAM hook: start PAM
	 * pass the name of current user, pam handler, and conversation struct.  
	 */
	pam_start("pam_app_conv", pw->pw_name, &my_conv, &pamh);
	pam_set_item(pamh, PAM_TTY, "/dev/tty");

	/* PAM hook: do authentication with PAM 
	 * The hook will verify the user's certificate.
	 */
	if ((res = pam_authenticate(pamh, 0)) != PAM_SUCCESS) {
		/* Fail to authenticte user, denied */
		fprintf(stderr, "Oops: %s\n", pam_strerror(pamh, res));
		fprintf(stderr,"\nHints:\n");
		fprintf(stderr,"(1) Have you correctly configured /etc/pam.d/pam_app_conv ?\n");
		fprintf(stderr,"(2) Have you had valid certificate ?\n");
		exit(1);
	}


	/* PAM hook: do account management (apply account) 
	 * After authentication, the identity information will be fetched,
	 * and pass to the PAM hook, which will connect with AMS and finish
	 * user mapping (if the user has not been mapped).
	 *
	 * Another thing will be done here is policy translation. After
	 * mapping, the access rights of mapped account will be setuped
	 * before open session.
	 */
	res = pam_acct_mgmt(pamh, 0);
	if (res != PAM_SUCCESS) {
		/* Fail to apply account, denied */
		fprintf(stderr, "Oops: %s\n", pam_strerror(pamh, res));
		exit(1);
	}

	/* PAM hook: Open session 
	 * Setup the quota limitation prepared in last hook.
	 */
	res = pam_open_session(pamh, 0);
	if (res != PAM_SUCCESS) {
		/* Fail to open session, denied */
		fprintf(stderr, "Oops: %s\n", pam_strerror(pamh, res));
		exit(1);
	}

	char idtoken[2048];
	char vo_buffer[1024];
	char role_buffer[1024];
	strcpy(vo_buffer, vop);
	strcpy(role_buffer, rolep);
	sprintf(idtoken, "%s#%s#%s", dn_buffer, vo_buffer, role_buffer);
	/* get the mapped account information 
	 * It will call NSS module.
	 */
	pw = getpwnam(idtoken);
	if (!pw) {
		printf("Can not get pw\n");
		return -1;
	}

	switch ((pid = fork())) {
	case -1:
		printf(" fork dump \n");
		exit(1);
	case 0:
		/* child: give up privs and start a shell */
		/* set uid and groups */
		if(!non_root) {
			if (initgroups(pw->pw_name, pw->pw_gid) == -1) {
				printf("initgroups()...");
				exit(1);
			}
			if (setgid(pw->pw_gid) == -1) {
				printf("setgid()...");
				exit(1);
			}
			if (setuid(pw->pw_uid) == -1) {
				printf("setuid()...");
				exit(1);
			}
		}

		/* PAM hook: do credentials setup
		 * It setups the authenticated certificate in 
		 * keyring.
		 * You can add a command to see the change of
		 * keyring,for example 
		 *    system("keyctl show");
		 */
		pam_setcred(pamh, PAM_ESTABLISH_CRED);
		chdir(pw->pw_dir);

		pid_1 = fork();
		if (pid_1 == 0)
			execl("/bin/echo", "echo", "Sucess in PAM checking !", NULL);
		else if (pid_1 > 0)
			waitpid(pid_1, &status, 0);

		/* PAM hook: do credentials clear
		 * It clears the keyring and mapped account
		 * in databases.
		 */

		pam_setcred(pamh, PAM_DELETE_CRED);
		/* If you want to see the change
		 * in keyring, please add the command here:
		 *      system("keyctl show");
		 */
		exit(1);
	default:
		/* parent: wait for child to exit */
		waitpid(pid, &status, 0);

	}


	/* PAM hook: close the session 
	 * The hook also clear the quota limitation setuped 
	 * by pam_open_session.
	 */
	res = pam_close_session(pamh, 0);

	/* PAM hook: release the PAM resources */
	pam_end(pamh, res);

	return 0;
}

void usage(char *grpnam)
{
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"%s  [-u] -pem  <cert>\n",grpnam);
	fprintf(stderr,"%s  [-u] [-h|--help] [-solution]\n",grpnam);
	fprintf(stderr,
			"\t-pem <cert>\t\tpath of tested certificate\n"
			"\t-u\t\t\tRun command without root\n"
			"\t-h|--help\t\tGet help of command\n"
			"\t-solution\t\tGet solution hints for output result\n"
		);
}

void self_solution()
{
	fprintf(stderr, 
		"\n1. Error: unable to get local issuer certificate \n"
		"   Solution: check the configuration in pam_xos.conf:\n"
		"   	VOCACertDir             /path/to/CA_ROOT_DIR \n"
		"	VOCACertFile            <CA_SELFSIGN_CERT>	\n\n"
		"2. Error: Assertion `path != ((void *)0)' failed \n"
		"   Solution: check the command parameters used, please run --help [-h] to get Usage\n\n"
		"3. Error: bss_file.c:122:fopen('/etc/xos/testcerts/xosca//ca-list.pem','r')\n"
		"   Solution: a) check whether the 'ca-list.pem' is in given location\n"
		"	    b) check the configuration in pam_xos.conf, like Error 1.\n\n"
		);
}
