/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/* usage:
 * xos-policy-admin-ru -dn xxx -vo xxx -grp xxx
 *                    [-cpu | -mem | -nproc | -nofile]
 *                    [-abs_quo | -rel_quo] xxx
 *
 */

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include "xos_policy_db.h"
#include "xos_policy_mgmt.h"

#define RES_TYPE_CPU		0
#define RES_TYPE_MEM		1
#define RES_TYPE_NPROC		4
#define RES_TYPE_NOFILE		5


void usage(char *program)
{
	printf("Usage: Set resource usage for given users or groups\n\t%s [-dn | -grp] xxx -vo xxx"
	       " [-cpu | -mem | -nproc | -nofile] "
	       "[-abs_quo | -rel_quo] xxx\n", program);
	printf("\n\t[-dn | -grp]\tFor account rules, '-dn' will pass the DN,\n"
			"\t\t\tfor group rules, use '-grp' to pass subgroup\n");
	printf("\n\t-vo\t\tpass the VO id\n");
	printf("\n\t[-cpu | -mem | -nproc | -nofile]\n"
			"\t\t-cpu     - Set CPU resources\n"
			"\t\t-mem     - Set Memory resource\n"
			"\t\t-nproc   - Set maximal number of forked processes\n"
			"\t\t-nofile  - Set maximal number of opened files\n");
	printf("\n\t[-abs_quo | -rel_quo] xxx\n"
			"\t\t-abs_quo - absolute value\n"
			"\t\t-rel_quo - relative value\n");

	printf("\nExamples:\n");
	printf("%s  -dn \\* -vo dff93ffe-87f3-476a-b573-983f92623f95  -mem  -abs_quo  100\n",program);
	printf("%s  -grp \\* -vo dff93ffe-87f3-476a-b573-983f92623f95 -mem  -rel_quo  0.75\n",program);
}

int is_alldigit(char *str)
{
	char *p = str;
	while (*p != '\n' && *p != '\0') {
		if (!isdigit(*p))
			return 0;
		p++;
	}

	return 1;
}

int main(int argc, char *argv[])
{

	char *dn_exp, *vo_exp, *grp_exp;
	char *quoval_exp;
	struct rule_t xrule;
	int i, res_type, quoval_type;

	dn_exp = vo_exp = grp_exp = quoval_exp = NULL;
	res_type = quoval_type = -1;

	/* parse parameters */
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-dn")) {
			dn_exp = argv[++i];
		} else if (!strcmp(argv[i], "-vo")) {
			vo_exp = argv[++i];
		} else if (!strcmp(argv[i], "-grp")) {
			grp_exp = argv[++i];
		} else if (!strcmp(argv[i], "-cpu")) {
			if (res_type == -1)
				res_type = RES_TYPE_CPU;
			else
				printf("Parameter: '-cpu' error!\n");
		} else if (!strcmp(argv[i], "-mem")) {
			if (res_type == -1)
				res_type = RES_TYPE_MEM;
			else
				printf("Parameter: '-mem' error!\n");
		} else if (!strcmp(argv[i], "-nproc")) {
			if (res_type == -1)
				res_type = RES_TYPE_NPROC;
			else
				printf("Parameter: '-nproc' error!\n");
		} else if (!strcmp(argv[i], "-nofile")) {
			if (res_type == -1)
				res_type = RES_TYPE_NOFILE;
			else
				printf("Parameter: '-nofile' error!\n");
		} else if (!strcmp(argv[i], "-abs_quo")) {
			if (quoval_type == -1) {
				quoval_type = 0;
				quoval_exp = argv[++i];
			} else
				printf("Parameter: '-abs_quo' error!\n");
		} else if (!strcmp(argv[i], "-rel_quo")) {
			if (quoval_type == -1) {
				quoval_type = 1;
				quoval_exp = argv[++i];
			} else
				printf("Parameter: '-rel_quo' error!\n");
		} else if (!strcmp(argv[i], "-help")||!strcmp(argv[i], "--help")||!strcmp(argv[i], "-h")) {
			usage(argv[0]);
			return 0;
		} else
			printf("Unknown parameters !\n");
	}

	if (getuid() != 0) {
		printf("run with root privilege,thanks\n");
		return 0;
	}

	memset(&xrule, 0, sizeof(struct rule_t));

	if (!vo_exp) {
		printf("Missing parameters: -vo !\n");
		usage(argv[0]);
		return 1;
	}

	/* check and fill in the rule */
	if (dn_exp) {
		xrule.r_type = XPLY_RT_RU;

		/* Subject */
		xrule.r_subject.sub_type = XPLY_ST_ACT;
		sprintf(xrule.r_subject.sub_content_1, "%s", dn_exp);
		sprintf(xrule.r_subject.sub_content_2, "%s", vo_exp);

	} else if (grp_exp) {
		xrule.r_type = XPLY_RT_RU;

		/* Subject */
		xrule.r_subject.sub_type = XPLY_ST_GRP;
		sprintf(xrule.r_subject.sub_content_1, "%s", vo_exp);
		sprintf(xrule.r_subject.sub_content_2, "%s", grp_exp);

	} else {
		printf("Missing parameters: -dn or -grp\n");
		usage(argv[0]);
		return 1;
	}


	/* Object */
	if (res_type != -1) {
		xrule.r_object.obj_type = XPLY_OT_QUOTA;
		xrule.r_object.obj_content.objc_restype = res_type;

	} else {
		printf
		    ("Missing paramter: [-cpu | -mem | -nproc | -nofile]\n");
		return 1;
	}

	/* quota */
	if (quoval_type != -1) {
		xrule.r_content.rc_quota.quo_type = quoval_type;

		if (is_alldigit(quoval_exp))
			xrule.r_content.rc_quota.quo_val =
			    atof(quoval_exp);
		else {
			printf("Invalid parameter: '%s'\n", quoval_exp);
			return 1;
		}
	}

	xrule.r_valid = 1;
	xrule.r_time = time(0);

	char *handler = NULL;
	if (xos_policy_addrule(&xrule, &handler))
		printf("fail to add rule\n");


	return 0;
}
