/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file xos_readconf.c
 * @brief Interpret the PAM configure file in /etc/xos/pam_xos.conf
 * 
 *
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "xos_utils.h"
#include "xos_readconf.h"


/** Keyword tokens. */
typedef enum {
	iRemoteUnsecureLogin,
	iUseShareBuffer,
	iMyProxy,
	iProxyServer,
	iProxyServerPort,
	iLocalSecureLogin,
	iUsePass,
	iUseCertfile,
	iUseExternalAMS,
	iAMSServer,
	iAMSPort,
	iUseExternalAES,
	iAESServer,
	iAESPort,
	iVOACConf,
	iCSubsys,
	iVOCADir,
	iVOCACertF,
	iNodePrvKeyF,
	iUseDebug,
	iUseSyslog,
	iUseCgroups,
	iAutoMount,
	iFixedIDAlloc,
        iAMSServerIsLocal,
	iUnknownItem
} ConfItems;

/** Textual representations of the tokens. */
static struct {
	const char *name;
	ConfItems item;
} keywords[] = {
	{
	"RemoteUnsecureLogin", iRemoteUnsecureLogin}, {
	"UseShareBuffer", iUseShareBuffer}, {
	"MyProxy", iMyProxy}, {
	"ProxyServer", iProxyServer}, {
	"ProxyServerPort", iProxyServerPort}, {
	"LocalSecureLogin", iLocalSecureLogin}, {
	"UsePass", iUsePass}, {
	"UseCertfile", iUseCertfile}, {
	"UseExternalAMS", iUseExternalAMS}, {
	"AMSServer", iAMSServer}, {
	"AMSPort", iAMSPort}, {
	"UseExternalAES", iUseExternalAES}, {
	"AESServer", iAESServer}, {
	"AESPort", iAESPort}, {
	"VOACConf", iVOACConf}, {
	"Subsystems", iCSubsys}, {
	"VOCACertDir", iVOCADir}, {
	"VOCACertFile", iVOCACertF}, {
	"NodePrivateKeyfile", iNodePrvKeyF}, {
	"UseDebug", iUseDebug}, {
	"UseSyslog", iUseSyslog}, {
	"UseCgroups", iUseCgroups}, {
        "OpenAutoMount", iAutoMount}, {
	"FixedIDAlloc", iFixedIDAlloc}, {
        "AMSServerIsLocal", iAMSServerIsLocal}, {
	NULL, iUnknownItem}
};



/**parse_token() - parse the token
 * @param cp
 * @param filename configure file
 * @param linenum the number of current line
 * @Returns the number of the token pointed to by cp or iUnknownItem.
 */

static
ConfItems parse_token(const char *cp, const char *filename, int linenum)
{
	u_int i;

	for (i = 0; keywords[i].name; i++)
		if (strcasecmp(cp, keywords[i].name) == 0)
			return keywords[i].item;

	DEBUG("%s: line %d: Bad configuration option: %s",
	      filename, linenum, cp);
	return iUnknownItem;
}


/** character of return*/
#define WHITESPACE " \t\r\n"
/** character of double quotation marks */
#define QUOTE	"\""

/** strdelim() -           
 * Processes a single option line as used in the configuration files. This
 * only sets those values that have not already been set.
 * @param s
 * @ return next token in configuration line
 */
char *strdelim(char **s)
{
	char *old;
	int wspace = 0;

	if (*s == NULL)
		return NULL;

	old = *s;

	*s = strpbrk(*s, WHITESPACE QUOTE "=");
	if (*s == NULL)
		return (old);

	if (*s[0] == '\"') {
		memmove(*s, *s + 1, strlen(*s));	/* move nul too */
		/* Find matching quote */
		if ((*s = strpbrk(*s, QUOTE)) == NULL) {
			return (NULL);	/* no matching quote */
		} else {
			*s[0] = '\0';
			return (old);
		}
	}

	/* Allow only one '=' to be skipped */
	if (*s[0] == '=')
		wspace = 1;
	*s[0] = '\0';

	/* Skip any extra whitespace after first token */
	*s += strspn(*s + 1, WHITESPACE) + 1;
	if (*s[0] == '=' && !wspace)
		*s += strspn(*s + 1, WHITESPACE) + 1;

	return (old);
}

/** process_config_line()- process each line of configure file
 * @param confdata The struct of ConfData
 * @param line The line in file
 * @param filename the configure file
 * @param linenum the number of current processed line
 * @param activep The pointer of total number of error 
 * @return  zero if success, or non-zero 
 */

int process_config_line(ConfData * confdata, char *line,
			const char *filename, int linenum, int *activep)
{
	char *s, **charptr, *endofnumber, *keyword, *arg;
	int item, *intptr, value, value2;
	size_t len;

	/* Strip trailing whitespace */
	for (len = strlen(line) - 1; len > 0; len--) {
		if (strchr(WHITESPACE, line[len]) == NULL)
			break;
		line[len] = '\0';
	}

	s = line;
	/* Get the keyword. (Each line is supposed to begin with a keyword). */
	if ((keyword = strdelim(&s)) == NULL)
		return 0;
	/* Ignore leading whitespace. */
	if (*keyword == '\0')
		keyword = strdelim(&s);
	if (keyword == NULL || !*keyword || *keyword == '\n'
	    || *keyword == '#')
		return 0;

	item = parse_token(keyword, filename, linenum);


	switch (item) {
	case iUnknownItem:
		/* don't panic, but count bad confdata */
		return -1;

	case iRemoteUnsecureLogin:
		intptr = &confdata->mode;
		value2 = M_REMOTEUNSECURELOGIN;
	      parse_yesno:
		arg = strdelim(&s);
		if (!arg || *arg == '\0')
			fatal("%.200s line %d: Missing yes/no argument.",
			      filename, linenum);
		value = 0;	/* To avoid compiler warning... */
		if (strcmp(arg, "yes") == 0 || strcmp(arg, "true") == 0)
			value = value2;
		else if (strcmp(arg, "no") == 0
			 || strcmp(arg, "false") == 0)
			value = 0;
		else
			fatal("%.200s line %d: Bad yes/no argument.",
			      filename, linenum);
		if (*activep)
			*intptr |= value;

		break;
	case iUseShareBuffer:
		intptr = &confdata->mode;
		value2 = M_USESHAREBUFFER;
		goto parse_yesno;
	case iUseDebug:
		intptr = &confdata->use_debug;
		value2 = 1;
		goto parse_yesno;
	case iUseCgroups:
		intptr = &confdata->use_cgroups;
		value2 = 1;
		goto parse_yesno;
        case iAutoMount:
                intptr = &confdata->use_automount;
                value2 = 1;
                goto parse_yesno;
        case iFixedIDAlloc:
                intptr = &confdata->use_fixedidalloc;
                value2 = 1;
                goto parse_yesno;
	case iUseSyslog:
		intptr = &confdata->use_syslog;
		value2 = 1;
		goto parse_yesno;
	case iMyProxy:
		intptr = &confdata->mode;
		value2 = M_MYPROXY;
		goto parse_yesno;
	case iProxyServer:
		charptr = &confdata->myproxy_server;
	      parse_string:
		arg = strdelim(&s);
		if (!arg || *arg == '\0')
			fatal("%.200s line %d: Missing argument.",
			      filename, linenum);
		if (*activep && *charptr == NULL)
			*charptr = strdup(arg);
		break;
	case iProxyServerPort:
		intptr = &confdata->myproxy_port;
	      parse_int:
		arg = strdelim(&s);
		if (!arg || *arg == '\0')
			fatal("%.200s line %d: Missing argument.",
			      filename, linenum);
		if (arg[0] < '0' || arg[0] > '9')
			fatal("%.200s line %d: Bad number.", filename,
			      linenum);

		/* Octal, decimal, or hex format? */
		value = strtol(arg, &endofnumber, 0);
		if (arg == endofnumber)
			fatal("%.200s line %d: Bad number.", filename,
			      linenum);
		if (*activep && *intptr == -1)
			*intptr = value;
		break;
	case iLocalSecureLogin:
		intptr = &confdata->mode;
		value2 = M_LOGINSECURELOGIN;
		goto parse_yesno;
	case iUsePass:
		intptr = &confdata->usepass;
		value2 = USE_PASSPHRASE;
	      parse_yesno2:
		arg = strdelim(&s);
		if (!arg || *arg == '\0')
			fatal("%.200s line %d: Missing yes/no argument.",
			      filename, linenum);
		value = 0;	/* To avoid compiler warning... */
		if (strcmp(arg, "yes") == 0 || strcmp(arg, "true") == 0)
			value = value2;
		else if (strcmp(arg, "no") == 0
			 || strcmp(arg, "false") == 0)
			value = 0;
		else
			fatal("%.200s line %d: Bad yes/no argument.",
			      filename, linenum);
		if (*activep && value != 0)
			*intptr = value;
		break;
	case iUseCertfile:
		intptr = &confdata->usepass;
		value2 = USE_LOCALCERT;
		goto parse_yesno2;
	case iUseExternalAMS:
		intptr = &confdata->mode;
		value2 = M_USEEXTERNALAMS;
		goto parse_yesno;
        case iAMSServerIsLocal:
                intptr= &confdata->ams_server_is_local;
                value2=1;
                goto parse_yesno;
	case iAMSServer:
		charptr = &confdata->ams_server;
		goto parse_string;
	case iAMSPort:
		intptr = &confdata->ams_port;
		goto parse_int;
	case iUseExternalAES:
		intptr = &confdata->mode;
		value2 = M_USEEXTERNALAES;
		goto parse_yesno;
	case iAESServer:
		charptr = &confdata->aes_server;
		goto parse_string;
	case iAESPort:
		intptr = &confdata->aes_port;
		goto parse_int;
	case iVOACConf:
		charptr = &confdata->vo_acconf;
		goto parse_string;
	case iCSubsys:
		charptr = &confdata->cgroup_subsys;
		goto parse_string;
	case iVOCADir:
		charptr = &confdata->vo_cacertdir;
		goto parse_string;
	case iVOCACertF:
                charptr = &confdata->vo_cacertf;
                goto parse_string;
	case iNodePrvKeyF:
		charptr = &confdata->node_prvkeyf;
		goto parse_string;
	default:
		fatal("process_config_line: Unimplemented opcode %d",
		      item);
	}

	/* Check that there is no garbage at end of line. */
	if ((arg = strdelim(&s)) != NULL && *arg != '\0') {
		fatal
		    ("%.200s line %d: garbage at end of line; \"%.200s\".",
		     filename, linenum, arg);
	}
	return 0;
}


/** Reads the config file and modifies the confdata accordingly.  ConfData
 * should already be initialized before this call.
 * @param filename configure file.
 * @param confdata ConfData to be stored in configure parameters.
 * @param checkperm whether check the permission of configure file.
 *                 check if 1,or not.
 * @return This never returns if there is an error.
 *              If the file does not exist, this returns 0.
 */
int xos_read_config_file(const char *filename, ConfData * confdata,
		     int checkperm)
{
	FILE *f;
	char line[1024];
	int active, linenum;
	int bad_options = 0;

	/* Open the file. */
	if ((f = fopen(filename, "r")) == NULL)
		return 0;

	if (checkperm) {
		struct stat sb;

		if (fstat(fileno(f), &sb) == -1)
			fatal("fstat %s: %s", filename, strerror(errno));
		if (((sb.st_uid != 0 && sb.st_uid != getuid()) ||
		     (sb.st_mode & 022) != 0))
			fatal("Bad owner or permissions on %s", filename);
	}


	/*
	 * Mark that we are now processing the confdata.  This flag is turned
	 * on/off by Host specifications.
	 */
	active = 1;
	linenum = 0;
	while (fgets(line, sizeof(line), f)) {
		/* Update line number counter. */
		linenum++;
		if (process_config_line
		    (confdata, line, filename, linenum, &active) != 0)
			bad_options++;
	}
	fclose(f);
	if (bad_options > 0)
		fatal("%s: terminating, %d bad configuration confdata\n",
		      filename, bad_options);
	return 1;
}


/** initialize the ConfData struct
 * @param confdata it should be declared in other codes.
 */
void initialize_confdata(ConfData * confdata)
{
	memset(confdata, 'X', sizeof(*confdata));

	confdata->mode = 0;
	confdata->use_debug = 0;
	confdata->use_syslog = 0;
	confdata->use_cgroups = 0;
	confdata->use_automount = 1;
	confdata->use_fixedidalloc = 0;
	confdata->myproxy_server = NULL;
	confdata->myproxy_port = -1;
	confdata->usepass = -1;
        confdata->ams_server_is_local= 0;
	confdata->ams_server = NULL;
	confdata->ams_port = -1;
	confdata->aes_server = NULL;
	confdata->aes_port = -1;
	confdata->vo_acconf = NULL;
	confdata->cgroup_subsys = NULL;
	confdata->vo_cacertdir = NULL;
	confdata->vo_cacertf = NULL;
	confdata->node_prvkeyf = NULL;
}


/** Fill the default value in ConfData struct
 * @param confdata it should be declared in other codes.
 */
void fill_default_confdata(ConfData * confdata)
{
	if (confdata->mode == 0)
		confdata->mode = M_LOGINSECURELOGIN;
	if (confdata->usepass == -1)
		confdata->usepass = USE_LOCALCERT;
	if (confdata->vo_acconf == NULL)
		confdata->vo_acconf = VO_ACCONF;
        if (confdata->cgroup_subsys == NULL)
		confdata->cgroup_subsys = CGROUP_SUBSYS_DEFAULT;
	if (confdata->vo_cacertdir == NULL)
		confdata->vo_cacertdir = VO_CAPUBKEYF;
	if (confdata->node_prvkeyf == NULL)
		confdata->node_prvkeyf = NODE_PRVKEYF;
        if (confdata->vo_cacertf == NULL)
                confdata->vo_cacertf = VO_CACERTF;

}


/** Print ConfData struct parsed from configure file
 * @param confdata pointer of ConfData to be printed.
 * @param flag control whether prints debug information (out of date)
 */
void print_confdata(ConfData * confdata, int flag)
{
	if (!confdata)
		return;

	if (confdata->mode == 0 && flag) {
		printf("Configure not filled\n");
		return;
	}


	printf("=============== PAM configure data ===============\n");

	if (confdata->mode & M_REMOTEUNSECURELOGIN)
		printf("work mode:\t\tRemote unsecure login\n");
	if (confdata->mode & M_USESHAREBUFFER)
		printf("work mode:\t\t\tUse Share Buffer\n");
	if (confdata->mode & M_MYPROXY) {
		printf("work mode:\t\t\tMyProxy\n");
		if (confdata->myproxy_server)
			printf("-- Myproxy server:\t\t%s\n",
			       confdata->myproxy_server);
		else
			printf("-- Myproxy server error\n");
		if (confdata->myproxy_port != -1)
			printf("-- Myproxy port:\t\t%d\n",
			       confdata->myproxy_port);
		else
			printf("-- Myproxy port error\n");
	}
	if (confdata->mode & M_LOGINSECURELOGIN) {
		printf("work mode:\t\t\tlocal secure login\n");
		if (confdata->usepass == USE_PASSPHRASE)
			printf("-- Use pass\n");
		else if (confdata->usepass == USE_LOCALCERT)
			printf("-- Use user certificate\n");
		else
			printf("-- pass option error[%d]\n",
			       confdata->usepass);
	}
	if (confdata->mode & M_USEEXTERNALAMS) {
		printf("work mode:\t\t\tUse external AMS\n");
		if (confdata->ams_server)
			printf("-- External AMS server:\t\t%s\n",
			       confdata->ams_server);
		else
			printf("-- External AMS server error\n");
		if (confdata->ams_port != -1)
			printf("-- External AMS port:\t\t%d\n",
			       confdata->ams_port);
		else
			printf("-- External AMS port error\n");
	}
	if (confdata->mode & M_USEEXTERNALAES) {
		printf("work mode:\t\t\tUse external AES\n");
		if (confdata->aes_server)
			printf("-- External AES server:\t\t%s\n",
			       confdata->aes_server);
		else
			printf("-- External AES server error\n");
		if (confdata->aes_port != -1)
			printf("-- External AES port:\t\t%d\n",
			       confdata->aes_port);
		else
			printf("-- External AES port error\n");
	}

	if (confdata->vo_acconf != NULL)
		printf("VO Access Configure file: %s\n",
		       confdata->vo_acconf);

	if (confdata->vo_cacertdir != NULL)
		printf("VO CA dir: %s\n", confdata->vo_cacertdir);

        if (confdata->vo_cacertf != NULL)
               printf("VO CA cert name:\t\t%s\n", confdata->vo_cacertf);
	
	if (confdata->node_prvkeyf != NULL)
		printf("Node private key file: %s\n",
		       confdata->node_prvkeyf);

	printf("Use Debug:\t\t\t");
	if (confdata->use_debug)
		printf("YES\n");
	else
		printf("NO\n");

	printf("Use Syslog:\t\t\t");
	if (confdata->use_syslog)
		printf("YES\n");
	else
		printf("NO\n");

        printf("Open Automount:\t\t\t");
        if (confdata->use_automount)
                printf("YES\n");
        else
                printf("NO\n");


        printf("cgroup support:\t\t\t");
        if (confdata->use_cgroups)
                printf("YES\n");
        else
                printf("NO\n");

        if (confdata->cgroup_subsys != NULL)
                printf("  Subsys: %s\n", confdata->cgroup_subsys);

        printf("UseFixedID:\t\t\t");
        if (confdata->use_fixedidalloc)
                printf("YES\n");
        else
                printf("NO\n");

	printf("==================================================\n");
}
