/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file bdb_common.c
 * @brief The implementation of common database interfaces
 *
 */
#include "bdb_common.h"



/** Fetch the DN as the key of secondary database.
 * Used to extract an dababase item's name from an
 * dababase database record. This function is used to create
 * keys for secondary database records. 
 * @param dbp The pointer of database
 * @param pkey The \a pkey pointed to key item of primary database
 * @param pdata The \a pdata pointed to data item
 * @param skey The \a skey pointed to key item of secondary database
 * @return zero if success, or non-zero
 */
int
get_item_name(DB * dbp, const DBT * pkey, const DBT * pdata, DBT * skey)
{
	u_int offset;

	dbp = NULL;
	pkey = NULL;

	offset = sizeof(int) + sizeof(int) + sizeof(int) + sizeof(long);

	/* Check to make sure there's data */
	if (pdata->size < offset)
		return (-1);

	/* Now set the secondary key's data to be the item name */
	memset(skey, 0, sizeof(DBT));
	skey->data = (u_int8_t *) pdata->data + offset;
	skey->size = (u_int32_t) strlen(skey->data) + 1;

	return (0);
}

/** Fetch the uid as the key of secondary database. 
 * Used to extract an dababase item's name from an                          
 * dababase database record. This function is used to create                 
 * keys for secondary database records.                                      
 * @param dbp The pointer of database                                        
 * @param pkey The \a pkey pointed to key item of primary database           
 * @param pdata The \a pdata pointed to data item                            
 * @param skey The \a skey pointed to key item of secondary database         
 * @return zero if success, or non-zero                                      
 */

int get_item_uid(DB * dbp, const DBT * pkey, const DBT * pdata, DBT * skey)
{
	u_int offset;

	dbp = NULL;
	pkey = NULL;

	offset = 0;

	/* Check to make sure there's data */
	if (pdata->size < offset)
		return (-1);

	/* Now set the secondary key's data to be the item name */
	memset(skey, 0, sizeof(DBT));
	skey->data = (int *) pdata->data + offset;
	skey->size = sizeof(int);

	return (0);
}


/** Fetch the subgroup name as the key of secondary database 
 * Used to extract an dababase item's name from an
 * dababase database record. This function is used to create
 * keys for secondary database records.
 * @param dbp The pointer of database
 * @param pkey The \a pkey pointed to key item of primary database
 * @param pdata The \a pdata pointed to data item
 * @param skey The \a skey pointed to key item of secondary database
 * @return zero if success, or non-zero
 */

int
get_item_group(DB * dbp, const DBT * pkey, const DBT * pdata, DBT * skey)
{
	u_int offset;

	dbp = NULL;
	pkey = NULL;

	offset = sizeof(int) + sizeof(long);

	/* Check to make sure there's data */
	if (pdata->size < offset)
		return (-1);

	/* Now set the secondary key's data to be the item name */
	memset(skey, 0, sizeof(DBT));
	skey->data = (u_int8_t *) pdata->data + offset;
	skey->size = (u_int32_t) strlen(skey->data) + 1;

	return (0);
}

/** Fetch the gid as the key of secondary database 
 * Used to extract an dababase item's name from an
 * dababase database record. This function is used to create
 * keys for secondary database records.
 * @param dbp The pointer of database
 * @param pkey The \a pkey pointed to key item of primary database
 * @param pdata The \a pdata pointed to data item
 * @param skey The \a skey pointed to key item of secondary database
 * @return zero if success, or non-zero
 */
int get_item_gid(DB * dbp, const DBT * pkey, const DBT * pdata, DBT * skey)
{
	u_int offset;

	dbp = NULL;
	pkey = NULL;


	offset = 0;

	/* Check to make sure there's data */
	if (pdata->size < offset)
		return (-1);

	/* Now set the secondary key's data to be the item name */
	memset(skey, 0, sizeof(DBT));
	skey->data = (int *) pdata->data + offset;
	skey->size = sizeof(int);

	return (0);
}

/** Open a database
 * @param dbpp the pointer pointed the the pointer of database
 * @param file_name the database file for creatation
 * @param program_name the program name, using to log error
 * @param error_file_pointer the error file
 * @param is_secondary the \a is_secondary shows whether the database is
 *        index database.
 * @return 0 if success, or non-zeor
 */
int
open_database(DB ** dbpp, const char *file_name,
	      const char *program_name, FILE * error_file_pointer,
	      int is_secondary)
{
	DB *dbp;		/* For convenience */
	u_int32_t open_flags;
	int ret;

	/* Initialize the DB handle */
	ret = db_create(&dbp, NULL, 0);
	if (ret != 0) {
		fprintf(error_file_pointer, "%s: %s\n", program_name,
			db_strerror(ret));
		return (ret);
	}
	/* Point to the memory malloc'd by db_create() */
	*dbpp = dbp;

	/* Set up error handling for this database */
	dbp->set_errfile(dbp, error_file_pointer);
	dbp->set_errpfx(dbp, program_name);

	/*
	 * If this is a secondary database, then we want to allow
	 * sorted duplicates.
	 */
	if (is_secondary) {
		ret = dbp->set_flags(dbp, DB_DUPSORT);
		if (ret != 0) {
			dbp->err(dbp, ret,
				 "Attempt to set DUPSORT flags failed.",
				 file_name);
			return (ret);
		}
	}

	/* Set the open flags */
	if (getuid() == 0)
		open_flags = DB_CREATE;	/* Allow database creation */
	else
		open_flags = DB_RDONLY;

	/* Now open the database */
	ret = dbp->open(dbp,	/* Pointer to the database */
			NULL,	/* Txn pointer */
			file_name,	/* File name */
			NULL,	/* Logical db name */
			DB_BTREE,	/* Database type (using btree) */
			open_flags,	/* Open flags */
			0644);	/** \todo File mode should be 0600 for the AMS */
	if (ret != 0) {
		dbp->err(dbp, ret, "Database '%s' open failed.",
			 file_name);
		return (ret);
	}

	return (0);
}

/** opens all databases
 * @param my_stock the set of databases
 * @param program_name the program name for loging error
 * @param err_file_pointer the pointer of error file
 * @return 0 if success, or non-zero 
 */
int
databases_setup(STOCK_DBS * my_stock, const char *program_name,
		FILE * error_file_pointer)
{
	int ret;


	/* Open the PASSWD database */
	ret = open_database(&(my_stock->gpasswd_dbp),
			    my_stock->gpasswd_db_name,
			    program_name, error_file_pointer, PRIMARY_DB);
	if (ret != 0)
		return (ret);

	/* Open the index  - dn index */
	ret = open_database(&(my_stock->igpasswd_sdbp),
			    my_stock->ipasswd_db_name,
			    program_name, error_file_pointer,
			    SECONDARY_DB);
	if (ret != 0)
		return (0);

	/*
	 * Associate the itemname db with its primary db
	 * (PASSWD db).
	 */
	my_stock->gpasswd_dbp->associate(my_stock->gpasswd_dbp,	/* Primary db */
					 NULL,	/* txn id */
					 my_stock->igpasswd_sdbp,	/* Secondary db */
					 get_item_name,	/* Secondary key creator */
					 0);	/* Flags */

	/* Open the index database - uid index */
	ret = open_database(&(my_stock->ugpasswd_sdbp),
			    my_stock->upasswd_db_name,
			    program_name, error_file_pointer,
			    SECONDARY_DB);
	if (ret != 0)
		return (0);

	/*
	 * Associate the itemname db with its primary db
	 * (PASSWD db).
	 */
	my_stock->gpasswd_dbp->associate(my_stock->gpasswd_dbp,	/* Primary db */
					 NULL,	/* txn id */
					 my_stock->ugpasswd_sdbp,	/* Secondary db */
					 get_item_uid,	/* Secondary key creator */
					 0);


	/* Open the GROUP database */
	ret = open_database(&(my_stock->ggroup_dbp),
			    my_stock->ggroup_db_name,
			    program_name, error_file_pointer, PRIMARY_DB);
	if (ret != 0)
		return (ret);
	/* Open the index - vo index */
	ret = open_database(&(my_stock->iggroup_sdbp),
			    my_stock->igroup_db_name,
			    program_name, error_file_pointer,
			    SECONDARY_DB);
	if (ret != 0)
		return (0);

	/*
	 * Associate the itemname db with its primary db
	 * (GROUPS db).
	 */
	my_stock->ggroup_dbp->associate(my_stock->ggroup_dbp,	/* Primary db */
					NULL,	/* txn id */
					my_stock->iggroup_sdbp,	/* Secondary db */
					get_item_group,	/* Secondary key creator */
					0);	/* Flags */

	/* Open the index - gid index */
	ret = open_database(&(my_stock->uggroup_sdbp),
			    my_stock->ugroup_db_name,
			    program_name, error_file_pointer,
			    SECONDARY_DB);
	if (ret != 0)
		return (0);

	/*
	 * Associate the itemname db with its primary db
	 * (GROUPS db).
	 */
	my_stock->ggroup_dbp->associate(my_stock->ggroup_dbp,	/* Primary db */
					NULL,	/* txn id */
					my_stock->uggroup_sdbp,	/* Secondary db */
					get_item_gid,	/* Secondary key creator */
					0);



	return (0);
}


/** Initializes the STOCK_DBS struct.
 * @param my_stock the set of databases
 * 
 */
void initialize_stockdbs(STOCK_DBS * my_stock)
{
	my_stock->db_home_dir = DEFAULT_HOMEDIR;
	my_stock->gpasswd_dbp = NULL;
	my_stock->igpasswd_sdbp = NULL;
	my_stock->ugpasswd_sdbp = NULL;

	my_stock->ggroup_dbp = NULL;
	my_stock->iggroup_sdbp = NULL;
	my_stock->uggroup_sdbp = NULL;

	my_stock->gpasswd_db_name = NULL;
	my_stock->ipasswd_db_name = NULL;
	my_stock->upasswd_db_name = NULL;

	my_stock->ggroup_db_name = NULL;
	my_stock->igroup_db_name = NULL;
	my_stock->ugroup_db_name = NULL;
}

/** Identify all the files that will hold our databases. 
 * @param my_stock the set of databases
 */
void set_db_filenames(STOCK_DBS * my_stock)
{
	size_t size;

	/* Create the PASSWD DB file name */
	size = strlen(my_stock->db_home_dir) + strlen(GPASSWD_DB) + 1;
	my_stock->gpasswd_db_name = malloc(size);
	snprintf(my_stock->gpasswd_db_name, size, "%s%s",
		 my_stock->db_home_dir, GPASSWD_DB);


	size = strlen(my_stock->db_home_dir) + strlen(IGPASSWD_DB) + 1;
	my_stock->ipasswd_db_name = malloc(size);
	snprintf(my_stock->ipasswd_db_name, size, "%s%s",
		 my_stock->db_home_dir, IGPASSWD_DB);

	size = strlen(my_stock->db_home_dir) + strlen(UGPASSWD_DB) + 1;
	my_stock->upasswd_db_name = malloc(size);
	snprintf(my_stock->upasswd_db_name, size, "%s%s",
		 my_stock->db_home_dir, UGPASSWD_DB);


	/* Create the GROUP DB file name */
	size = strlen(my_stock->db_home_dir) + strlen(GGROUP_DB) + 1;
	my_stock->ggroup_db_name = malloc(size);
	snprintf(my_stock->ggroup_db_name, size, "%s%s",
		 my_stock->db_home_dir, GGROUP_DB);



	size = strlen(my_stock->db_home_dir) + strlen(IGGROUP_DB) + 1;
	my_stock->igroup_db_name = malloc(size);
	snprintf(my_stock->igroup_db_name, size, "%s%s",
		 my_stock->db_home_dir, IGGROUP_DB);

	size = strlen(my_stock->db_home_dir) + strlen(UGGROUP_DB) + 1;
	my_stock->ugroup_db_name = malloc(size);
	snprintf(my_stock->ugroup_db_name, size, "%s%s",
		 my_stock->db_home_dir, UGGROUP_DB);

}

/** Closes all the databases and secondary databases.
 * @param my_stock the set of databases
 * @return 0 if success, or non-zero 
 */
int databases_close(STOCK_DBS * my_stock)
{
	int ret;
	/*
	 * Note that closing a database automatically flushes its cached data
	 * to disk, so no sync is required here.
	 */
	if (my_stock->igpasswd_sdbp != NULL) {
		ret =
		    my_stock->igpasswd_sdbp->close(my_stock->igpasswd_sdbp,
						   0);
		if (ret != 0)
			fprintf(stderr,
				"Itemname database close failed: %s\n",
				db_strerror(ret));
	}
	if (my_stock->ugpasswd_sdbp != NULL) {
		ret =
		    my_stock->ugpasswd_sdbp->close(my_stock->ugpasswd_sdbp,
						   0);
		if (ret != 0)
			fprintf(stderr,
				"Itemname database close failed: %s\n",
				db_strerror(ret));
	}
	if (my_stock->gpasswd_dbp != NULL) {
		ret =
		    my_stock->gpasswd_dbp->close(my_stock->gpasswd_dbp, 0);
		if (ret != 0)
			fprintf(stderr,
				"Inventory database close failed: %s\n",
				db_strerror(ret));
	}

	if (my_stock->iggroup_sdbp != NULL) {
		ret =
		    my_stock->iggroup_sdbp->close(my_stock->iggroup_sdbp,
						  0);
		if (ret != 0)
			fprintf(stderr,
				"Itemname database close failed: %s\n",
				db_strerror(ret));
	}
	if (my_stock->uggroup_sdbp != NULL) {
		ret =
		    my_stock->uggroup_sdbp->close(my_stock->uggroup_sdbp,
						  0);
		if (ret != 0)
			fprintf(stderr,
				"Itemname database close failed: %s\n",
				db_strerror(ret));
	}
	if (my_stock->ggroup_dbp != NULL) {
		ret = my_stock->ggroup_dbp->close(my_stock->ggroup_dbp, 0);
		if (ret != 0)
			fprintf(stderr,
				"Inventory database close failed: %s\n",
				db_strerror(ret));
	}

	return (0);
}
