/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file xos_dbaux_grp.c
 * @brief The implementation of all interfaces in GGROUPS database
 * 
 * Adopted the Berkeley-DB as the basic Database, storing the Identity token of user.
 *
 *
 */

#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <db.h>



#include "xos_dbaux.h"
#include "xos_utils.h"
#include "bdb_common.h"




/** create a new GGROUPS entiry
 * @return pointer of GGROUPS struct if success, or NULL
 */

GGROUPS *ggrp_new()
{
	GGROUPS *ggrp = NULL;
	ggrp = malloc(sizeof(*ggrp));

	if (ggrp == NULL) {
		goto err;
	}
	memset(ggrp, 0, sizeof(GGROUPS));
      err:
	return ggrp;
}



/** get GGROUPS struct from database
 * @param vBuf Input buffer from database
 * @param gpwd returned GPASSWD struct after parsing the string buffer.
 * @return key of record in PASSWD database if success, or NULL
 */
char *get_db_item_grp(void *vBuf, GGROUPS * ggrp)
{

	int gid;
	long time;
	size_t buf_pos;
	char *VO, *vogroup, *lname, *sku, *members;
	char *buf = (char *) vBuf;

	gid = *((int *) buf);
	buf_pos = sizeof(int);


	time = *((long *) (buf + buf_pos));
	buf_pos += sizeof(long);

	vogroup = buf + buf_pos;
	buf_pos += strlen(vogroup) + 1;

	VO = buf + buf_pos;
	buf_pos += strlen(VO) + 1;

	sku = buf + buf_pos;
	buf_pos += strlen(sku) + 1;

	lname = buf + buf_pos;
	buf_pos += strlen(lname) + 1;


	members = buf + buf_pos;


	if (!ggrp)
		goto err;

	strncpy(ggrp->g_grptoken.g_vo, VO, MAXFIELD);
	strncpy(ggrp->g_grptoken.g_subgroup, vogroup, MAXFIELD);
	strncpy(ggrp->l_grptoken.g_mappedname, lname, MAXFIELD);
	ggrp->l_grptoken.g_mappedgid = gid;
	strncpy(ggrp->l_grptoken.g_members, members, MAXFIELD);
	ggrp->g_time = time;

      err:
	return sku;
}


/** Fetch GGROUPS struct by given local gid
 * @param gid user's local gid
 * @param ggrp GGROUPS structure holding group mapping information
 * @return key of record in PASSWD database if success, or NULL
 */
char *bdb_getdbgid_grp(gid_t gid, GGROUPS * ggrp)
{
	STOCK_DBS my_stock;
	DBC *itemgid_cursorp;
	DBT key, data;

	int ret, loop, flag = 0;
	char *skey = NULL;

	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}


	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));


	my_stock.uggroup_sdbp->cursor(my_stock.uggroup_sdbp, NULL,
				      &itemgid_cursorp, 0);

	key.data = &gid;
	key.size = sizeof(int);


	ret = itemgid_cursorp->c_get(itemgid_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {
			skey = get_db_item_grp(data.data, ggrp);
			if (skey) {
				DEBUG_ASSERT(ggrp->l_grptoken.
					     g_mappedgid != 0);
				if (gid == ggrp->l_grptoken.g_mappedgid) {
					flag = 1;
					skey = strdup(skey);
					break;
				}
			}

			loop =
			    itemgid_cursorp->c_get(itemgid_cursorp, &key,
						   &data, DB_NEXT_DUP);
		} while (loop == 0);

		if (!loop && !flag)
			skey = NULL;
	} else {
		if (DEBUG_INFO) DEBUG("No records found for '%i'", gid);
	}

	/* Close the cursor */
	itemgid_cursorp->c_close(itemgid_cursorp);
	databases_close(&my_stock);


	return skey;

}



/** Fetch GGROUPS struct by given local group name
 * @param gname local group name
 * @param ggrp GGROUPS structure holding group mapping information
 * @return key of record in PASSWD database if success, or NULL
 */

char *bdb_getdbgname_grp(char *gname, GGROUPS * ggrp)
{
	STOCK_DBS my_stock;
	DBC *cursor;
	DBT key, data;

	int ret, found = 0;
	char *skey = NULL;

	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}
	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	my_stock.ggroup_dbp->cursor(my_stock.ggroup_dbp, NULL, &cursor, 0);

	/* key.data = gname;
	   key.size = (u_int32_t)strlen(gname) + 1; */

	while (!found && cursor->c_get(cursor, &key, &data, DB_NEXT) == 0) {
		skey = get_db_item_grp(data.data, ggrp);
		if (skey) {
			DEBUG_ASSERT(ggrp != NULL);
			DEBUG_ASSERT(ggrp->l_grptoken.g_mappedname[0] !=
				     '\0');
			if (!strcmp(gname, ggrp->l_grptoken.g_mappedname)) {
				found = 1;
				skey = strdup(skey);
			}
		}
	}

	if (!found)
		skey = NULL;

	/* Close the cursor */
	cursor->c_close(cursor);
	databases_close(&my_stock);

	return skey;
}


/** get gid from database by his subgroup name
 * @param gname the subgroup name
 * @param the pointer of GROUPS struct for storing data
 * @return key of data if success, or NULL  
 */
char *bdb_getdbname_grp(char *gname, GGROUPS * ggrp)
{
	STOCK_DBS my_stock;
	DBC *itemname_cursorp;
	DBT key, data;

	int ret, loop, flag = 0;
	char *skey = NULL;

	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}
	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	/* Get a cursor to the itemname db */
	my_stock.iggroup_sdbp->cursor(my_stock.iggroup_sdbp, NULL,
				      &itemname_cursorp, 0);

	key.data = gname;
	key.size = (u_int32_t) strlen(gname) + 1;


	ret =
	    itemname_cursorp->c_get(itemname_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {
			skey = get_db_item_grp(data.data, ggrp);
			if (skey) {
				DEBUG_ASSERT(ggrp->l_grptoken.
					     g_mappedname[0] != '\0');
				if (!strncmp
				    (gname, ggrp->l_grptoken.g_mappedname,
				     strlen(gname))) {
					flag = 1;
					skey = strdup(skey);
					break;
				}
			}

			loop =
			    itemname_cursorp->c_get(itemname_cursorp, &key,
						    &data, DB_NEXT_DUP);
		} while (loop == 0);

		if (!loop && !flag)
			skey = NULL;
	} else {
		if (DEBUG_INFO) DEBUG("No records found for '%s'", gname);
	}

	/* Close the cursor */
	itemname_cursorp->c_close(itemname_cursorp);
	databases_close(&my_stock);


	return skey;
}



/** Fetch GGROUPS struct by given global identity token
 * @param gtk global identity token
 * @param gprp GGROUPS structure holding group mapping information
 * @return key of record in PASSWD database if success, or NULL
 */

char *bdb_getdbgtk_grp(GGRPTOKEN * gtk, GGROUPS * ggrp)
{
	STOCK_DBS my_stock;
	DBC *itemname_cursorp;
	DBT key, data;

	int ret, loop, flag = 0;
	char *skey = NULL;

	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}

	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	/* Get a cursor to the itemname db */
	my_stock.iggroup_sdbp->cursor(my_stock.iggroup_sdbp, NULL,
				      &itemname_cursorp, 0);

	key.data = gtk->g_subgroup;
	key.size = (u_int32_t) strlen(gtk->g_subgroup) + 1;

	ret =
	    itemname_cursorp->c_get(itemname_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {

			skey = get_db_item_grp(data.data, ggrp);
			if (ggrp) {
				if ((!strcmp
				     (ggrp->g_grptoken.g_vo, gtk->g_vo)
				     && strlen(ggrp->g_grptoken.g_vo) ==
				     strlen(gtk->g_vo))
				    &&
				    (!strcmp
				     (ggrp->g_grptoken.g_subgroup,
				      gtk->g_subgroup)
				     && strlen(ggrp->g_grptoken.
					       g_subgroup) ==
				     strlen(gtk->g_subgroup))
				    ) {
					skey = strdup(skey);
					flag = 1;
					break;
				}
			}

			loop =
			    itemname_cursorp->c_get(itemname_cursorp, &key,
						    &data, DB_NEXT_DUP);

		} while (loop == 0);

		if (!flag) {
			if (!loop)
				skey = NULL;
		}
	} else {
		if (DEBUG_INFO) DEBUG("No records found for the token");
	}

	/* Close the cursor */
	itemname_cursorp->c_close(itemname_cursorp);
	databases_close(&my_stock);

	return skey;
}



/** Fetch GGROUPS struct by given local identity token
 * @param ltk local identity token
 * @param ggrp GGROUPS structure holding group mapping information
 * @return key of record in PASSWD database if success, or NULL
 */

char *bdb_getdbltk_grp(LGRPTOKEN * ltk, GGROUPS * ggrp)
{
	STOCK_DBS my_stock;
	DBC *itemgid_cursorp;
	DBT key, data;

	int ret, loop;
	char *skey = NULL;

	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}


	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));


	my_stock.uggroup_sdbp->cursor(my_stock.uggroup_sdbp, NULL,
				      &itemgid_cursorp, 0);

	key.data = &(ltk->g_mappedgid);
	key.size = sizeof(int);


	ret = itemgid_cursorp->c_get(itemgid_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {
			skey = get_db_item_grp(data.data, ggrp);
			if (ggrp) {
				if ((ltk->g_mappedgid ==
				     ggrp->l_grptoken.g_mappedgid)) {
					skey = strdup(skey);
					break;
				}
			}

			loop =
			    itemgid_cursorp->c_get(itemgid_cursorp, &key,
						   &data, DB_NEXT_DUP);
		} while (loop == 0);

		if (!loop)
			skey = NULL;
	} else {
		if (DEBUG_INFO) DEBUG("No records found for the token");
	}

	/* Close the cursor */
	itemgid_cursorp->c_close(itemgid_cursorp);
	databases_close(&my_stock);


	return skey;
}


/** Store GGROUPS struct to GROUP database
 * @param gpren GGROUPS structure holding group mapping information
 * @return 0 if success, or !0
 */
int bdb_storedben_grp(GGROUPS * ggren)
{
	STOCK_DBS my_stock;
	DBT key, data;
	int ret;
	char databuf[MAXDATABUF];
	size_t bufLen, dataLen;

	char randkey[20 + 1];

	create_sranddbkey(randkey, 20);
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return (ret);
	}

	memset(databuf, 0, MAXDATABUF);
	bufLen = 0;
	dataLen = 0;

	dataLen = sizeof(int);
	memcpy(databuf, &(ggren->l_grptoken.g_mappedgid), dataLen);
	bufLen += dataLen;


	dataLen = sizeof(long);
	memcpy(databuf + bufLen, &(ggren->g_time), dataLen);
	bufLen += dataLen;

	bufLen =
	    pack_string(databuf, ggren->g_grptoken.g_subgroup, bufLen);
	bufLen = pack_string(databuf, ggren->g_grptoken.g_vo, bufLen);
	bufLen = pack_string(databuf, randkey, bufLen);
	bufLen =
	    pack_string(databuf, ggren->l_grptoken.g_mappedname, bufLen);
	bufLen = pack_string(databuf, ggren->l_grptoken.g_members, bufLen);


	/* Zero out the DBTs */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	/* The key is the item's SKU */
	key.data = randkey;
	key.size = (u_int32_t) strlen(randkey) + 1;

	/* The data is the information that we packed into databuf. */
	data.data = databuf;
	data.size = (u_int32_t) bufLen;

	/* Put the data into the database */
	my_stock.ggroup_dbp->put(my_stock.ggroup_dbp, 0, &key, &data, 0);

	databases_close(&my_stock);
	return ret;
}



/** Delete GGROUPS struct in GROUP database by given global identity token
 * @param gtk GGRPTOKEN structure holding global identity token
 * @return 0 if success, or !0
 */

int bdb_deldbgtk_grp(GGRPTOKEN * gtk)
{
	STOCK_DBS my_stock;
	DBT key, data;
	GGROUPS *ggrp = NULL;
	int ret;
	char *skey = NULL;

	ggrp = ggrp_new();
	skey = bdb_getdbgtk_grp(gtk, ggrp);
	if (!skey) {
		if (DEBUG_INFO) DEBUG("Not found the record");
		return -1;
	}

	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return (ret);
	}

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	key.data = skey;
	key.size = (u_int32_t) strlen(skey) + 1;

	my_stock.ggroup_dbp->del(my_stock.ggroup_dbp, 0, &key, 0);

	databases_close(&my_stock);
	return ret;
}




/** Print data in GGROUPS struct as readable format
 * @param gpwen GGROUPS structure holding group mapping information
 */

void bdb_printdata_grp(GGROUPS * ggren)
{
	printf
	    ("VO:\t%s\nVOgroup:\t%s\nL_name:\t%s\nL_gid:\t%d\nmember:\t%s\ntime:\t%ld\n",
	     ggren->g_grptoken.g_vo, ggren->g_grptoken.g_subgroup,
	     ggren->l_grptoken.g_mappedname, ggren->l_grptoken.g_mappedgid,
	     ggren->l_grptoken.g_members, ggren->g_time);
}


/** Print data in buffer as readable format
 * @param vBuf buffer holding user's mapping information
 * @return key of this record if success, or NULL
 */

char *print_db_item_grp(void *vBuf)
{

	int gid;
	long time;
	size_t buf_pos;
	char *VO, *vogroup, *lname, *sku, *member;
	char *buf = (char *) vBuf;

	gid = *((int *) buf);
	buf_pos = sizeof(int);


	time = *((long *) (buf + buf_pos));
	buf_pos += sizeof(long);

	vogroup = buf + buf_pos;
	buf_pos += strlen(vogroup) + 1;

	VO = buf + buf_pos;
	buf_pos += strlen(VO) + 1;

	sku = buf + buf_pos;
	buf_pos += strlen(sku) + 1;

	lname = buf + buf_pos;
	buf_pos += strlen(lname) + 1;

	member = buf + buf_pos;


	printf("VO: %s\n", VO);
	printf("VOgroup: %s\n", vogroup);
	printf("\tSKU: %s\n", sku);
	printf("\tname: %s\n", lname);
	printf("\tgid: %i\n", gid);
	printf("\tmember:%s\n", member);
	printf("\ttime: %ld\n", time);

	return (sku);
}


/** Print all the data in GPASSWD database
 */
void bdb_printdb_grp(void)
{
	STOCK_DBS my_stock;
	DBC *iggroup_cursorp;
	DBT key, data;
	char *gname;
	int exit_value, ret;

	init_DBT(&key, &data);
	/* Identify the files that hold our databases */
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	/* Open all databases */
	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret != 0) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return;
	}

	my_stock.ggroup_dbp->cursor(my_stock.ggroup_dbp, NULL,
				    &iggroup_cursorp, 0);

	exit_value = 0;
	while ((ret =
		iggroup_cursorp->c_get(iggroup_cursorp, &key, &data,
				       DB_NEXT)) == 0) {
		gname = print_db_item_grp(data.data);
		if (ret) {
			exit_value = ret;
			break;
		}
	}

	/* Close the cursor */
	iggroup_cursorp->c_close(iggroup_cursorp);

	/* close our databases */
	databases_close(&my_stock);


}




/** Dispose the given data in database by self-defined function
 * @param gtk global identity token to find the GGROUPS struct
 * @param optfunc self-defined function to deal with the data. It
 *        must be the type of OPTGRPFUNC.
 * @return 0 if success, or !0
 */

int bdb_operatedata_grp(GGRPTOKEN * gtk, OPTGRPFUNC optfunc)
{
#if 0
	GGROUPS *ggrp = NULL;
	char *skey = NULL;
	int ret = 0;

	ggrp = ggrp_new();
	skey = bdb_getdbgtk_grp(gtk, ggrp);
	if (!skey) {
		if (DEBUG_INFO) DEBUG("No found the record");
		ret = -1;
		goto err;
	}

	ret = optfunc(ggrp);
      err:
	if (ggrp)
		free(ggrp);
	return ret
#else
	return -1;
#endif
}



/** Dispose all the  data in database by self-defined function
 * @param optfunc self-defined function to deal with the data. It
 *        must be the type of OPTGRPFUNC.
 * @return 0 if success, or !0
 */
#if 0
int bdb_operatedb_grp(GGROUPS * t_grp, OPTGRPFUNC optfunc)
{
	STOCK_DBS my_stock;
	DBC *iggroup_cursorp;
	DBT key, data;
	char *skey;
	int exit_ret, ret;
	GGROUPS *ggrp = NULL;

	init_DBT(&key, &data);
	/* Identify the files that hold our databases */
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	/* Open all databases */
	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret != 0) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return -1;
	}

	my_stock.ggroup_dbp->cursor(my_stock.ggroup_dbp, NULL,
				    &iggroup_cursorp, 0);

	exit_ret = 0;
	ggrp = ggrp_new();
	while ((ret =
		iggroup_cursorp->c_get(iggroup_cursorp, &key, &data,
				       DB_NEXT)) == 0) {
		skey = get_db_item_grp(data.data, ggrp);
		if (skey) {
			ret = optfunc(ggrp, t_grp);
		}

		if (ret) {
			exit_ret = ret;
			break;
		}
	}

	/* Close the cursor */
	iggroup_cursorp->c_close(iggroup_cursorp);

	/* close our databases */
	databases_close(&my_stock);

	return exit_ret;
}
#else
int bdb_operatedb_grp(void *t_grp, void *r_grp, OPTGRPFUNC optfunc)
{
	STOCK_DBS my_stock;
	DBC *iggroup_cursorp;
	DBT key, data;
	char *skey;
	int exit_ret, ret;
	GGROUPS *ggrp = NULL;

	init_DBT(&key, &data);
	/* Identify the files that hold our databases */
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	/* Open all databases */
	ret = databases_setup(&my_stock, "xos_dbaux_grp", stderr);
	if (ret != 0) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return -1;
	}

	my_stock.ggroup_dbp->cursor(my_stock.ggroup_dbp, NULL,
				    &iggroup_cursorp, 0);

	exit_ret = 0;
	ggrp = ggrp_new();
	while ((ret =
		iggroup_cursorp->c_get(iggroup_cursorp, &key, &data,
				       DB_NEXT)) == 0) {
		skey = get_db_item_grp(data.data, ggrp);
		if (skey) {
			ret = optfunc(ggrp, t_grp, r_grp);
		}

		if (ret) {
			exit_ret = ret;
			break;
		}
	}

	/* Close the cursor */
	iggroup_cursorp->c_close(iggroup_cursorp);

	/* close our databases */
	databases_close(&my_stock);

	return exit_ret;
}
#endif
