/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file xos_dbaux_pwd.c
 * @brief The implementation of all interfaces in GPASSWD database
 * 
 * adopted the Berkeley-DB as the basic Database, storing the Identity token of user.
 *
 *
 */

#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <db.h>


#include "xos_dbaux.h"
#include "xos_utils.h"
#include "bdb_common.h"




/*
 * internal functions
 */

/** Print error as readable string.
 * @param ret error code of database
 * @param func function name
 */
void print_error(int ret, const char *func)
{
	if (ret != 0)
		DEBUG("%s: ERROR: %s", func, db_strerror(ret));
}



/** Initiate DBT structure of key and data
 * @param key
 * @param data
 */
void init_DBT(DBT * key, DBT * data)
{
	memset(key, 0, sizeof(DBT));
	memset(data, 0, sizeof(DBT));
}

/** Create random string key
 * @param randstr buffer to hold random string
 * @param len length of random string
 */
void create_sranddbkey(char *randstr, int len)
{
	int index;
	int i;
	static int init = 0;
	static unsigned int seed;

	if (!init) {
		seed = (unsigned int) time((time_t *) 0);
		init = 1;
	}

	for (i = 0; i < len - 1; i++) {
		index =
		    (int) (((float) 62) * rand_r(&seed) /
			   (RAND_MAX + 1.0));

		if (index < 10)
			randstr[i] = (48 + index);
		else if (index < 36)
			randstr[i] = (55 + index);
		else if (index < 62)
			randstr[i] = (61 + index);
	}
	randstr[len - 1] = '\0';
}

/** Pack string to buffer
 * @param buffer pointer of buffer
 * @param string  string to pack
 * @param start_pos position to start storing.
 * @return current length of buffer
 */
size_t pack_string(char *buffer, char *string, size_t start_pos)
{
	size_t string_size;

	string_size = strlen(string) + 1;
	memcpy(buffer + start_pos, string, string_size);

	return (start_pos + string_size);
}



/** create a new passwd entiry
 * @return pointer of GPASSWD struct if success, or NULL
 */

GPASSWD *gpwd_new()
{
	GPASSWD *gpwd = NULL;
	gpwd = malloc(sizeof(*gpwd));

	if (gpwd == NULL) {
		goto err;
	}
	memset(gpwd, 0, sizeof(GPASSWD));
      err:
	return gpwd;
}



/** get GPASSWD struct  from database
 * @param vBuf Input buffer from database
 * @param gpwd returned GPASSWD struct after parsing the string buffer.
 * @return key of record in PASSWD database if success, or NULL
 */
char *get_db_item_pwd(void *vBuf, GPASSWD * gpwd)
{

	int uid, gid, mount;
	long time;
	size_t buf_pos;
	char *shell, *DNname, *VO, *Role, *GUID, *lname, *sku, *home;
	char *buf = (char *) vBuf;

	uid = *((int *) buf);
	buf_pos = sizeof(int);

	gid = *((int *) (buf + buf_pos));
	buf_pos += sizeof(int);

        mount = *((int *) (buf + buf_pos));
	buf_pos += sizeof(int);

	time = *((long *) (buf + buf_pos));
	buf_pos += sizeof(long);

	DNname = buf + buf_pos;
	buf_pos += strlen(DNname) + 1;

	VO = buf + buf_pos;
	buf_pos += strlen(VO) + 1;

	Role = buf + buf_pos;
	buf_pos += strlen(Role) + 1;

	GUID = buf + buf_pos;
	buf_pos += strlen(GUID) + 1;
		
	lname = buf + buf_pos;
	buf_pos += strlen(lname) + 1;

	sku = buf + buf_pos;
	buf_pos += strlen(sku) + 1;

	shell = buf + buf_pos;
	buf_pos += strlen(shell) + 1;

	home = buf + buf_pos;


	if (!gpwd)
		return NULL;

	strncpy(gpwd->g_idtoken.g_dn, DNname, MAXFIELD);
	strncpy(gpwd->g_idtoken.g_vo, VO, MAXFIELD);
	strncpy(gpwd->g_idtoken.g_role, Role, MAXFIELD);
	strncpy(gpwd->g_idtoken.g_guid, GUID, MAXFIELD);
	strncpy(gpwd->l_idtoken.g_mappedname, lname, MAXFIELD);
	gpwd->l_idtoken.g_mappeduid = uid;
	gpwd->l_idtoken.g_mappedgid = gid;
	strncpy(gpwd->g_shell, shell, MAXFIELD);
	strncpy(gpwd->g_home, home, MAXFIELD);
	gpwd->g_time = time;
	gpwd->g_mount = mount;

	return sku;
}


/** Fetch GPASSWD struct by given user's local account name
 * @param uname user's local account name
 * @param gpwd GPASSWD structure holding user's mapping information
 * @return key of record in PASSWD database if success, or NULL
 */
char *bdb_getdbuname_pwd(char *uname, GPASSWD * gpwd)
{
	STOCK_DBS my_stock;
	DBC *cursor;
	DBT key, data;

	int ret, found = 0;
	char *skey = NULL;


	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}
	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	my_stock.gpasswd_dbp->cursor(my_stock.gpasswd_dbp, NULL, &cursor,
				     0);

	/* key.data = &uid;
	   key.size = sizeof(int); */

	while (!found && cursor->c_get(cursor, &key, &data, DB_NEXT) == 0) {
		skey = get_db_item_pwd(data.data, gpwd);
		if (skey) {
			DEBUG_ASSERT(gpwd != NULL);
			DEBUG_ASSERT(gpwd->l_idtoken.g_mappedname[0] !=
				     '\0');
			if (!strcmp(uname, gpwd->l_idtoken.g_mappedname)) {
				found = 1;
				skey = strdup(skey);
			}
		}
	}

	if (!found)
		skey = NULL;

	/* Close the cursor */
	cursor->c_close(cursor);
	databases_close(&my_stock);


	return skey;
}


/** Fetch GPASSWD struct by given user's local uid
 * @param uid user's local uid
 * @param gpwd GPASSWD structure holding user's mapping information
 * @return key of record in PASSWD database if success, or NULL
 */
char *bdb_getdbuid_pwd(uid_t uid, GPASSWD * gpwd)
{
	STOCK_DBS my_stock;
	DBC *itemuid_cursorp;
	DBT key, data;

	int ret, loop, flag = 0;
	char *skey = NULL;

	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}


	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));


	my_stock.ugpasswd_sdbp->cursor(my_stock.ugpasswd_sdbp, NULL,
				       &itemuid_cursorp, 0);

	key.data = &uid;
	key.size = sizeof(int);


	ret = itemuid_cursorp->c_get(itemuid_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {
			skey = get_db_item_pwd(data.data, gpwd);
			if (skey) {
				DEBUG_ASSERT(gpwd->l_idtoken.g_mappeduid !=
					     0);
				if (uid == gpwd->l_idtoken.g_mappeduid) {
					flag = 1;
					skey = strdup(skey);
					break;
				}
			}

			loop =
			    itemuid_cursorp->c_get(itemuid_cursorp, &key,
						   &data, DB_NEXT_DUP);
		} while (loop == 0);

		if (!loop && !flag)
			skey = NULL;
	} else {
		if (DEBUG_INFO) DEBUG("No records found for '%i'", uid);
	}

	/* Close the cursor */
	itemuid_cursorp->c_close(itemuid_cursorp);
	databases_close(&my_stock);

	return skey;

}





/** Fetch GPASSWD struct by given user's DN
 * @param DNname user's DN
 * @param gpwd GPASSWD structure holding user's mapping information
 * @return key of record in PASSWD database if success, or NULL
 */
char *bdb_getdbname_pwd(char *DNname, GPASSWD * gpwd)
{

	STOCK_DBS my_stock;
	DBC *itemname_cursorp;
	DBT key, data;

	int ret, loop, flag = 0;
	char *skey = NULL;
	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}

	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	/* Get a cursor to the itemname db */
	my_stock.igpasswd_sdbp->cursor(my_stock.igpasswd_sdbp, NULL,
				       &itemname_cursorp, 0);

	key.data = DNname;
	key.size = (u_int32_t) strlen(DNname) + 1;


	ret =
	    itemname_cursorp->c_get(itemname_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {

			skey = get_db_item_pwd(data.data, gpwd);
			if (skey) {
				if (!strncmp
				    (DNname, gpwd->g_idtoken.g_dn,
				     strlen(DNname))) {
					flag = 1;
					skey = strdup(skey);
					break;
				}
			}

			loop =
			    itemname_cursorp->c_get(itemname_cursorp, &key,
						    &data, DB_NEXT_DUP);

		} while (loop == 0);

		if (!flag) {
			if (!loop) {
				if (DEBUG_INFO) DEBUG("%s: loop is null",
					    __FUNCTION__);
				skey = NULL;
			} else
			    if (strncmp
				(DNname, gpwd->l_idtoken.g_mappedname,
				 strlen(DNname))) {
				if (DEBUG_INFO) DEBUG
				    ("%s: DNname is not same (%s,%s,[%d,%d])",
				     __FUNCTION__, DNname,
				     gpwd->l_idtoken.g_mappedname,
				     strlen(DNname),
				     strlen(gpwd->l_idtoken.g_mappedname));
				skey = NULL;
			}
		}

	} else {
		 if (DEBUG_INFO)
		          DEBUG("No records found for '%s'", DNname);
		
	}

	/* Close the cursor */
	itemname_cursorp->c_close(itemname_cursorp);
	databases_close(&my_stock);

	return skey;
}



/** Fetch GPASSWD struct by given user's global identity token
 * @param gtk user's global identity token
 * @param gpwd GPASSWD structure holding user's mapping information
 * @return key of record in PASSWD database if success, or NULL
 */
char *bdb_getdbgtk_pwd(GIDTOKEN * gtk, GPASSWD * gpwd)
{
	STOCK_DBS my_stock;
	DBC *itemname_cursorp;
	DBT key, data;

	int ret, loop, flag = 0;
	char *skey = NULL;

	ret = loop = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}

	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	/* Get a cursor to the itemname db */
	my_stock.igpasswd_sdbp->cursor(my_stock.igpasswd_sdbp, NULL,
				       &itemname_cursorp, 0);

	key.data = gtk->g_dn;
	key.size = (u_int32_t) strlen(gtk->g_dn) + 1;

	ret =
	    itemname_cursorp->c_get(itemname_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {

			skey = get_db_item_pwd(data.data, gpwd);
			if (skey) {

				if (!strncmp
				    (gpwd->g_idtoken.g_dn, gtk->g_dn,
				     strlen(gtk->g_dn))
				    && !strncmp(gpwd->g_idtoken.g_vo,
						gtk->g_vo,
						strlen(gtk->g_vo))
				    && !strncmp(gpwd->g_idtoken.g_role,
						gtk->g_role,
						strlen(gtk->g_role))) {
					flag = 1;
					skey = strdup(skey);
					break;
				}

			}

			loop =
			    itemname_cursorp->c_get(itemname_cursorp, &key,
						    &data, DB_NEXT_DUP);
		} while (loop == 0);

		if (!flag) {
			if (!loop) {
				if (DEBUG_INFO) DEBUG("%s: loop is null",
					    __FUNCTION__);
				skey = NULL;
			} else
			    if (strncmp
				(gpwd->g_idtoken.g_dn, gtk->g_dn,
				 strlen(gtk->g_dn))) {
				if (DEBUG_INFO) DEBUG("%s: dn is not same",
					    __FUNCTION__);
				skey = NULL;
			} else
			    if (strncmp
				(gpwd->g_idtoken.g_vo, gtk->g_vo,
				 strlen(gtk->g_vo))) {
				if (DEBUG_INFO) DEBUG("%s: vo is not same",
					    __FUNCTION__);
				skey = NULL;
#if 1
				/*YVON: removed test on role because of inconsistency*/
#else
			} else
			    if (strncmp
				(gpwd->g_idtoken.g_role, gtk->g_role,
				 strlen(gtk->g_role))) {
				if (DEBUG_INFO) DEBUG("%s: role is not same %s %s",
						      __FUNCTION__, gpwd->g_idtoken.g_role, gtk->g_role);
				skey = NULL;
#endif
			}
		}

	} else {
		if (DEBUG_INFO) DEBUG("No records found for the token");
	}

	/* Close the cursor */
	itemname_cursorp->c_close(itemname_cursorp);
	databases_close(&my_stock);


	return skey;
}



/** Fetch GPASSWD struct by given user's local identity token
 * @param ltk user's local identity token
 * @param gpwd GPASSWD structure holding user's mapping information
 * @return key of record in PASSWD database if success, or NULL
 */
char *bdb_getdbltk_pwd(LIDTOKEN * ltk, GPASSWD * gpwd)
{
	STOCK_DBS my_stock;
	DBC *itemuid_cursorp;
	DBT key, data;

	int ret, loop;
	char *skey = NULL;

	loop = ret = -1;
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return NULL;
	}


	/* Initialize our DBTs. */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));


	my_stock.ugpasswd_sdbp->cursor(my_stock.ugpasswd_sdbp, NULL,
				       &itemuid_cursorp, 0);

	key.data = &(ltk->g_mappeduid);
	key.size = sizeof(int);


	ret = itemuid_cursorp->c_get(itemuid_cursorp, &key, &data, DB_SET);
	if (!ret) {
		do {
			skey = get_db_item_pwd(data.data, gpwd);
			if (skey) {
				if ((ltk->g_mappeduid ==
				     gpwd->l_idtoken.g_mappeduid)
				    && (ltk->g_mappedgid ==
					gpwd->l_idtoken.g_mappedgid)) {
					skey = strdup(skey);
					break;
				}
			}

			loop =
			    itemuid_cursorp->c_get(itemuid_cursorp, &key,
						   &data, DB_NEXT_DUP);
		} while (loop == 0);

		if (!loop)
			skey = NULL;
	} else {
		if (DEBUG_INFO) DEBUG("No records found for the token");
	}

	/* Close the cursor */
	itemuid_cursorp->c_close(itemuid_cursorp);
	databases_close(&my_stock);


	return skey;
}



/** Store GPASSWD struct to PASSWD database
 * @param gpwen GPASSWD structure holding user's mapping information
 * @return 0 if success, or !0
 */
int bdb_storedben_pwd(GPASSWD * gpwen)
{
	STOCK_DBS my_stock;
	DBT key, data;
	int ret;
	char databuf[MAXDATABUF];
	size_t bufLen, dataLen;

	char randkey[20 + 1];

	create_sranddbkey(randkey, 20);
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return (ret);
	}

	memset(databuf, 0, MAXDATABUF);
	bufLen = 0;
	dataLen = 0;

	dataLen = sizeof(int);
	memcpy(databuf, &(gpwen->l_idtoken.g_mappeduid), dataLen);
	bufLen += dataLen;

	dataLen = sizeof(int);
	memcpy(databuf + bufLen, &(gpwen->l_idtoken.g_mappedgid), dataLen);
	bufLen += dataLen;

        dataLen = sizeof(int);
        memcpy(databuf + bufLen, &(gpwen->g_mount), dataLen);
        bufLen += dataLen;

	dataLen = sizeof(long);
	memcpy(databuf + bufLen, &(gpwen->g_time), dataLen);
	bufLen += dataLen;

	bufLen = pack_string(databuf, gpwen->g_idtoken.g_dn, bufLen);
	bufLen = pack_string(databuf, gpwen->g_idtoken.g_vo, bufLen);
	bufLen = pack_string(databuf, gpwen->g_idtoken.g_role, bufLen);
	bufLen = pack_string(databuf, gpwen->g_idtoken.g_guid, bufLen);
	bufLen =
	    pack_string(databuf, gpwen->l_idtoken.g_mappedname, bufLen);
	bufLen = pack_string(databuf, randkey, bufLen);
	bufLen = pack_string(databuf, gpwen->g_shell, bufLen);
	bufLen = pack_string(databuf, gpwen->g_home, bufLen);

	/* Zero out the DBTs */
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	/* The key is the item's SKU */
	key.data = randkey;
	key.size = (u_int32_t) strlen(randkey) + 1;

	/* The data is the information that we packed into databuf. */
	data.data = databuf;
	data.size = (u_int32_t) bufLen;

	/* Put the data into the database */
	my_stock.gpasswd_dbp->put(my_stock.gpasswd_dbp, 0, &key, &data, 0);

	databases_close(&my_stock);
	return ret;
}



/** Delete GPASSWD struct in PASSWD database by given user's global identity token
 * @param gtk GIDTOKEN structure holding user's global identity token
 * @return 0 if success, or !0
 */
int bdb_deldbgtk_pwd(GIDTOKEN * gtk)
{
	STOCK_DBS my_stock;
	DBT key, data;
	GPASSWD *gpwd = NULL;
	int ret;
	char *skey = NULL;

	gpwd = gpwd_new();
	skey = bdb_getdbgtk_pwd(gtk, gpwd);
	if (!skey) {
		if (DEBUG_INFO) DEBUG("Not found the record");
		return -1;
	}

	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return (ret);
	}

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	key.data = skey;
	key.size = (u_int32_t) strlen(skey) + 1;

	my_stock.gpasswd_dbp->del(my_stock.gpasswd_dbp, 0, &key, 0);

	databases_close(&my_stock);
	return ret;
}




/** Print data in GPASSWD struct as readable format
 * @param gpwen GPASSWD structure holding user's mapping information
 */
void bdb_printdata_pwd(GPASSWD * gpwen)
{
	printf
	    ("DN:\t%s\nVO:\t%s\nRole:\t%s\nGUID:\t%s\nL_name:\t%s\nL_uid:\t%d\nL_gid:\t%d\nHome:\t%s\nShell:\t%s\ntime:\t%ld\nmountcount:\t%d\n",
	     gpwen->g_idtoken.g_dn, gpwen->g_idtoken.g_vo,
	     gpwen->g_idtoken.g_role,gpwen->g_idtoken.g_guid, gpwen->l_idtoken.g_mappedname,
	     gpwen->l_idtoken.g_mappeduid, gpwen->l_idtoken.g_mappedgid,
	     gpwen->g_home, gpwen->g_shell, gpwen->g_time,
	     gpwen->g_mount);
}


/** Print data in buffer as readable format
 * @param vBuf buffer holding user's mapping information
 * @return key of this record if success, or NULL
 */
char *print_db_item_pwd(void *vBuf)
{
	int uid;
	int gid,mount;
	long time;
	size_t buf_pos;
	char *shell, *DNname, *VO, *Role, *GUID, *lname, *sku, *home;
	char *buf = (char *) vBuf;

	uid = *((int *) buf);
	buf_pos = sizeof(int);

	gid = *((int *) (buf + buf_pos));
	buf_pos += sizeof(int);

        mount = *((int *) (buf + buf_pos));
        buf_pos += sizeof(int);

	time = *((long *) (buf + buf_pos));
	buf_pos += sizeof(long);

	DNname = buf + buf_pos;
	buf_pos += strlen(DNname) + 1;

	VO = buf + buf_pos;
	buf_pos += strlen(VO) + 1;

	Role = buf + buf_pos;
	buf_pos += strlen(Role) + 1;

	GUID = buf + buf_pos;
	buf_pos += strlen(GUID) + 1;
		
	lname = buf + buf_pos;
	buf_pos += strlen(lname) + 1;

	sku = buf + buf_pos;
	buf_pos += strlen(sku) + 1;

	shell = buf + buf_pos;
	buf_pos += strlen(shell) + 1;

	home = buf + buf_pos;

	printf("DNname: %s\n", DNname);
	printf("\tVO: %s\n", VO);
	printf("\tRole: %s\n", Role);
	printf("\tGUID: %s\n", GUID);
	printf("\tSKU: %s\n", sku);
	printf("\tname: %s\n", lname);
	printf("\tuid: %i\n", uid);
	printf("\tgid: %i\n", gid);
	printf("\thome:%s\n", home);
	printf("\tShell: %s\n", shell);
	printf("\ttime: %ld\n", time);
	printf("\tmnt counter: %i\n",mount);

	return (DNname);
}


/** Print all the data in GPASSWD database
 */
void bdb_printdb_pwd(void)
{
	STOCK_DBS my_stock;
	DBC *igpasswd_cursorp;
	DBT key, data;
	char *DNname;
	int exit_value, ret;

	init_DBT(&key, &data);
	/* Identify the files that hold our databases */
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	/* Open all databases */
	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret != 0) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return;
	}

	my_stock.gpasswd_dbp->cursor(my_stock.gpasswd_dbp, NULL,
				     &igpasswd_cursorp, 0);

	exit_value = 0;
	while ((ret =
		igpasswd_cursorp->c_get(igpasswd_cursorp, &key, &data,
					DB_NEXT)) == 0) {
		DNname = print_db_item_pwd(data.data);
		if (ret) {
			exit_value = ret;
			break;
		}
	}

	/* Close the cursor */
	igpasswd_cursorp->c_close(igpasswd_cursorp);

	/* close our databases */
	databases_close(&my_stock);


}




/** Dispose the given data in database by self-defined function
 * @param gtk global identity token to find the GPASSWD struct
 * @param optfunc self-defined function to deal with the data. It
 *        must be the type of OPTPWDFUNC.
 * @return 0 if success, or !0
 */

int bdb_operatedata_pwd(GIDTOKEN * gtk, OPTPWDFUNC optfunc)
{
#if 0
	GPASSWD *gpwd = NULL;
	char *skey = NULL;
	int ret = 0;

	gpwd = gpwd_new();
	skey = bdb_getdbgtk_pwd(gtk, gpwd);
	if (!skey) {
		if (DEBUG_INFO) DEBUG("No found the record");
		ret = -1;
		goto err;
	}

	ret = optfunc(gpwd);

      err:
	if (gpwd)
		free(gpwd);
	return ret;

#else
	return -1;
#endif
}



/** Dispose all the  data in database by self-defined function
 * @param optfunc self-defined function to deal with the data. It
 *        must be the type of OPTPWDFUNC.
 * @return 0 if success, or !0
 */
#if 0
int bdb_operatedb_pwd(GPASSWD * t_gpwd, OPTPWDFUNC optfunc)
{
	STOCK_DBS my_stock;
	DBC *igpasswd_cursorp;
	DBT key, data;
	char *skey;
	int exit_ret, ret;
	GPASSWD *gpwd = NULL;

	init_DBT(&key, &data);
	/* Identify the files that hold our databases */
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	/* Open all databases */
	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret != 0) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return -1;
	}

	my_stock.gpasswd_dbp->cursor(my_stock.gpasswd_dbp, NULL,
				     &igpasswd_cursorp, 0);

	exit_ret = 0;
	gpwd = gpwd_new();
	while ((ret =
		igpasswd_cursorp->c_get(igpasswd_cursorp, &key, &data,
					DB_NEXT)) == 0) {
		skey = get_db_item_pwd(data.data, gpwd);
		if (skey) {
			ret = optfunc(gpwd, t_gpwd);
		}

		if (ret) {
			exit_ret = ret;
			break;
		}
	}

	/* Close the cursor */
	igpasswd_cursorp->c_close(igpasswd_cursorp);

	/* close our databases */
	databases_close(&my_stock);

	return exit_ret;
}
#else
int bdb_operatedb_pwd(void *t_gpwd, void *r_gpwd, OPTPWDFUNC optfunc)
{
	STOCK_DBS my_stock;
	DBC *igpasswd_cursorp;
	DBT key, data;
	char *skey;
	int exit_ret, ret;
	GPASSWD *gpwd = NULL;

	init_DBT(&key, &data);
	/* Identify the files that hold our databases */
	initialize_stockdbs(&my_stock);
	set_db_filenames(&my_stock);

	/* Open all databases */
	ret = databases_setup(&my_stock, "xos_dbaux", stderr);
	if (ret != 0) {
		print_error(ret, "databases_setup");
		databases_close(&my_stock);
		return -1;
	}

	my_stock.gpasswd_dbp->cursor(my_stock.gpasswd_dbp, NULL,
				     &igpasswd_cursorp, 0);

	exit_ret = 0;
	gpwd = gpwd_new();
	while ((ret =
		igpasswd_cursorp->c_get(igpasswd_cursorp, &key, &data,
					DB_NEXT)) == 0) {
		skey = get_db_item_pwd(data.data, gpwd);
		if (skey) {
			ret = optfunc(gpwd, t_gpwd, r_gpwd);
		}

		if (ret) {
			exit_ret = ret;
			break;
		}
	}

	/* Close the cursor */
	igpasswd_cursorp->c_close(igpasswd_cursorp);

	/* close our databases */
	databases_close(&my_stock);

	return exit_ret;
}
#endif
