/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <db.h>
#include "xos_utils.h"
#include "bdb_common.h"
#include "xos_dbaux_tmp.h"

#ifndef _XOS_DBAUX_H_

/** Print error as readable string.
 * @param ret error code of database
 * @param func function name
 */
void print_error(int ret, const char *func)
{
	if (ret != 0)
		DEBUG("%s: ERROR: %s\n", func, db_strerror(ret));
}

/** Initiate DBT structure of key and data
 * @param key
 * @param data
 */
void init_DBT(DBT * key, DBT * data)
{
	memset(key, 0, sizeof(DBT));
	memset(data, 0, sizeof(DBT));
}

/** Create random string key
 * @param randstr buffer to hold random string
 * @param len length of random string
 */
void create_sranddbkey(char *randstr, int len)
{
	int index;
	int i;
	static int init = 0;
	static unsigned int seed;

	if (!init) {
		seed = (unsigned int) time((time_t *) 0);
		init = 1;
	}

	for (i = 0; i < len; i++) {
		index =
		    (int) (((float) 62) * rand_r(&seed) /
			   (RAND_MAX + 1.0));

		if (index < 10)
			randstr[i] = (48 + index);
		else if (index < 36)
			randstr[i] = (55 + index);
		else if (index < 62)
			randstr[i] = (61 + index);
	}
	randstr[len] = '\0';
}

/** Pack string to buffer
 * @param buffer pointer of buffer
 * @param string  string to pack
 * @param start_pos position to start storing.
 * @return current length of buffer
 */
size_t pack_string(char *buffer, char *string, size_t start_pos)
{
	size_t string_size;

	string_size = strlen(string) + 1;
	memcpy(buffer + start_pos, string, string_size);

	return (start_pos + string_size);
}

#endif





/** initialize the temporarial info database
 * 
 */
int tmpinfo_db_init(TMPINFO_DBS * tmpdbp)
{
	size_t size;
	int ret = 0;

	if (tmpdbp == NULL) {
		if (DEBUG_INFO) DEBUG("db is null\n");
		return -1;
	}

	tmpdbp->tmpinfo_dbp = NULL;

	tmpdbp->tmpdb_home_dir = TMPDB_DIR;

	size = strlen(TMPDB_DIR) + strlen(TMPINFO_DB) + 1;
	tmpdbp->tmpinfo_db_name = malloc(size);
	snprintf(tmpdbp->tmpinfo_db_name, size, "%s%s",
		 TMPDB_DIR, TMPINFO_DB);

	ret = open_database(&(tmpdbp->tmpinfo_dbp),
			    tmpdbp->tmpinfo_db_name,
			    "xos_dbtmp", stderr, PRIMARY_DB);

	return ret;
}


int tmpinfo_db_release(TMPINFO_DBS * tmpdbp)
{
	int ret = 0;

	if (tmpdbp == NULL) {
		return ret;
	}

	if (tmpdbp->tmpinfo_dbp != NULL) {
		ret = tmpdbp->tmpinfo_dbp->close(tmpdbp->tmpinfo_dbp, 0);
		if (ret != 0)
			fprintf(stderr,
				"tmpinfo database close failed: %s\n",
				db_strerror(ret));
	}

	return ret;
}


/** get item from database 
 */
int tmpinfo_db_get(char *skey, char **databuf, int *buf_len)
{
	TMPINFO_DBS tmpdbp;
	DBT key, data;
	int ret = 0, length = 0;


	if (!skey)
		return -1;
	if (tmpinfo_db_init(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_init\n");
		return -1;
	}

	init_DBT(&key, &data);

	key.data = skey;
	key.size = (u_int32_t) strlen(skey);


	data.flags = DB_DBT_MALLOC;
	ret =
	    tmpdbp.tmpinfo_dbp->get(tmpdbp.tmpinfo_dbp, NULL, &key, &data,
				    0);
	if (ret != 0) {
		tmpdbp.tmpinfo_dbp->err(tmpdbp.tmpinfo_dbp, ret,
					"Error getting for key: '%s'",
					skey);
		print_error(ret, "tmpinfo_db_get:dbp->get");
	} else {
		//if (DEBUG_INFO) DEBUG("key: %s\nvalue: %s\n", skey, (char*)data.data);
		length = strlen((char *) data.data);
		*buf_len = length;
		*databuf = (char *) malloc(length);
		sprintf(*databuf, "%s", (char *) data.data);

	}

	if (tmpinfo_db_release(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG(" fail in tmpinfo_db_releast\n");
		return -1;
	}

	return ret;
}

/** put item to database 
 */
int tmpinfo_db_put(char *skey, char *databuf)
{
	TMPINFO_DBS tmpdbp;
	DBT key, data;
	int ret = 0;


	if (!skey || !databuf)
		return -1;
	if (tmpinfo_db_init(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_int\n");
		return -1;
	}

	init_DBT(&key, &data);

	key.data = skey;
	key.size = (u_int32_t) strlen(skey);

	data.data = databuf;
	data.size = (u_int32_t) strlen((char *) databuf) + 1;

	ret =
	    tmpdbp.tmpinfo_dbp->put(tmpdbp.tmpinfo_dbp, 0, &key, &data, 0);
	print_error(ret, "tmpinfo_db_put");

	if (tmpinfo_db_release(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG(" fail in tmpinfo_db_releast\n");
		return -1;
	}

	return ret;
}


/** delete item from database 
 */
int tmpinfo_db_delete(char *skey)
{
	TMPINFO_DBS tmpdbp;
	DBT key, data;
	int ret = 0;


	if (!skey)
		return -1;
	if (tmpinfo_db_init(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_init\n");
		return -1;
	}

	init_DBT(&key, &data);

	key.data = skey;
	key.size = (u_int32_t) strlen(skey);

	ret = tmpdbp.tmpinfo_dbp->del(tmpdbp.tmpinfo_dbp, 0, &key, 0);
	print_error(ret, "tmpinfo_db_delete");

	if (tmpinfo_db_release(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG(" fail in tmpinfo_db_releast\n");
		return -1;
	}

	return ret;
}

/** Clear all the data */
int tmpinfo_db_clearall()
{
	TMPINFO_DBS tmpdbp;
	DBT key, data;
	DBC *icursorp;
	int ret = 0, pre_ret = 0;

	if (tmpinfo_db_init(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_init\n");
		return -1;
	}


	init_DBT(&key, &data);


	tmpdbp.tmpinfo_dbp->cursor(tmpdbp.tmpinfo_dbp, NULL, &icursorp, 0);


	while ((ret =
		icursorp->c_get(icursorp, &key, &data, DB_NEXT)) == 0) {
		//if (DEBUG_INFO) DEBUG("key -> %s\nvalue -> %s\n\n", (char*)key.data, (char*)data.data);
		if (ret)
			break;
		pre_ret =
		    tmpdbp.tmpinfo_dbp->del(tmpdbp.tmpinfo_dbp, 0, &key,
					    0);
		print_error(ret, "tmpinfo_db_clearall");

	}


	icursorp->c_close(icursorp);

	if (tmpinfo_db_release(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_releast\n");
		return -1;
	}
	if (ret)
		return ret;
	else
		return pre_ret;
}

/** Print all the items in temp database */
void tmpinfo_db_printdb()
{
	TMPINFO_DBS tmpdbp;
	DBT key, data;
	DBC *icursorp;
	int ret = 0;

	if (tmpinfo_db_init(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_init\n");
		return;
	}


	init_DBT(&key, &data);


	tmpdbp.tmpinfo_dbp->cursor(tmpdbp.tmpinfo_dbp, NULL, &icursorp, 0);

	printf("----------- tmp database ---------\n");
	while ((ret =
		icursorp->c_get(icursorp, &key, &data, DB_NEXT)) == 0) {
		printf("key -> %s\nvalue -> %s\n\n", (char *) key.data,
		       (char *) data.data);
		if (ret) {

			break;
		}
	}
	printf("----------------------------------\n");

	icursorp->c_close(icursorp);

	if (tmpinfo_db_release(&tmpdbp)) {
		if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_releast\n");
		return;
	}
}
