/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <db.h>
#include "xos_utils.h"
#include "bdb_common.h"
#include "xos_policy_rule.h"
#include "xos_policy_db.h"

#define	INDEX_HEAD	"index_head"

#ifndef _XOS_DBAUX_H_

/** Print error as readable string.
 * @param ret error code of database
 * @param func function name
 */
void print_error(int ret, const char *func)
{
	if (ret != 0)
		DEBUG("%s: ERROR: %s\n", func, db_strerror(ret));
}

/** Initiate DBT structure of key and data
 * @param key
 * @param data
 */
void init_DBT(DBT * key, DBT * data)
{
	memset(key, 0, sizeof(DBT));
	memset(data, 0, sizeof(DBT));
}

/** Create random string key
 * @param randstr buffer to hold random string
 * @param len length of random string
 */
void create_sranddbkey(char *randstr, int len)
{
	int index;
	int i;
	static int init = 0;
	static unsigned int seed;

	if (!init) {
		seed = (unsigned int) time((time_t *) 0);
		init = 1;
	}

	for (i = 0; i < len - 1; i++) {
		index =
		    (int) (((float) 62) * rand_r(&seed) /
			   (RAND_MAX + 1.0));

		if (index < 10)
			randstr[i] = (48 + index);
		else if (index < 36)
			randstr[i] = (55 + index);
		else if (index < 62)
			randstr[i] = (61 + index);
	}
	randstr[len - 1] = '\0';
}

/** Pack string to buffer
 * @param buffer pointer of buffer
 * @param string  string to pack
 * @param start_pos position to start storing.
 * @return current length of buffer
 */
size_t pack_string(char *buffer, char *string, size_t start_pos)
{
	size_t string_size;

	string_size = strlen(string) + 1;
	memcpy(buffer + start_pos, string, string_size);

	return (start_pos + string_size);
}

#endif


/** initialize the temporarial info database
 * 
 */
int ruleinfo_db_init(RULEINFO_DBS * ruledbp)
{
	size_t size;
	int ret = 0;

	if (ruledbp == NULL) {
		 if (DEBUG_INFO) DEBUG("db is null\n");
		return -1;
	}

	ruledbp->ruleinfo_dbp = NULL;
	ruledbp->ruleindex_dbp = NULL;
	ruledbp->ruledb_home_dir = RULEDB_DIR;

	size = strlen(RULEDB_DIR) + strlen(RULEINFO_DB) + 1;
	ruledbp->ruleinfo_db_name = malloc(size);
	snprintf(ruledbp->ruleinfo_db_name, size, "%s%s",
		 RULEDB_DIR, RULEINFO_DB);
	size = strlen(RULEDB_DIR) + strlen(RULEINDEX_DB) + 1;
	ruledbp->ruleindex_db_name = malloc(size);
	snprintf(ruledbp->ruleindex_db_name, size, "%s%s",
		 RULEDB_DIR, RULEINDEX_DB);

	ret = open_database(&(ruledbp->ruleinfo_dbp),
			    ruledbp->ruleinfo_db_name,
			    "xos_ruledb", stderr, PRIMARY_DB);
	ret = open_database(&(ruledbp->ruleindex_dbp),
			    ruledbp->ruleindex_db_name,
			    "xos_ruledb", stderr, PRIMARY_DB);
	return ret;
}


int ruleinfo_db_release(RULEINFO_DBS * ruledbp)
{
	int ret = 0;

	if (ruledbp == NULL) {
		return ret;
	}

	if (ruledbp->ruleinfo_dbp != NULL) {
		ret =
		    ruledbp->ruleinfo_dbp->close(ruledbp->ruleinfo_dbp, 0);
		if (ret != 0)
			fprintf(stderr,
				"ruleinfo database close failed: %s\n",
				db_strerror(ret));
	}

	if (ruledbp->ruleindex_dbp != NULL) {
		ret =
		    ruledbp->ruleindex_dbp->close(ruledbp->ruleindex_dbp,
						  0);
		if (ret != 0)
			fprintf(stderr,
				"ruleinfo database close failed: %s\n",
				db_strerror(ret));
	}

	return ret;
}


struct rule_t *xrule_new()
{
	struct rule_t *rulep = NULL;
	rulep = malloc(sizeof(*rulep));

	if (rulep == NULL) {
		goto err;
	}
	memset(rulep, 0, sizeof(*rulep));
      err:
	return rulep;
}

/** get item from database 
 */
int ruleinfo_db_getbykey(unsigned int skey, struct rule_t **xrule,
			 char **handler)
{
	RULEINFO_DBS ruledbp;
	DBT ikey, idata;
	DBT key, data;
	int ret = -1;
	//int length = 0;
	//char *dataindex = NULL;

	if (!skey)
		return -1;
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return -1;
	}

	init_DBT(&ikey, &idata);
	init_DBT(&key, &data);

	ikey.data = &skey;
	ikey.size = sizeof(unsigned int);


	idata.flags = DB_DBT_MALLOC;
	ret =
	    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &ikey,
				       &idata, 0);
	if (ret != 0) {
		//ruledbp.ruleindex_dbp->err(ruledbp.ruleindex_dbp, ret, "Error getting for key: '%s'",
		//skey);
		//print_error(ret,"ruleinfo_db_get:dbp->get -> index");
		goto err;
	} else {
		// if (DEBUG_INFO) DEBUG("key: %s\nvalue: %s\n", skey, (char*)data.data);
//              dataindex = (char*) malloc(32);
//              sprintf(dataindex,"%s",(char*)idata.data);
//
//              key.data = dataindex;
//              key.size = (u_int32_t)strlen(dataindex);        

		if (handler) {
			*handler = strdup((char *) idata.data);
			(*handler)[idata.size] = '\0';
		}
		if (!xrule) {
			ret = 0;
			goto err;
		}

		*xrule = xrule_new();
		data.data = *xrule;
		data.ulen = sizeof(struct rule_t);
		data.flags = DB_DBT_USERMEM;

		ret =
		    ruledbp.ruleinfo_dbp->get(ruledbp.ruleinfo_dbp, NULL,
					      &idata, &data, 0);
		if (ret != 0) {
		//	ruledbp.ruleinfo_dbp->err(ruledbp.ruleinfo_dbp,
		//				  ret,
		//				  "Error getting for key: '%s'",
		//				  skey);
			print_error(ret,
				    "ruleinfo_db_get:dbp->get -> rule data");
			goto err;
		} else {
			// if (DEBUG_INFO) DEBUG("key: %s\nvalue: %s\n", skey, (char*)data.data);
			//length = strlen((char*)data.data);
			//*buf_len = length;
			//*buf_len = data.size;
			//*databuf = (char*) malloc(data.size + 1);
			//sprintf(*databuf,"%s",(char*)data.data);
			//*(*databuf + data.size) = '\0';

		}


	}

	ret = 0;

      err:
	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in ruleinfo_db_releast\n");
		return -1;
	}

	return ret;
}


/** get item from database 
 */
int ruleinfo_db_getbyhdler(char *skey, struct rule_t **xrule)
{
	RULEINFO_DBS ruledbp;
	DBT key, data;
	int ret = 0;		// length = 0;


	if (!skey)
		return -1;
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in tmpinfo_db_init\n");
		return -1;
	}

	init_DBT(&key, &data);

	key.data = skey;
	key.size = (u_int32_t) strlen(skey);


	*xrule = xrule_new();
	data.data = *xrule;
	data.ulen = sizeof(struct rule_t);
	data.flags = DB_DBT_USERMEM;

	ret =
	    ruledbp.ruleinfo_dbp->get(ruledbp.ruleinfo_dbp, NULL, &key,
				      &data, 0);
	if (ret != 0) {
		ruledbp.ruleinfo_dbp->err(ruledbp.ruleinfo_dbp, ret,
					  "Error getting for key: '%s'",
					  skey);
		print_error(ret, "tmpinfo_db_get:dbp->get");
	} else {
		// if (DEBUG_INFO) DEBUG("key: %s\nvalue: %s\n", skey, (char*)data.data);
		//length = strlen((char*)data.data);
		//*buf_len = length;
		//*buf_len = data.size;
		//*databuf = (char*) malloc(data.size);
		//sprintf(*databuf,"%s",(char*)data.data);

	}

	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in tmpinfo_db_releast\n");
		return -1;
	}

	return ret;
}


/** put item to database 
 */
int ruleinfo_db_put(struct rule_t *xrule, char **handler)
{
	RULEINFO_DBS ruledbp;
	DBT ikey, idata;
	DBT key, data;
	int ret = 0, indkey = 0, inddata = 0;;
	char newkey[32];
	char *firstkey = INDEX_HEAD;


	if (!xrule)
		return -1;
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_int\n");
		return -1;
	}

	init_DBT(&ikey, &idata);
	init_DBT(&key, &data);

	/* store in ruleinfo_db */
	create_sranddbkey(newkey, 32);
	key.data = newkey;
	key.size = (u_int32_t) strlen(newkey);

	if (handler)
		*handler = strdup(newkey);

	data.data = xrule;
	data.size = sizeof(struct rule_t);
	ret =
	    ruledbp.ruleinfo_dbp->put(ruledbp.ruleinfo_dbp, 0, &key, &data,
				      0);
	if (ret) {
		print_error(ret, "ruleinfo_db_put");
		goto err;
	}


	/* store in ruleindex_db */
	indkey = ruleinfo_db_newhandler();
	ikey.data = &indkey;
	ikey.size = sizeof(unsigned int);
	idata.data = newkey;
	idata.size = (u_int32_t) strlen((char *) newkey);

	ret =
	    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0, &ikey,
				       &idata, 0);
	if (ret) {
		print_error(ret, "ruleinfo_db_put");
		goto err;
	}

	/* change the counter in ruleindex_db */
	init_DBT(&ikey, &idata);
	ikey.data = firstkey;
	ikey.size = (u_int32_t) strlen((char *) firstkey);
	ret =
	    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &ikey,
				       &idata, 0);
	if (ret) {
		print_error(ret, "ruleinfo_db_get");
		goto err;
	}
	inddata = *((unsigned int *) idata.data) + 1;
	idata.data = &inddata;
	idata.size = sizeof(unsigned int);
	ret =
	    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0, &ikey,
				       &idata, 0);
	print_error(ret, "ruleinfo_db_get");
      err:
	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in ruleinfo_db_releast\n");
		return -1;
	}


	return ret;
}



int ruleinfo_db_put_aux(struct rule_t *xrule, int given_key,
			char **handler)
{
	RULEINFO_DBS ruledbp;
	DBT ikey, idata;
	DBT key, data;
	int ret = 0, indkey = 0;	// inddata = 0;;
	char newkey[32];
	//char *firstkey = INDEX_HEAD;


	if (!xrule || given_key > -1)
		return -1;
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_int\n");
		return -1;
	}

	init_DBT(&ikey, &idata);
	init_DBT(&key, &data);

	/* store in ruleinfo_db */
	create_sranddbkey(newkey, 32);
	key.data = newkey;
	key.size = (u_int32_t) strlen(newkey);

	if (handler)
		*handler = strdup(newkey);

	data.data = xrule;
	data.size = sizeof(struct rule_t);
	ret =
	    ruledbp.ruleinfo_dbp->put(ruledbp.ruleinfo_dbp, 0, &key, &data,
				      0);
	if (ret) {
		print_error(ret, "ruleinfo_db_put");
		goto err;
	}


	/* store in ruleindex_db */
	indkey = given_key;

	ikey.data = &indkey;
	ikey.size = sizeof(unsigned int);
	idata.data = newkey;
	idata.size = (u_int32_t) strlen((char *) newkey);

	ret =
	    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0, &ikey,
				       &idata, 0);
	if (ret) {
		print_error(ret, "ruleinfo_db_put");
		goto err;
	}

      err:
	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in ruleinfo_db_releast\n");
		return -1;
	}


	return ret;
}

unsigned int ruleinfo_db_newhandler()
{
	RULEINFO_DBS ruledbp;
	DBT key, data;
	int ret = 0, fret = -1;

	/* default 0 record */
	char *skey = INDEX_HEAD;
	//char *databuf = NULL;
	int head_data = -1;

	//if(!skey) return -1;
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return -1;
	}

	init_DBT(&key, &data);

	key.data = skey;
	key.size = (u_int32_t) strlen((char *) skey);


	data.flags = DB_DBT_MALLOC;
	ret =
	    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &key,
				       &data, 0);
	if (ret != 0) {
		//ruledbp.ruleindex_dbp->err(ruledbp.ruleindex_dbp, ret, "Error getting for key: '%s'",
		//skey);
		//print_error(ret,"ruleinfo_db_get:dbp->get");
	} else {
		// if (DEBUG_INFO) DEBUG("key: %s\nvalue: %s\n", skey, (char*)data.data);
		//length = strlen((char*)data.data);
		//databuf = (char*) malloc(length);
		//sprintf(databuf,"%s",(char*)data.data);
		head_data = *((int *) data.data);

	}

	if (ret == 0 && head_data != -1)
		fret = head_data + 1;
	else {
		/* add a new record of key = 0 */
		int firstdata = 0;
		//init_DBT(&key, &data);
		//key.data = &skey;
		//key.size = (u_int32_t) strlen((char*)skey);           
		data.data = &firstdata;
		data.size = sizeof(unsigned int);
		ret =
		    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0,
					       &key, &data, 0);
		if (ret) {
			print_error(ret, "ruleinfo_db_newhandler");
			fret = -1;
		} else
			fret = 1;

	}

	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in ruleinfo_db_releast\n");
		return -1;
	}

	return fret;
}


int ruleinfo_db_search(struct rule_t *xrule, char **handler)
{
	RULEINFO_DBS ruledbp;
	DBT ikey, idata;
	DBT key, data;
	DBC *icursorp;
	int ret = 0;		// length = 0;
	char hd[33];

	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return -1;
	}


	init_DBT(&ikey, &idata);
	init_DBT(&key, &data);

	ruledbp.ruleindex_dbp->cursor(ruledbp.ruleindex_dbp, NULL,
				      &icursorp, 0);
	while ((ret =
		icursorp->c_get(icursorp, &ikey, &idata, DB_NEXT)) == 0) {
		//printf("key -> %d\nvalue -> %s\n\n", *((unsigned int *)ikey.data), (char*)idata.data);

		if (ret)
			break;

		/* Search the database and compare */
		data.flags = DB_DBT_MALLOC;
		ret =
		    ruledbp.ruleinfo_dbp->get(ruledbp.ruleinfo_dbp, NULL,
					      &idata, &data, 0);
		if (ret != 0) {
			continue;
			ruledbp.ruleinfo_dbp->err(ruledbp.ruleinfo_dbp,
						  ret,
						  "Error getting for key: '%s'",
						  (char *) idata.data);
			print_error(ret,
				    "ruleinfo_db_get:dbp->get -> rule data");
		} else {
			// if (DEBUG_INFO) DEBUG("key: %s\nvalue: %s\n", (char*)idata.data, (char*)data.data);
			if (!memcmp
			    (data.data, (void *) xrule,
			     sizeof(struct rule_t))) {
				snprintf(hd, idata.size + 1, "%s",
					 (char *) idata.data);
				*handler = strdup(hd);
				break;
			}

		}
	}
	icursorp->c_close(icursorp);

	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_releast\n");
		return -1;
	}

	return ret;
}


/** delete item from database 
 */
int ruleinfo_db_delete(char *handler)
{
	RULEINFO_DBS ruledbp;
	DBT key, data;
	DBT ikeyh, idatah;
	DBT ikey, idata;
	int ret = 0, pre_ret = 0;
	//int key_value = 0;


	if (!handler)
		return -1;
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return -1;
	}

	init_DBT(&key, &data);
	init_DBT(&ikey, &idata);
	init_DBT(&ikeyh, &idatah);


	/* remove the rule in ruleinfo_db */
	key.data = handler;
	key.size = (u_int32_t) strlen(handler);

	ret = ruledbp.ruleinfo_dbp->del(ruledbp.ruleinfo_dbp, 0, &key, 0);
	if (ret) {
		print_error(ret, "ruleinfo_db_delete");
		goto err;
	}

	/* adjust and sort the items in ruleindex_db */
	ikeyh.data = INDEX_HEAD;
	ikeyh.size = (u_int32_t) strlen((char *) INDEX_HEAD);
	idatah.flags = DB_DBT_MALLOC;
	ret =
	    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &ikeyh,
				       &idatah, 0);
	if (ret != 0) {
		ruledbp.ruleindex_dbp->err(ruledbp.ruleindex_dbp, ret,
					   "Error getting for key: '%s'",
					   (char *) ikeyh.data);
		print_error(ret, "ruleinfo_db_get:dbp->get -> index");
	} else {
		int total = *((int *) idatah.data);
		int i, j, flag = 0;
		ikey.size = sizeof(int);
		idata.flags = DB_DBT_MALLOC;
		for (i = 1; i <= total; i++) {
			ikey.data = &i;
			ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp,
						   NULL, &ikey, &idata, 0);
			if (!memcmp(key.data, idata.data, idata.size)
			    || flag) {
				pre_ret =
				    ruledbp.ruleindex_dbp->del(ruledbp.
							       ruleindex_dbp,
							       0, &ikey,
							       0);
				print_error(pre_ret,
					    "ruleinfo_db_search:dbp->del");
				if (!flag) {
					j = total - 1;
					idatah.data = &j;
					pre_ret =
					    ruledbp.ruleindex_dbp->
					    put(ruledbp.ruleindex_dbp, 0,
						&ikeyh, &idatah, 0);
					print_error(pre_ret,
						    "ruleinfo_db_search:dbp->del");
					flag = 1;
					continue;
				}
			}

			if (flag) {
				j = i - 1;
				ikey.data = &j;
				pre_ret =
				    ruledbp.ruleindex_dbp->put(ruledbp.
							       ruleindex_dbp,
							       0, &ikey,
							       &idata, 0);
				print_error(pre_ret,
					    "ruleinfo_db_search:dbp->del");
			}
		}
	}
      err:
	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in ruleinfo_db_releast\n");
		return -1;
	}

	return ret;
}

/** Clear all the data */
int ruleinfo_db_clearall()
{
	RULEINFO_DBS ruledbp;
	DBT key, data;
	DBC *icursorp;
	int ret = 0, pre_ret = 0;

	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return -1;
	}


	init_DBT(&key, &data);


	ruledbp.ruleindex_dbp->cursor(ruledbp.ruleindex_dbp, NULL,
				      &icursorp, 0);


	while ((ret =
		icursorp->c_get(icursorp, &key, &data, DB_NEXT)) == 0) {
		// if (DEBUG_INFO) DEBUG("key -> %s\nvalue -> %s\n\n", (char*)key.data, (char*)data.data);
		if (ret)
			break;
		pre_ret =
		    ruledbp.ruleindex_dbp->del(ruledbp.ruleindex_dbp, 0,
					       &key, 0);
		print_error(ret, "ruleinfo_db_clearall");
		pre_ret =
		    ruledbp.ruleinfo_dbp->del(ruledbp.ruleinfo_dbp, 0,
					      &data, 0);
		print_error(ret, "ruleinfo_db_clearall");
	}
	icursorp->c_close(icursorp);

	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_releast\n");
		return -1;
	}
	if (ret)
		return ret;
	else
		return pre_ret;
}

/** Print all the items in temp database */
void ruleinfo_db_printdb()
{
	RULEINFO_DBS ruledbp;
	DBT key, data;
	DBT ikey, idata;
	DBC *icursorp;
	int ret = 0, pre_ret = 0;
	struct rule_t *xrule = NULL;

	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return;
	}


	init_DBT(&key, &data);
	init_DBT(&ikey, &idata);


	xrule = xrule_new();
	data.data = xrule;
	data.ulen = sizeof(struct rule_t);
	data.flags = DB_DBT_USERMEM;

	ruledbp.ruleindex_dbp->cursor(ruledbp.ruleindex_dbp, NULL,
				      &icursorp, 0);

	printf("----------- rule database ---------\n");
	while ((ret =
		icursorp->c_get(icursorp, &ikey, &idata, DB_NEXT)) == 0) {
		//printf("[key:%d]:\nvalue -> %s\n\n", *((unsigned int *)ikey.data), (char*)idata.data);
		if (ret) {

			break;
		}
		if (ikey.size != 4) {
			printf("[key:%s]:%d\n", (char *) ikey.data,
			       *((unsigned int *) idata.data));
		} else {
			printf("[key:%d]:%32s\n",
			       *((unsigned int *) ikey.data),
			       (char *) idata.data);

			pre_ret =
			    ruledbp.ruleinfo_dbp->get(ruledbp.ruleinfo_dbp,
						      NULL, &idata, &data,
						      0);
			if (pre_ret) {
				print_error(ret, "ruleinfo_db_printdb");
				continue;
			}
			ruleinfo_db_printdata(xrule);
			printf("\n\n");
		}

	}
	printf("----------------------------------\n");

	icursorp->c_close(icursorp);

	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_releast\n");
		return;
	}
}


void ruleinfo_db_printdata(struct rule_t *xrule)
{
	int invalid_rule = 0;

	if (!xrule)
		return;

	/* print rule type */
	printf("Rule Type:");
	switch (xrule->r_type) {
	case XPLY_RT_NUL:
		printf("NULL\n");
		invalid_rule = 1;
		break;
	case XPLY_RT_MA:
		printf("Mapping rule\n");
		break;
	case XPLY_RT_AC:
		printf("Access Control rule\n");
		break;
	case XPLY_RT_RU:
		printf("Resource usage\n");
		break;
	case XPLY_RT_AP:
		printf("Application Specific\n");
		break;
	default:
		printf("Undefine rule type !!\n");
		invalid_rule = 1;
	}


	if (invalid_rule)
		return;

	/* print subject_t */
	printf("\tSubject:\n");
	printf("\t\tType: ");
	switch (xrule->r_subject.sub_type) {
	case XPLY_ST_NUL:
		printf("NULL\n");
		break;
	case XPLY_ST_ACT:
		printf("<DN,VO,ROLE>\n");
		break;
	case XPLY_ST_GRP:
		printf("<VO,ROLE,ATTRS>\n");
		break;
	default:
		printf("Undefine subjcet type !!\n");
	}
	printf("\t\tContent_1: %s\n", xrule->r_subject.sub_content_1);
	printf("\t\tContent_2: %s\n", xrule->r_subject.sub_content_2);

	/* print object_t */
	printf("\tObject:\n");
	printf("\t\tType: ");
	switch (xrule->r_object.obj_type) {
	case XPLY_OT_NUL:
		printf("NULL\n");
		break;
	case XPLY_OT_ACT:
		printf("local account\n");
		break;
	case XPLY_OT_GRP:
		printf("local groups\n");
		break;
	case XPLY_OT_FILE:
		printf("file object\n");
		break;
	case XPLY_OT_NODE:
		printf("VO node object\n");
		break;
	case XPLY_OT_QUOTA:
		printf("quota object\n");
		break;
	default:
		printf("Undefine subjcet type !!\n");
	}

	printf("\t\tContent:");
	if(xrule->r_object.obj_type == XPLY_OT_QUOTA) {
		switch (xrule->r_object.obj_content.objc_restype) {
		case XPLY_REST_CPU:
			printf("CPU\n");
			break;
		case XPLY_REST_MEM:
			printf("MEMORY\n");
			break;
		case XPLY_REST_DISK:
			printf("DISK\n");
			break;
		case XPLY_REST_NPROC:
			printf("NPROC\n");
			break;
		case XPLY_REST_NOFILE:
			printf("NOFILE\n");
			break;
		case XPLY_REST_NETWORK:
			printf("NETWORK bandwidth\n");
			break;
		default:
			printf("Undefine object type !! \n");
		}
	} else
		printf("%s\n", xrule->r_object.obj_content.objc_object);


	/* print rule content */
	printf("\tRule Content:\n");
	switch (xrule->r_type) {
	case XPLY_RT_AC:
		printf("\t\tPermission: %o\n", xrule->r_content.rc_perm);
		break;
	case XPLY_RT_RU:
		printf("\t\tquota: ");
		if (xrule->r_content.rc_quota.quo_type == XPLY_QT_ABS)
			printf("[ABS|");
		else
			printf("[REL|");
		printf("%f]\n", xrule->r_content.rc_quota.quo_val);
		break;
	case XPLY_RT_AP:
	case XPLY_RT_MA:
		printf("\t\tDriver Name: %s\n",
		       xrule->r_content.rc_driver.driver_name);
		printf("\t\tDriver Params: %s\n",
		       xrule->r_content.rc_driver.driver_params);
		break;
	default:
		printf("Undefine rule type !!\n");
	}


	/* print valid */
	printf("\tValid: %d\n", xrule->r_valid);

	/* print time stamp */
	printf("\tTimeStamp: %ld\n", (long) xrule->r_time);
}



int ruleinfo_db_change_location(int pre_key, int next_key)
{
	RULEINFO_DBS ruledbp;
	DBT ikey, idata;
	DBT ikey2, idata2;
	DBT key, data;
	int ret = -1;
	int maxrules = 0, i, j, loop_flag;
	//char *dataindex = NULL;

	maxrules = ruleinfo_db_newhandler();
	if (pre_key <= 0 || next_key <= 0
		|| pre_key >= maxrules || next_key >= maxrules)
		return -1;
	else if(pre_key == next_key)
		return 0;
	
	if(pre_key > next_key) loop_flag = 1;
	else loop_flag = 0;	
	
	if (ruleinfo_db_init(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG("fail in ruleinfo_db_init\n");
		return -1;
	}

	init_DBT(&ikey, &idata);
	init_DBT(&ikey2, &idata2);
	init_DBT(&key, &data);

	ikey.data = &pre_key;
	ikey.size = sizeof(unsigned int);


	idata.flags = DB_DBT_MALLOC;
	ret =
	    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &ikey,
				       &idata2, 0);
	if (ret != 0) {
		goto err;
	} else {

		for(i = pre_key, j= i-1; j >= next_key && loop_flag; i--,j--) {
			ikey.data = &j;
			ikey.size = sizeof(unsigned int);
			ret =
			    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &ikey,
						       &idata, 0);
			if (ret != 0)
				goto err;

			ikey.data = &i;
				
			ret =
			    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0, &ikey,
						       &idata, 0);
			if (ret) {
				print_error(ret, "ruleinfo_db_put");
				goto err;
			}
					
		}
		
		for(i = pre_key, j= i+1; j <= next_key && !loop_flag; i++,j++) {
			ikey.data = &j;
			ikey.size = sizeof(unsigned int);
			ret =
			    ruledbp.ruleindex_dbp->get(ruledbp.ruleindex_dbp, NULL, &ikey,
						       &idata, 0);
			if (ret != 0)
				goto err;

			ikey.data = &i;
				
			ret =
			    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0, &ikey,
						       &idata, 0);
			if (ret) {
				print_error(ret, "ruleinfo_db_put");
				goto err;
			}
					
		}
		
	}
	
	ikey.data = &next_key;

	ret =
	    ruledbp.ruleindex_dbp->put(ruledbp.ruleindex_dbp, 0, &ikey,
				       &idata2, 0);
	if (ret) {
		print_error(ret, "ruleinfo_db_put");
		goto err;
	}
			
	ret = 0;

      err:
	if (ruleinfo_db_release(&ruledbp)) {
		 if (DEBUG_INFO) DEBUG(" fail in ruleinfo_db_releast\n");
		return -1;
	}

	return ret;	
}



