/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/**
 * @file dn_mapper.c
 * @brief Global identity mapping functionality.
 *
 * This file implements identity mapping module support.
 */
#include <stdio.h>		/* snprintf, sscanf */
#include <string.h>		/* strcpy, strlen */
#include <pthread.h>		/* pthread_mutex* */

#include "libnss_xos.h"		/* _nss_xos_opts */

/* mapper modules */
#include "dn_mapper_rawdn.h"
#include "dn_mapper_scapechar.h"

/** Lock to protect module_initialized */
static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
/** Set to 1 if mapper_module has already been initialized */
static int module_initialized = 0;
/** Struct with callbacks to mapper functions */
static struct _nss_xos_dn_mapper_module mapper_module;

static int init_module(struct _nss_xos_dn_mapper_module *m)
{
	int error = 0;

	pthread_mutex_lock(&lock);
	if (!module_initialized) {	/* initialize module */
		NSS_DEBUG("NSS:%s: Initializing module (%d)", __FUNCTION__,
			  _nss_xos_opts.m_scheme);
		switch (_nss_xos_opts.m_scheme) {
		case MAP_SCH_RAWDN:
			error = _nss_xos_dn_mapper_rawdn_new(m);
			break;
		case MAP_SCH_SCAPECHAR:
			error = _nss_xos_dn_mapper_scapechar_new(m);
			break;
		/** \todo Add support for more mappers? */
		default:
			error = -1;	/* error */
		}
		NSS_DEBUG("NSS:%s: error = %d", __FUNCTION__, error);
		if (!error)
			module_initialized = 1;
	}
	pthread_mutex_unlock(&lock);
	return error;
}

int _nss_xos_dn_mapper_is_map(const char *name)
{
	if (init_module(&mapper_module) != 0)
		return 0;

	return (*mapper_module.is_map) (name);
}

size_t _nss_xos_dn_mapper_dn2name(const char *dn, char *buf, size_t bufsiz)
{
	if (init_module(&mapper_module) != 0)
		return (size_t) 0;

	return (*mapper_module.dn2name) (dn, buf, bufsiz);
}

size_t _nss_xos_dn_mapper_name2dn(const char *name, char *buf,
				  size_t bufsiz)
{
	if (init_module(&mapper_module) != 0)
		return (size_t) 0;

	return (*mapper_module.name2dn) (name, buf, bufsiz);
}
