/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/**
 * @file dn_mapper_scapechar.c
 * @brief SCAPECHAR mapper module.
 *
 * This mapper module that performs a mapping based on introducing a scape char
 * ('x') followed by an hexadecimal number to avoid characters that may cause
 * conflicts when used in local usernames (i.e. '/', '=', ' ', etc.).
 */
#include <stdio.h>		/* snprintf, sscanf */
#include <string.h>		/* strlen */

#include "dn_mapper_scapechar.h"

static int dn_mapper_is_map_scapechar(const char *name)
{
	return name[0] == SCAPE_CHAR && name[1] == '2' && name[2] == 'f';
}

static size_t dn_mapper_dn2name_scapechar(const char *dn, char *buf,
					  size_t bufsiz)
{
	size_t len, i_dn, i_buf;
	int done = 0;

	len = strlen(dn);
	i_dn = i_buf = 0;
	while (!done && i_buf < bufsiz) {
		char c = dn[i_dn++];
		if (c == '\0') {
			buf[i_buf++] = '\0';
			done = 1;	/* done! */
		} else if ((c != SCAPE_CHAR) &&
			   ((c >= 'a' && c <= 'z') ||
			    (c >= '0' && c <= '9' && i_dn > 1))) {
			buf[i_buf++] = c;	/* no mapping needed */
		} else if (i_buf + 3 < bufsiz) {
			char b[3];
			snprintf(b, 3, "%02x", (unsigned char) c);
			buf[i_buf++] = SCAPE_CHAR;
			buf[i_buf++] = b[0];
			buf[i_buf++] = b[1];
		} else {
			return 0;	/* buffer too small */
		}
	}
	if (!done)
		return 0;	/* buffer too small */
	else
		return i_buf;
}

static size_t dn_mapper_name2dn_scapechar(const char *name, char *buf,
					  size_t bufsiz)
{
	size_t len, i_name, i_buf;
	int done = 0;

	len = strlen(name);
	i_name = i_buf = 0;
	while (!done && i_buf < bufsiz) {
		char c = name[i_name++];
		if (c == '\0') {
			buf[i_buf++] = '\0';
			done = 1;	/* done! */
		} else if (c != SCAPE_CHAR) {
			buf[i_buf++] = c;	/* no mapping needed */
		} else if (i_name + 2 <= len) {
			unsigned int val;
			sscanf(name + i_name, "%2x", &val);
			i_name = i_name + 2;
			buf[i_buf++] = (char) val;
		} else {
			return 0;	/* buffer too small */
			/** \todo Error control: what if name = "x2"? "xz"? */
		}
	}
	if (!done)
		return 0;	/* buffer too small */
	else
		return i_buf;
}

int _nss_xos_dn_mapper_scapechar_new(struct _nss_xos_dn_mapper_module *m)
{
	if (m == NULL)
		return 1;
	m->is_map = &dn_mapper_is_map_scapechar;
	m->dn2name = &dn_mapper_dn2name_scapechar;
	m->name2dn = &dn_mapper_name2dn_scapechar;
	return 0;
}
