/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/**
 * @file group.c
 * @brief Group database handling for XtreemOS NSSwitch module.
 *
 * This file handles the different NSS functions related to group information
 * (i.e. getgrnam(), getgrgid(), etc.).
 */
#include "libnss_xos.h"		/* BUF_*, NSS_STATUS*, _nss_xos_* */
#if 0
#include "xos_dbaux.h"
#else
#include "xos_ams.h"
#endif
#include "xos_utils.h"		/* DEBUG */
#include "dn_mapper.h"		/* _nss_xos_dn_mapper_* */

NSS_STATUS _nss_xos_getgrnam_r(const char *name, GR_PARAMS, DEF_PARAMS)
{
	int n_mem = 1;		/* Show only the current user */
	GGROUPS *ggrp = NULL;
	char *skey;
	NSS_STATUS nss_ret = NSS_STATUS_SUCCESS;
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s(\"%s\") called (buflen=%d)",
		  __FUNCTION__, name, (int) buflen);
	if (name == NULL || name[0] == '\0')
		return NSS_STATUS_NOTFOUND;
	ggrp = ggrp_new();
	if (ggrp == NULL)
		return NSS_STATUS_NOTFOUND;

#if 0
	skey = bdb_getdbname_grp(name, ggrp);
	if (!skey)
		skey = bdb_getdbgname_grp((char *) name, ggrp);
	if (!skey) {
		free(ggrp);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("Freeing skey...");
	free(skey);
#else
	if (amsclient_invmappinginfo
	    (NULL, 0, (char *) name, 0, NULL, ggrp)) {
		NSS_DEBUG("NSS:Can not get GGROUPS");
		return NSS_STATUS_NOTFOUND;
	}
#endif




	DEBUG_ASSERT(ggrp != NULL);
	DEBUG_ASSERT(ggrp->l_grptoken.g_mappedname[0] != '\0');
	DEBUG_ASSERT(ggrp->l_grptoken.g_mappedgid != 0);

	if(ggrp == NULL || ggrp->l_grptoken.g_mappedname[0] == '\0' || ggrp->l_grptoken.g_mappedgid == 0)
		return NSS_STATUS_NOTFOUND;

	/* DEBUG_ASSERT(ggrp->l_grptoken.g_members[0] != 0); */
	BUF_CHR(buffer, buflen, gr->gr_name, ggrp->l_grptoken.g_mappedname,
		errnop);
	BUF_CHR(buffer, buflen, gr->gr_passwd, "XOS-cert", errnop);
	gr->gr_gid = ggrp->l_grptoken.g_mappedgid;
	/* n_mem = 3; */
	gr->gr_mem = (char **) buffer;/** \todo Check for overflow first? */
	BUF_SKIP(buffer, buflen, (n_mem + 1) * sizeof(char *), errnop);
	BUF_CHR(buffer, buflen, gr->gr_mem[0], ggrp->l_grptoken.g_members,
		errnop);
#if 0
	/** \todo add VO members? */
	BUF_CHR(buffer, buflen, gr->gr_mem[1], "usr2", errnop);
	BUF_CHR(buffer, buflen, gr->gr_mem[2], "usr3", errnop);
#endif
	gr->gr_mem[n_mem] = NULL;
      buff_overflow:
	if (ggrp != NULL)
		free(ggrp);
	return nss_ret;
}

NSS_STATUS _nss_xos_getgrgid_r(gid_t gid, GR_PARAMS, DEF_PARAMS)
{
	int n_mem = 1;		/* Show only the current user */
	GGROUPS *ggrp = NULL;
	char *skey;
	NSS_STATUS nss_ret = NSS_STATUS_SUCCESS;
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s(%u) called (buflen=%d)",
		  __FUNCTION__, (unsigned int) gid, (int) buflen);

	ggrp = ggrp_new();
	if (ggrp == NULL)
		return NSS_STATUS_NOTFOUND;
#if 0
	skey = bdb_getdbgid_grp(gid, ggrp);
	if (!skey) {
		free(ggrp);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("Freeing skey...");
	free(skey);
#else
	if (amsclient_invmappinginfo(NULL, 0, NULL, gid, NULL, ggrp)) {
		NSS_DEBUG("NSS: Can not get GGROUPS by gid");
		return NSS_STATUS_NOTFOUND;
	}
#endif


	DEBUG_ASSERT(ggrp != NULL);
	DEBUG_ASSERT(ggrp->l_grptoken.g_mappedname[0] != '\0');
	DEBUG_ASSERT(ggrp->l_grptoken.g_mappedgid != 0);
	DEBUG_ASSERT(ggrp->l_grptoken.g_members[0] != 0);

	if(ggrp == NULL || ggrp->l_grptoken.g_mappedname[0] == '\0' || ggrp->l_grptoken.g_members[0] == 0)
		return NSS_STATUS_NOTFOUND;

	BUF_CHR(buffer, buflen, gr->gr_name, ggrp->l_grptoken.g_mappedname,
		errnop);
	BUF_CHR(buffer, buflen, gr->gr_passwd, "XOS-cert", errnop);
	gr->gr_gid = ggrp->l_grptoken.g_mappedgid;
	/* n_mem = 3; */
	gr->gr_mem = (char **) buffer;/** \todo Check for overflow first? */
	BUF_SKIP(buffer, buflen, (n_mem + 1) * sizeof(char *), errnop);
	BUF_CHR(buffer, buflen, gr->gr_mem[0], ggrp->l_grptoken.g_members,
		errnop);
#if 0
	/** \todo add VO members? */
	BUF_CHR(buffer, buflen, gr->gr_mem[1], "usr2", errnop);
	BUF_CHR(buffer, buflen, gr->gr_mem[2], "usr3", errnop);
#endif
	gr->gr_mem[n_mem] = NULL;
      buff_overflow:
	if (ggrp != NULL)
		free(ggrp);
	return nss_ret;
}

NSS_STATUS _nss_xos_getgrent_r(GR_PARAMS, DEF_PARAMS)
{
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s called (buflen=%d)", __FUNCTION__, (int) buflen);
	/** \todo This function could be problematic in XtreemOS, too many entries? privacy? */
	/* Possible return values:
	   - NSS_STATUS_UNAVAIL
	   - return no users
	   - return only the current user
	   - return only users in the same default VO */
	return NSS_STATUS_NOTFOUND;
}

NSS_STATUS _nss_xos_endgrent(void)
{
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s called", __FUNCTION__);
	/** \todo Close connections? */
	return NSS_STATUS_NOTFOUND;
}

NSS_STATUS _nss_xos_setgrent(void)
{
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s called", __FUNCTION__);
	/** \todo Open connection to an XtreemOS server? */
	return NSS_STATUS_NOTFOUND;
}
