/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/**
 * @file libnss_xos.c
 * @brief XtreemOS NSSwitch module support functions.
 *
 * This file implements several support functions for the XtreemOS NSSwitch
 * module. Currently, parsing of nss_xos.conf is implemented here.
 */
#include <pwd.h>		/* struct passwd */
#include <stdio.h>		/* fprintf,fopen,fgets,fclose */
#include <sys/types.h>		/* uid_t */
#include <unistd.h>		/* getuid */
#include <pthread.h>		/* pthread_mutex* */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif				/* HAVE_CONFIG_H */

#include "libnss_xos.h"		/* _nss_xos_* */
#include "xos.h"		/* BUFFERSIZE, NSS_CONF */
#include "xos_utils.h"		/* DEBUG, xos_is_user_global */

#if ENABLE_DEBUG_NSS
/* nothing*/
#else				/* ENABLE_DEBUG_NSS */
# define DEBUG(args...) ((void)(0))
#endif				/* ENABLE_DEBUG_NSS */

/* XOS-NSS default options */
struct _nss_xos_config_options _nss_xos_opts = {
	0,
	MAP_SCH_RAWDN,
	DB_ACCESS_ALL
};

static void trim(char *buffer)
{
	/* remove trailing [ \t]*\r?\n */
	size_t len;

	len = strlen(buffer);
	if (len > 0 && buffer[len - 1] == '\n')
		buffer[len-- - 1] = '\0';
	if (len > 0 && buffer[len - 1] == '\r')
		buffer[len-- - 1] = '\0';
	while (len > 0 &&
	       (buffer[len - 1] == ' ' || buffer[len - 1] == '\t')) {
		buffer[len-- - 1] = '\0';
	}
}

static int parse_line(char *buffer, struct _nss_xos_config_options *opts)
{
	char *key, *key2, *val, *val2;

	trim(buffer);
	key = buffer;
	while (*key == ' ' || *key == '\t')
		key++;		/* skip spaces */
	/* skip empty lines and comments */
	if (*key == '\0' || *key == '#')
		return 0;

	key2 = key;
	while (*key2 != ' ' && *key2 != '\t' && *key2 != '='
	       && *key2 != '\0')
		key2++;
	if (*key2 == '\0')
		return -1;	/* parse error */
	*key2 = '\0';

	val = key2 + 1;		/* search for value */
	while (*val == ' ' || *val == '\t' || *val == '=')
		val++;
	val2 = val;
	while (*val2 != '\0' && *val2 != ' ' && *val2 != '\t')
		val2++;
	*val2 = '\0';

	if (!strcmp(key, "debug")) {
		if (!strcmp(val, "true")) {
			/* temporarily disable
			   DEBUG("NSS: Debug mode on"); */
			opts->debug = 1;
		} else if (!strcmp(val, "false")) {
			opts->debug = 0;
		} else {
			DEBUG("NSS: Debug mode not recognized");
		}
	} else if (!strcmp(key, "mapping_scheme")) {
		if (!strcmp(val, "rawdn")) {
			/* temporarily disable
			   DEBUG("NSS: Mapping scheme: Raw DN"); */
			opts->m_scheme = MAP_SCH_RAWDN;
		} else if (!strcmp(val, "scapechar")) {
			/* temporarily disable
			   DEBUG("NSS: Mapping scheme: Scape Char"); */
			opts->m_scheme = MAP_SCH_SCAPECHAR;
		} else {
			DEBUG("NSS: Mapping scheme not recognized");
		}
	} else if (!strcmp(key, "db_access")) {
		if (!strcmp(val, "all")) {
			/* temporarily disable
			   DEBUG("NSS: Access to database: All users"); */
			opts->db_access = DB_ACCESS_ALL;
		} else if (!strcmp(val, "root")) {
			/* temporarily disable
			   DEBUG("NSS: Access to database: Global users and root"); */
			opts->db_access = DB_ACCESS_ROOT;
		} else if (!strcmp(val, "restricted")) {
			/* temporarily disable
			   DEBUG("NSS: Access to database: Restricted to global "
			   "users"); */
			opts->db_access = DB_ACCESS_RESTRICTED;
		} else {
			DEBUG
			    ("NSS: Access to database mode not recognized");
		}
	}
	return 0;
}

/* returns 0 if success */
int _nss_xos_loadconf(struct _nss_xos_config_options *opts)
{
	static pthread_mutex_t opt_mutex = PTHREAD_MUTEX_INITIALIZER;
	static int opt_loaded = 0;	/* options have been loaded (protected by
					 * opt_mutex) */

	FILE *fd;
	char buffer[BUFFERSIZE];
	int lineno = 0;

	pthread_mutex_lock(&opt_mutex);
	if (!opt_loaded) {
		if (_nss_xos_opts.debug)
			DEBUG("NSS:%s: Loading config file", __FUNCTION__);
		fd = fopen(NSS_CONF, "r");
		if (fd == NULL) {
			DEBUG("NSS:%s: Error opening config file %s",
			      __FUNCTION__, NSS_CONF);
			opt_loaded = 1;	/* use defaults */
			pthread_mutex_unlock(&opt_mutex);
			return -1;
		}
		while (fgets(buffer, BUFFERSIZE, fd)) {
			lineno++;
			if (parse_line(buffer, opts) < 0) {
				/* parse error: ignore line and continue */
				DEBUG
				    ("NSS:%s: Parse error in config file %s "
				     "line %d. Continuing...",
				     __FUNCTION__, NSS_CONF, lineno);
			}
		}
		fclose(fd);
		opt_loaded = 1;
	}
	pthread_mutex_unlock(&opt_mutex);
	return 0;
}

int _nss_xos_db_access_allowed(const struct _nss_xos_config_options *opts)
{
	int ret;

	switch (opts->db_access) {
	case DB_ACCESS_ALL:
		ret = 1;
		break;
	case DB_ACCESS_ROOT:
		ret = getuid() == ((uid_t) 0) || xos_is_user_global();
		break;
	case DB_ACCESS_RESTRICTED:
		ret = xos_is_user_global();
		break;
	default:
		ret = 0;
	}
	return ret;
}
