/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/**
 * @file passwd.c
 * @brief Passwd database handling for XtreemOS NSSwitch module.
 *
 * This file handles the different NSS functions related to user information
 * (i.e. getpwnam(), getpwuid(), etc.).
 */
#include "libnss_xos.h"		/* BUF_*, NSS_STATUS*, _nss_xos_* */
#include "xos_ams.h"
#include "xos_utils.h"		/* DEBUG */
#include "dn_mapper.h"		/* _nss_xos_dn_mapper_* */
/* return value:
 * 0: success
 * 1: error
 */
static int get_idtoken(char *idstr, char **mapped_dn, char **mapped_vo,
		       char **mapped_role)
{
	char *field, *last;
	int no = 0;

	if (idstr == NULL || idstr[0] == '\0')
		return 1;
	if (mapped_dn != NULL)
		*mapped_dn = idstr;
	if (mapped_vo != NULL)
		*mapped_vo = NULL;
	if (mapped_role != NULL)
		*mapped_role = NULL;

	if ((field = strtok_r(idstr, "#", &last)) == NULL)
		return 1;
	if (mapped_dn != NULL)
		*mapped_dn = field;
	while ((field = strtok_r(NULL, "#", &last)) != NULL && no++ < 3) {
		if (no == 1 && mapped_vo != NULL)
			*mapped_vo = field;
		if (no == 2 && mapped_role != NULL)
			*mapped_role = field;
	}
	return 0;
}

NSS_STATUS _nss_xos_getpwnam_r(const char *name, PW_PARAMS, DEF_PARAMS)
{
	char dn[1024];
	GPASSWD *gpwd = NULL, *tgpwd = NULL;
	char *skey;
	char *mapped_vo, *mapped_dn, *mapped_role;
	char pw_gecos[MAXFIELD + MAXFIELD + MAXFIELD + 3];
	int ret;
	NSS_STATUS nss_ret;

	syslog(LOG_INFO,"%s: enter ...",__FUNCTION__);
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s(\"%s\") called (buflen=%d)", __FUNCTION__, name,
		  (int) buflen);
	if (name == NULL || name[0] == '\0')
		return NSS_STATUS_NOTFOUND;
	if (_nss_xos_dn_mapper_is_map(name)) {
		/* name is a DN */
		if (_nss_xos_dn_mapper_name2dn(name, dn, 1024) == 0)
			return NSS_STATUS_NOTFOUND;
		ret =
		    get_idtoken(dn, &mapped_dn, &mapped_vo, &mapped_role);
		if (ret != 0)
			return NSS_STATUS_NOTFOUND;
		NSS_DEBUG("NSS:%s: get_idtoken => \"%s\" \"%s\" \"%s\"",
			  __FUNCTION__, mapped_dn, mapped_vo, mapped_role);
		if (mapped_vo == NULL) {
			tgpwd = gpwd_new();
			if (tgpwd == NULL)
				return NSS_STATUS_NOTFOUND;
			skey = bdb_getdbuid_pwd(getuid(), tgpwd);
			if (skey) {	/* XXX: grid user: same VO? */
				NSS_DEBUG("Freeing skey...");
				free(skey);
				mapped_vo = tgpwd->g_idtoken.g_vo;
			} else {	/* XXX: default VO? */
				mapped_vo = "ictgrid";
			}
		}
		if (mapped_role == NULL)
			mapped_role = "OTHER";

		gpwd = gpwd_new();
		if (amsclient_mappinginfo
		    (mapped_dn, mapped_vo, mapped_role, gpwd, NULL)) {
			NSS_DEBUG
			    ("Can not get mapped information from ams");
			if (gpwd){
				free(gpwd);
				gpwd= NULL;
			}
		}


	} else {
		/* name is not a DN, check if it's a local mapped account */
		gpwd = gpwd_new();
		if (gpwd == NULL)
			return NSS_STATUS_NOTFOUND;
		if (amsclient_invmappinginfo
		    ((char *) name, 0, NULL, 0, gpwd, NULL))
			return NSS_STATUS_NOTFOUND;
		mapped_dn = gpwd->g_idtoken.g_dn;
		mapped_vo = gpwd->g_idtoken.g_vo;
		mapped_role = gpwd->g_idtoken.g_role;
	}
	nss_ret = NSS_STATUS_NOTFOUND;
	if (gpwd) {
		NSS_DEBUG("NSS:%s: NAME: %s", __FUNCTION__,
			  gpwd->l_idtoken.g_mappedname);
		/** \todo this looks like a bug in xos_mapping */
		DEBUG_ASSERT(gpwd->l_idtoken.g_mappedname != NULL &&
			     gpwd->l_idtoken.g_mappedname[0] != '\0');
		pw->pw_uid = gpwd->l_idtoken.g_mappeduid;
		DEBUG_ASSERT(gpwd->l_idtoken.g_mappeduid != 0);
		pw->pw_gid = gpwd->l_idtoken.g_mappedgid;
		DEBUG_ASSERT(gpwd->l_idtoken.g_mappedgid != 0);

		//snprintf(pw_gecos, strlen(mapped_dn)+strlen(mapped_vo)+strlen(mapped_role), "%s#%s#%s", mapped_dn,
		//       mapped_vo, mapped_role);
		/* XXX: is this correct? */
		//setenv("XOS_ENV", pw_gecos, 1);
/*XXX:		BUF_CHR(buffer, buflen, pw->pw_name,
			gpwd->l_idtoken.g_mappedname, errnop); */
/** \todo name should be gpwd->l_idtoken.g_mappedname */
		BUF_CHR(buffer, buflen, pw->pw_name,
			gpwd->l_idtoken.g_mappedname, errnop);
		BUF_CHR(buffer, buflen, pw->pw_passwd, "XOS-cert", errnop);
		BUF_CHR(buffer, buflen, pw->pw_gecos, name, errnop);
		BUF_CHR(buffer, buflen, pw->pw_dir, gpwd->g_home, errnop);
		BUF_CHR(buffer, buflen, pw->pw_shell, gpwd->g_shell,
			errnop);
		nss_ret = NSS_STATUS_SUCCESS;
	}
/* the label buff_overflow is used by BUF_* */
      buff_overflow:
	/* Free memory when buffer too small! */

	if (tgpwd != NULL)
		free(tgpwd);
	NSS_DEBUG("NSS:%s:  finish ...",__FUNCTION__);
	syslog(LOG_INFO,"%s: finish ...",__FUNCTION__);
	return nss_ret;
}

NSS_STATUS _nss_xos_getpwuid_r(uid_t uid, PW_PARAMS, DEF_PARAMS)
{
	GPASSWD *gpwd = NULL;
	char *skey;
	char gecos[MAXFIELD + MAXFIELD + MAXFIELD + 3];
	NSS_STATUS nss_ret;

	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s(%u) called (buflen=%d)", __FUNCTION__,
		  (unsigned int) uid, (int) buflen);
	nss_ret = NSS_STATUS_NOTFOUND;
	gpwd = gpwd_new();
	if (gpwd == NULL)
		return NSS_STATUS_NOTFOUND;
	if (amsclient_invmappinginfo(NULL, uid, NULL, 0, gpwd, NULL))
		return NSS_STATUS_NOTFOUND;
	//bdb_printdata_pwd(gpwd);
	//sprintf(gecos, "%s#%s#%s", gpwd->g_idtoken.g_dn,
	//	gpwd->g_idtoken.g_vo, gpwd->g_idtoken.g_role);
	
	NSS_DEBUG("NSS:%s: enter ...", __FUNCTION__);
	DEBUG_ASSERT(gpwd->l_idtoken.g_mappedname != NULL
		     && gpwd->l_idtoken.g_mappedname[0] != '\0');
	DEBUG_ASSERT(gpwd->g_idtoken.g_dn != NULL
		     && gpwd->g_idtoken.g_dn[0] != '\0');
	DEBUG_ASSERT(gpwd->g_idtoken.g_vo != NULL
		     && gpwd->g_idtoken.g_vo[0] != '\0');
	DEBUG_ASSERT(gpwd->g_idtoken.g_role != NULL
		     && gpwd->g_idtoken.g_role[0] != '\0');

	pw->pw_uid = gpwd->l_idtoken.g_mappeduid;
	DEBUG_ASSERT(gpwd->l_idtoken.g_mappeduid != 0);
	pw->pw_gid = gpwd->l_idtoken.g_mappedgid;
	DEBUG_ASSERT(gpwd->l_idtoken.g_mappedgid != 0);

	//BUF_CHR(buffer, buflen, pw->pw_name, gpwd->l_idtoken.g_mappedname, errnop);
	/* BUF_CHR(buffer, buflen, pw->pw_name,
	   gpwd->g_idtoken.g_dn, errnop); */
	/* XXX: Use BUF_MAP instead */
	BUF_MAP(dn2name, buffer, buflen, pw->pw_name, gpwd->g_idtoken.g_dn,
		errnop);
	BUF_CHR(buffer, buflen, pw->pw_passwd, "XOS-cert", errnop);
	BUF_CHR(buffer, buflen, pw->pw_gecos, gecos, errnop);
	BUF_CHR(buffer, buflen, pw->pw_dir, gpwd->g_home, errnop);
	BUF_CHR(buffer, buflen, pw->pw_shell, gpwd->g_shell, errnop);
	nss_ret = NSS_STATUS_SUCCESS;
/* the label buff_overflow is used by BUF_* */
	NSS_DEBUG("NSS:%s: finish ...",__FUNCTION__);
      buff_overflow:
	free(gpwd);
	return nss_ret;
}

NSS_STATUS _nss_xos_getpwent_r(PW_PARAMS, DEF_PARAMS)
{
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s called (buflen=%d)", __FUNCTION__, (int) buflen);
	/** \todo This function could be problematic in XtreemOS, too many entries? privacy? */
	/* Possible return values:
	   - NSS_STATUS_UNAVAIL
	   - return no users
	   - return only the current user
	   - return only users in the same default VO */
	return NSS_STATUS_NOTFOUND;
}

NSS_STATUS _nss_xos_endpwent(void)
{
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s called", __FUNCTION__);
	/** \todo Close connections? */
	return NSS_STATUS_NOTFOUND;
}

NSS_STATUS _nss_xos_setpwent(void)
{
	if (_nss_xos_loadconf(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING, "NSS:%s: Couldn't load configuration",
			__FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	if (!_nss_xos_db_access_allowed(&_nss_xos_opts)) {
		NSS_LOG(LOG_WARNING,
			"NSS:%s: Unauthorized attempt to access "
			"global users' database", __FUNCTION__);
		return NSS_STATUS_NOTFOUND;
	}
	NSS_DEBUG("NSS:%s called", __FUNCTION__);
	/** \todo Open connection to an XtreemOS server? */
	return NSS_STATUS_NOTFOUND;
}
