/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file pam_modutil.c
 * @brief The auxility codes to support PAM module in low version
 *
 * The codes implemented based on the utility codes in PAM high version 
*/
#if(__LIBPAM_VERSION < 99)
#include "pam_modutil.h"

#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct passwd *pam_modutil_getpwnam(pam_handle_t * pamh, const char *user)
{
#ifdef HAVE_GETPWNAM_R

	void *buffer = NULL;
	size_t length = PWD_INITIAL_LENGTH;

	do {
		int status;
		void *new_buffer;
		struct passwd *result = NULL;

		new_buffer =
		    realloc(buffer, sizeof(struct passwd) + length);
		if (new_buffer == NULL) {

			D(("out of memory"));

			/* no memory for the user - so delete the memory */
			if (buffer) {
				free(buffer);
			}
			return NULL;
		}
		buffer = new_buffer;

		/* make the re-entrant call to get the pwd structure */
		errno = 0;
		status = getpwnam_r(user, buffer,
				    sizeof(struct passwd) +
				    (char *) buffer, length, &result);
		if (!status && (result == buffer)) {
			char *data_name;
			const void *ignore;
			int i;

			data_name =
			    malloc(strlen("_pammodutil_getpwnam") + 1 +
				   strlen(user) + 1 + intlen(INT_MAX) + 1);
			if ((pamh != NULL) && (data_name == NULL)) {
				D(("was unable to register the data item [%s]", pam_strerror(pamh, status)));
				free(buffer);
				return NULL;
			}

			if (pamh != NULL) {
				for (i = 0; i < INT_MAX; i++) {
					sprintf(data_name,
						"_pammodutil_getpwnam_%s_%d",
						user, i);
					_pammodutil_lock();
					status = PAM_NO_MODULE_DATA;
					if (pam_get_data
					    (pamh, data_name,
					     &ignore) != PAM_SUCCESS) {
						status =
						    pam_set_data(pamh,
								 data_name,
								 result,
								 pam_modutil_cleanup);
					}
					_pammodutil_unlock();
					if (status == PAM_SUCCESS) {
						break;
					}
				}
			} else {
				status = PAM_SUCCESS;
			}

			free(data_name);

			if (status == PAM_SUCCESS) {
				D(("success"));
				return result;
			}

			D(("was unable to register the data item [%s]",
			   pam_strerror(pamh, status)));

			free(buffer);
			return NULL;

		} else if (errno != ERANGE && errno != EINTR) {
			/* no sense in repeating the call */
			break;
		}

		length <<= 2;

	} while (length < PWD_ABSURD_PWD_LENGTH);

	D(("pwd structure took %u bytes or so of memory",
	   length + sizeof(struct passwd)));

	free(buffer);
	return NULL;

#else				/* ie. ifndef HAVE_GETPWNAM_R */

	/*
	 * If there does not appear to be a reentrant version of
	 * getpwnam(). So, we use the standard libc function.
	 */

	return getpwnam(user);

#endif				/* def HAVE_GETPWNAM_R */
}

struct passwd *pam_modutil_getpwuid(pam_handle_t * pamh, uid_t uid)
{
#ifdef HAVE_GETPWUID_R

	void *buffer = NULL;
	size_t length = PWD_INITIAL_LENGTH;

	do {
		int status;
		void *new_buffer;
		struct passwd *result = NULL;

		new_buffer =
		    realloc(buffer, sizeof(struct passwd) + length);
		if (new_buffer == NULL) {

			D(("out of memory"));

			/* no memory for the user - so delete the memory */
			if (buffer) {
				free(buffer);
			}
			return NULL;
		}
		buffer = new_buffer;

		/* make the re-entrant call to get the pwd structure */
		errno = 0;
		status = getpwuid_r(uid, buffer,
				    sizeof(struct passwd) +
				    (char *) buffer, length, &result);
		if (!status && (result == buffer)) {
			char *data_name;
			const void *ignore;
			int i;

			data_name =
			    malloc(strlen("_pammodutil_getpwuid") + 1 +
				   longlen((long) uid) + 1 +
				   intlen(INT_MAX) + 1);
			if ((pamh != NULL) && (data_name == NULL)) {
				D(("was unable to register the data item [%s]", pam_strerror(pamh, status)));
				free(buffer);
				return NULL;
			}

			if (pamh != NULL) {
				for (i = 0; i < INT_MAX; i++) {
					sprintf(data_name,
						"_pammodutil_getpwuid_%ld_%d",
						(long) uid, i);
					_pammodutil_lock();
					status = PAM_NO_MODULE_DATA;
					if (pam_get_data
					    (pamh, data_name,
					     &ignore) != PAM_SUCCESS) {
						status =
						    pam_set_data(pamh,
								 data_name,
								 result,
								 pam_modutil_cleanup);
					}
					_pammodutil_unlock();
					if (status == PAM_SUCCESS) {
						break;
					}
				}
			} else {
				status = PAM_SUCCESS;
			}

			free(data_name);

			if (status == PAM_SUCCESS) {
				D(("success"));
				return result;
			}

			D(("was unable to register the data item [%s]",
			   pam_strerror(pamh, status)));

			free(buffer);
			return NULL;

		} else if (errno != ERANGE && errno != EINTR) {
			/* no sense in repeating the call */
			break;
		}

		length <<= 2;

	} while (length < PWD_ABSURD_PWD_LENGTH);

	D(("pwd structure took %u bytes or so of memory",
	   length + sizeof(struct passwd)));

	free(buffer);
	return NULL;

#else				/* ie. ifndef HAVE_GETPWUID_R */

	/*
	 * If there does not appear to be a reentrant version of
	 * getpwuid(). So, we use the standard libc function.
	 */

	return getpwuid(uid);

#endif				/* def HAVE_GETPWUID_R */
}

struct group *pam_modutil_getgrgid(pam_handle_t * pamh, gid_t gid)
{
#ifdef HAVE_GETGRGID_R

	void *buffer = NULL;
	size_t length = PWD_INITIAL_LENGTH;

	do {
		int status;
		void *new_buffer;
		struct group *result = NULL;

		new_buffer =
		    realloc(buffer, sizeof(struct group) + length);
		if (new_buffer == NULL) {

			D(("out of memory"));

			/* no memory for the user - so delete the memory */
			if (buffer) {
				free(buffer);
			}
			return NULL;
		}
		buffer = new_buffer;

		/* make the re-entrant call to get the grp structure */
		errno = 0;
		status = getgrgid_r(gid, buffer,
				    sizeof(struct group) + (char *) buffer,
				    length, &result);
		if (!status && (result == buffer)) {
			char *data_name;
			const void *ignore;
			int i;

			data_name =
			    malloc(strlen("_pammodutil_getgrgid") + 1 +
				   longlen((long) gid) + 1 +
				   intlen(INT_MAX) + 1);
			if ((pamh != NULL) && (data_name == NULL)) {
				D(("was unable to register the data item [%s]", pam_strerror(pamh, status)));
				free(buffer);
				return NULL;
			}

			if (pamh != NULL) {
				for (i = 0; i < INT_MAX; i++) {
					sprintf(data_name,
						"_pammodutil_getgrgid_%ld_%d",
						(long) gid, i);
					_pammodutil_lock();
					status = PAM_NO_MODULE_DATA;
					if (pam_get_data
					    (pamh, data_name,
					     &ignore) != PAM_SUCCESS) {
						status =
						    pam_set_data(pamh,
								 data_name,
								 result,
								 pam_modutil_cleanup);
					}
					_pammodutil_unlock();
					if (status == PAM_SUCCESS) {
						break;
					}
				}
			} else {
				status = PAM_SUCCESS;
			}

			free(data_name);

			if (status == PAM_SUCCESS) {
				D(("success"));
				return result;
			}

			D(("was unable to register the data item [%s]",
			   pam_strerror(pamh, status)));

			free(buffer);
			return NULL;

		} else if (errno != ERANGE && errno != EINTR) {
			/* no sense in repeating the call */
			break;
		}

		length <<= 2;

	} while (length < PWD_ABSURD_PWD_LENGTH);

	D(("grp structure took %u bytes or so of memory",
	   length + sizeof(struct group)));

	free(buffer);
	return NULL;

#else				/* ie. ifndef HAVE_GETGRGID_R */

	/*
	 * If there does not appear to be a reentrant version of
	 * getgrgid(). So, we use the standard libc function.
	 */

	return getgrgid(gid);

#endif				/* def HAVE_GETGRGID_R */
}

struct group *pam_modutil_getgrnam(pam_handle_t * pamh, const char *group)
{
#ifdef HAVE_GETGRNAM_R

	void *buffer = NULL;
	size_t length = PWD_INITIAL_LENGTH;

	do {
		int status;
		void *new_buffer;
		struct group *result = NULL;

		new_buffer =
		    realloc(buffer, sizeof(struct group) + length);
		if (new_buffer == NULL) {

			D(("out of memory"));

			/* no memory for the group - so delete the memory */
			if (buffer) {
				free(buffer);
			}
			return NULL;
		}
		buffer = new_buffer;

		/* make the re-entrant call to get the grp structure */
		errno = 0;
		status = getgrnam_r(group, buffer,
				    sizeof(struct group) + (char *) buffer,
				    length, &result);
		if (!status && (result == buffer)) {
			char *data_name;
			const void *ignore;
			int i;

			data_name =
			    malloc(strlen("_pammodutil_getgrnam") + 1 +
				   strlen(group) + 1 + intlen(INT_MAX) +
				   1);
			if ((pamh != NULL) && (data_name == NULL)) {
				D(("was unable to register the data item [%s]", pam_strerror(pamh, status)));
				free(buffer);
				return NULL;
			}

			if (pamh != NULL) {
				for (i = 0; i < INT_MAX; i++) {
					sprintf(data_name,
						"_pammodutil_getgrnam_%s_%d",
						group, i);
					_pammodutil_lock();
					status = PAM_NO_MODULE_DATA;
					if (pam_get_data
					    (pamh, data_name,
					     &ignore) != PAM_SUCCESS) {
						status =
						    pam_set_data(pamh,
								 data_name,
								 result,
								 pam_modutil_cleanup);
					}
					_pammodutil_unlock();
					if (status == PAM_SUCCESS) {
						break;
					}
				}
			} else {
				status = PAM_SUCCESS;
			}

			free(data_name);

			if (status == PAM_SUCCESS) {
				D(("success"));
				return result;
			}

			D(("was unable to register the data item [%s]",
			   pam_strerror(pamh, status)));

			free(buffer);
			return NULL;

		} else if (errno != ERANGE && errno != EINTR) {
			/* no sense in repeating the call */
			break;
		}

		length <<= 2;

	} while (length < PWD_ABSURD_PWD_LENGTH);

	D(("grp structure took %u bytes or so of memory",
	   length + sizeof(struct group)));

	free(buffer);
	return NULL;

#else				/* ie. ifndef HAVE_GETGRNAM_R */

	/*
	 * If there does not appear to be a reentrant version of
	 * getgrnam(). So, we use the standard libc function.
	 */

	return getgrnam(group);

#endif				/* def HAVE_GETGRNAM_R */
}


void pam_modutil_cleanup(pam_handle_t * pamh, void *data, int error_status)
{
	if (data) {
		/* junk it */
		(void) free(data);
	}
}

#ifdef HAVE_GETGROUPLIST
static int checkgrouplist(const char *user, gid_t primary, gid_t target)
{
	gid_t *grouplist = NULL;
	int agroups, ngroups, i;
	ngroups = agroups = 3;
	do {
		grouplist = malloc(sizeof(gid_t) * agroups);
		if (grouplist == NULL) {
			return 0;
		}
		ngroups = agroups;
		i = getgrouplist(user, primary, grouplist, &ngroups);
		if ((i < 0) || (ngroups < 1)) {
			agroups *= 2;
			free(grouplist);
		} else {
			for (i = 0; i < ngroups; i++) {
				if (grouplist[i] == target) {
					free(grouplist);
					return 1;
				}
			}
			free(grouplist);
		}
	} while (((i < 0) || (ngroups < 1)) && (agroups < 10000));
	return 0;
}
#endif

static int
pam_modutil_user_in_group_common(pam_handle_t * pamh,
				 struct passwd *pwd, struct group *grp)
{
	int i;

	if (pwd == NULL) {
		return 0;
	}
	if (grp == NULL) {
		return 0;
	}

	if (pwd->pw_gid == grp->gr_gid) {
		return 1;
	}

	for (i = 0; (grp->gr_mem != NULL) && (grp->gr_mem[i] != NULL); i++) {
		if (strcmp(pwd->pw_name, grp->gr_mem[i]) == 0) {
			return 1;
		}
	}

#ifdef HAVE_GETGROUPLIST
	if (checkgrouplist(pwd->pw_name, pwd->pw_gid, grp->gr_gid)) {
		return 1;
	}
#endif

	return 0;
}

int
pam_modutil_user_in_group_nam_nam(pam_handle_t * pamh,
				  const char *user, const char *group)
{
	struct passwd *pwd;
	struct group *grp;

	pwd = pam_modutil_getpwnam(pamh, user);
	grp = pam_modutil_getgrnam(pamh, group);

	return pam_modutil_user_in_group_common(pamh, pwd, grp);
}


#endif
