/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file xos_plymgt.c
 * @brief poly management. Currently, it is built only for resouce quota 
 *       enforcement.
 * 
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <limits.h>

#include <utmp.h>
#ifndef UT_USER
#define UT_USER ut_user
#endif

#include "xos_utils.h"
#include "xos_readconf.h"
#include "pam_plymgt.h"

#define PAM_DEBUG_ARG       0x0001
#define PAM_DO_SETREUID     0x0002
#define PAM_UTMP_EARLY      0x0004



/** @brief Define type of seting limitation
 */
static const char *limits_def_names[] = {
	"USER",
	"GROUP",
	"ALLGROUP",
	"ALL",
	"DEFAULT",
	"NONE",
	NULL
};



/** Parse the configure parameters 
 * @param pam_handle_t handler of PAM module.
 * @param conf configure file
 * @param pl limit paramters defined in configure file.
 * @return PAM_SUCCESS if success, or PAM_SYS_ERR.
 */

int xospm_paraparse(const pam_handle_t * pamh, char *conf, PAM_LIMIT * pl)
{

	if (conf)
		strcpy(pl->conf_file, conf);
	else
		strcpy(pl->conf_file, VO_QUOTA_CONF);

	return PAM_SUCCESS;
}


/** Check login number of account
 */

int xospm_checklogin(pam_handle_t * pamh, const char *name, int limit,
		     int ctrl, PAM_LIMIT * pl)
{
	struct utmp *ut;
	int count;

	if (ctrl & USE_DEBUG) {
		DEBUG("PAM:checking logins for '%s' (maximum of %d)", name,
		      limit);
	}

	if (limit < 0)
		return 0;	/* no limits imposed */
	if (limit == 0) {	/* maximum 0 logins ? */
		DEBUG("PAM:No logins allowed for '%s'", name);
		return LOGIN_ERR;
	}

	setutent();


	count = 0;

	while ((ut = getutent())) {
#ifdef USER_PROCESS
		if (ut->ut_type != USER_PROCESS)
			continue;
#endif
		if (ut->UT_USER[0] == '\0')
			continue;

		if (!pl->flag_numsyslogins) {
			if (((pl->login_limit_def == LIMITS_DEF_USER)
			     || (pl->login_limit_def == LIMITS_DEF_GROUP)
			     || (pl->login_limit_def ==
				 LIMITS_DEF_DEFAULT))
			    && strncmp(name, ut->UT_USER,
				       sizeof(ut->UT_USER)) != 0)
				continue;

			if ((pl->login_limit_def == LIMITS_DEF_ALLGROUP)
			    && !pam_modutil_user_in_group_nam_nam(pamh,
								  ut->
								  UT_USER,
								  pl->
								  login_group))
				continue;

		}
		if (++count > limit)
			break;

	}
	endutent();

	if (count > limit) {
		if (name) {
			DEBUG("PAM:Too many logins (max %d) for %s", limit,
			      name);
		} else {
			DEBUG("PAM:Too many system logins (max %d)",
			      limit);
		}
		return LOGIN_ERR;
	}
	return 0;
}


/** Initiate the policy manager
 * @param pl limit paramters defined in configure file.
 * @return PAM_SUCCESS if success, or PAM_SYS_ERR.
 */
int xospm_plmgtinit(PAM_LIMIT * pl)
{
	int i;
	int retval = PAM_SUCCESS;


	for (i = 0; i < RLIM_NLIMITS; i++) {
		int r = getrlimit(i, &pl->limits[i].limit);
		if (r == -1) {
			pl->limits[i].supported = 0;
			if (errno != EINVAL)
				retval = !PAM_SUCCESS;

		} else {
			pl->limits[i].supported = 1;
			pl->limits[i].src_soft = LIMITS_DEF_NONE;
			pl->limits[i].src_hard = LIMITS_DEF_NONE;
		}
	}

	errno = 0;
	pl->priority = getpriority(PRIO_PROCESS, 0);
	if (pl->priority == -1 && errno != 0)
		retval = !PAM_SUCCESS;
	pl->login_limit = -2;
	pl->login_limit_def = LIMITS_DEF_NONE;

	return retval;
}

/** Limit resouce quota by setrlimit()
 * 
 */
void xospm_quotalimit(const pam_handle_t * pamh, int source,
		      const char *lim_type, const char *lim_item,
		      const char *lim_value, int ctrl, PAM_LIMIT * pl)
{
	int limit_item;
	int limit_type = 0;
	int int_value = 0;
	rlim_t rlimit_value = 0;
	char *endptr;
	const char *value_orig = lim_value;

	if (ctrl & USE_DEBUG)
		DEBUG("PAM:%s: processing %s %s %s for %s",
		      __FUNCTION__, lim_type, lim_item, lim_value,
		      limits_def_names[source]);

	if (strcmp(lim_item, "cpu") == 0)
		limit_item = RLIMIT_CPU;
	else if (strcmp(lim_item, "fsize") == 0)
		limit_item = RLIMIT_FSIZE;
	else if (strcmp(lim_item, "data") == 0)
		limit_item = RLIMIT_DATA;
	else if (strcmp(lim_item, "stack") == 0)
		limit_item = RLIMIT_STACK;
	else if (strcmp(lim_item, "core") == 0)
		limit_item = RLIMIT_CORE;
	else if (strcmp(lim_item, "rss") == 0)
		limit_item = RLIMIT_RSS;
	else if (strcmp(lim_item, "nproc") == 0)
		limit_item = RLIMIT_NPROC;
	else if (strcmp(lim_item, "nofile") == 0)
		limit_item = RLIMIT_NOFILE;
	else if (strcmp(lim_item, "memlock") == 0)
		limit_item = RLIMIT_MEMLOCK;
	else if (strcmp(lim_item, "mem") == 0)
		limit_item = RLIMIT_AS;
#ifdef RLIMIT_LOCKS
	else if (strcmp(lim_item, "locks") == 0)
		limit_item = RLIMIT_LOCKS;
#endif
#ifdef RLIMIT_SIGPENDING
	else if (strcmp(lim_item, "sigpending") == 0)
		limit_item = RLIMIT_SIGPENDING;
#endif
#ifdef RLIMIT_MSGQUEUE
	else if (strcmp(lim_item, "msgqueue") == 0)
		limit_item = RLIMIT_MSGQUEUE;
#endif
#ifdef RLIMIT_NICE
	else if (strcmp(lim_item, "nice") == 0)
		limit_item = RLIMIT_NICE;
#endif
#ifdef RLIMIT_RTPRIO
	else if (strcmp(lim_item, "rtprio") == 0)
		limit_item = RLIMIT_RTPRIO;
#endif
	else if (strcmp(lim_item, "maxlogins") == 0) {
		limit_item = LIMIT_LOGIN;
		pl->flag_numsyslogins = 0;
	} else if (strcmp(lim_item, "maxsyslogins") == 0) {
		limit_item = LIMIT_NUMSYSLOGINS;
		pl->flag_numsyslogins = 1;
	} else if (strcmp(lim_item, "priority") == 0) {
		limit_item = LIMIT_PRI;
	} else if (strcmp(lim_item, "disk") == 0) {
		DEBUG("PAM:No support limit item '%s' currently",
		      lim_item);
		return;
	} else {
		DEBUG("PAM:unknown limit item '%s'", lim_item);
		return;
	}

	if (strcmp(lim_type, "soft") == 0)
		limit_type = LIMIT_SOFT;
	else if (strcmp(lim_type, "hard") == 0)
		limit_type = LIMIT_HARD;
	else if (strcmp(lim_type, "-") == 0)
		limit_type = LIMIT_SOFT | LIMIT_HARD;
	else if (limit_item != LIMIT_LOGIN
		 && limit_item != LIMIT_NUMSYSLOGINS) {
		DEBUG("PAM:unknown limit type '%s'", lim_type);
		return;
	}
	if (limit_item != LIMIT_PRI
#ifdef RLIMIT_NICE
	    && limit_item != RLIMIT_NICE
#endif
	    && (strcmp(lim_value, "-1") == 0
		|| strcmp(lim_value, "-") == 0
		|| strcmp(lim_value, "unlimited") == 0
		|| strcmp(lim_value, "infinity") == 0)) {
		int_value = -1;
		rlimit_value = RLIM_INFINITY;
	} else if (limit_item == LIMIT_PRI || limit_item == LIMIT_LOGIN ||
#ifdef RLIMIT_NICE
		   limit_item == RLIMIT_NICE ||
#endif
		   limit_item == LIMIT_NUMSYSLOGINS) {
		long temp;
		temp = strtol(lim_value, &endptr, 10);
		temp = temp < INT_MAX ? temp : INT_MAX;
		int_value = temp > INT_MIN ? temp : INT_MIN;
		if (int_value == 0 && value_orig == endptr) {
			DEBUG
			    ("PAM:wrong limit value '%s' for limit type '%s'\n",
			     lim_value, lim_type);
			return;
		}
	} else {
#ifdef __USE_FILE_OFFSET64
		rlimit_value = strtoull(lim_value, &endptr, 10);
#else
		rlimit_value = strtoul(lim_value, &endptr, 10);
#endif
		if (rlimit_value == 0 && value_orig == endptr) {
			DEBUG
			    ("PAM:wrong limit value '%s' for limit type '%s'\n",
			     lim_value, lim_type);
			return;
		}
	}

	/* one more special case when limiting logins */
	if ((source == LIMITS_DEF_ALL || source == LIMITS_DEF_ALLGROUP)
	    && (limit_item != LIMIT_LOGIN)) {
		if (ctrl & USE_DEBUG)
			DEBUG
			    ("PAM:'%%' domain valid for maxlogins type only\n");
		return;
	}

	switch (limit_item) {
	case RLIMIT_CPU:
		if (rlimit_value != RLIM_INFINITY)
			rlimit_value *= 1;
		break;
	case RLIMIT_FSIZE:
	case RLIMIT_DATA:
	case RLIMIT_STACK:
	case RLIMIT_CORE:
	case RLIMIT_RSS:
	case RLIMIT_MEMLOCK:
	case RLIMIT_AS:
		if (rlimit_value != RLIM_INFINITY)
			rlimit_value *= 1024;
		break;
#ifdef RLIMIT_NICE
	case RLIMIT_NICE:
		if (int_value > 19)
			int_value = 19;
		rlimit_value = 19 - int_value;
#endif
		break;
	}

	if ((limit_item != LIMIT_LOGIN)
	    && (limit_item != LIMIT_NUMSYSLOGINS)
	    && (limit_item != LIMIT_PRI)) {
		if (limit_type & LIMIT_SOFT) {
			if (pl->limits[limit_item].src_soft < source)
				return;
			else {
				pl->limits[limit_item].limit.rlim_cur =
				    rlimit_value;
				pl->limits[limit_item].src_soft = source;
			}
		}
		if (limit_type & LIMIT_HARD) {
			if (pl->limits[limit_item].src_hard < source)
				return;
			else {
				pl->limits[limit_item].limit.rlim_max =
				    rlimit_value;
				pl->limits[limit_item].src_hard = source;
			}
		}
	} else {
		/* recent kernels support negative priority limits (=raise priority) */

		if (limit_item == LIMIT_PRI)
			pl->priority = int_value;
		else {
			if (pl->login_limit_def < source)
				return;
			else {
				pl->login_limit = int_value;
				pl->login_limit_def = source;
			}
		}
	}
	return;


}


/** get limitation parameters from configure file
 * @param pamh handler of PAM module
 * @param uname user to be limited
 * @param ctrl control code
 * @param pl limit paramters defined in configure file.
 * @return PAM_SUCCESS if success, or PAM_SYS_ERR.
 */
int xospm_getconfigure(pam_handle_t * pamh, const char *uname, int ctrl,
		       PAM_LIMIT * pl)
{
	struct stat file_stat;
	FILE *fp;
	char buf[LINE_LENGTH];
	int ret = PAM_SUCCESS;
	char *tptr;

	if (pl->conf_file[0] == '\0')	/* originally, NULL */
		strncpy(pl->conf_file, VO_QUOTA_CONF,
			sizeof(pl->conf_file) - 1);

	if (stat(pl->conf_file, &file_stat) != 0)
		return ret;

	fp = fopen(pl->conf_file, "r");
	if (fp == NULL) {
		DEBUG("PAM:cannot read settings from %s: %m",
		      VO_QUOTA_CONF);
		return PAM_SERVICE_ERR;
	}


	/* init things */
	memset(buf, 0, sizeof(buf));
	/* start the show */
	while (fgets(buf, LINE_LENGTH, fp) != NULL) {


		tptr = buf;
		/* skip the leading white space */
		while (*tptr && isspace(*tptr))
			tptr++;
		strncpy(buf, tptr, sizeof(buf) - 1);
		buf[sizeof(buf) - 1] = '\0';

		/* Rip off the comments */
		tptr = strchr(buf, '#');
		if (tptr)
			*tptr = '\0';

		/* Rip off the newline char */
		tptr = strchr(buf, '\n');
		if (tptr)
			*tptr = '\0';

		/* Anything left ? */
		if (!strlen(buf)) {
			memset(buf, 0, sizeof(buf));
			continue;
		}
		ret = xospam_parse_config(pamh, uname, buf, ctrl, pl);
	}
	fclose(fp);
	return ret;
}

int xospm_getconfigure_frombuf(pam_handle_t * pamh, const char *uname,
			       char *config_buf, int ctrl, PAM_LIMIT * pl)
{
	char *p = NULL, *pp = NULL, *tp = NULL;
	char buf[80];
	int i, flag, ret;

	if (!config_buf || !uname || !pl)
		return PAM_SERVICE_ERR;;

	memset(buf, 0, sizeof(buf));
	ret = PAM_SUCCESS;

	for (pp = p = config_buf; *p; p++) {
		flag = 0;

		if (*pp == '\n') {
			pp++;
			continue;
		}

		if (*p == '\n' && p != pp) {
			tp = pp;
			i = 0;
			while (tp != p) {
				*(buf + i) = *tp;
				tp++;
				i++;
			}
			*(buf + i) = '\0';
			flag = 1;
			pp = p + 1;
		}

		if (!flag)
			continue;

		ret = xospam_parse_config(pamh, uname, buf, ctrl, pl);
	}

	return ret;
}


/** setup the quota limitation by setrlimit
 * @param pamh handler of PAM module
 * @param uname user to be limited
 * @param uid uid of current user
 * @param ctrl control code
 * @param pl limit paramters defined in configure file.
 * @return PAM_SUCCESS if success, or PAM_SYS_ERR.
 */
int xospm_plysetup(pam_handle_t * pamh, const char *uname, uid_t uid,
		   int ctrl, PAM_LIMIT * pl)
{
	int i;
	int status;
	int retval = LIMITED_OK;

	for (i = 0, status = LIMITED_OK; i < RLIM_NLIMITS; i++) {
		if (!pl->limits[i].supported)
			/* skip it if its not known to the system */
			continue;

		if (pl->limits[i].src_soft == LIMITS_DEF_NONE &&
		    pl->limits[i].src_hard == LIMITS_DEF_NONE)
			/* skip it if its not initialized */
			continue;

		if (pl->limits[i].limit.rlim_cur >
		    pl->limits[i].limit.rlim_max)
			pl->limits[i].limit.rlim_cur =
			    pl->limits[i].limit.rlim_max;
		status |= setrlimit(i, &pl->limits[i].limit);
	}

	if (status)
		retval = LIMIT_ERR;


	status = setpriority(PRIO_PROCESS, 0, pl->priority);
	if (status != 0)
		retval = LIMIT_ERR;


	if (uid == 0) {
		if (ctrl & USE_DEBUG)
			DEBUG("PAM:skip login limit check for uid=0");
	} else if (pl->login_limit > 0) {
		if (xospm_checklogin
		    (pamh, uname, pl->login_limit, ctrl, pl) == LOGIN_ERR)
			retval |= LOGIN_ERR;

	} else if (pl->login_limit == 0)
		retval |= LOGIN_ERR;


	return retval;
}


int xospam_parse_config(pam_handle_t * pamh, const char *uname, char *buf,
			int ctrl, PAM_LIMIT * pl)
{
	char domain[LINE_LENGTH];
	char ltype[LINE_LENGTH];
	char item[LINE_LENGTH];
	char value[LINE_LENGTH];
	int i;
	size_t j;

	if (!uname || !buf || !pl)
		return PAM_SERVICE_ERR;

//      char *tptr;
//
//      tptr = buf;
//      /* skip the leading white space */
//      while (*tptr && isspace(*tptr))
//              tptr++;
//      strncpy(buf, tptr, sizeof(buf) - 1);
//      buf[sizeof(buf) - 1] = '\0';
//
//      /* Rip off the comments */
//      tptr = strchr(buf, '#');
//      if (tptr)
//              *tptr = '\0';
//
//      /* Rip off the newline char */
//      tptr = strchr(buf, '\n');
//      if (tptr)
//              *tptr = '\0';
//
//      /* Anything left ? */
//      if (!strlen(buf)) {
//              memset(buf, 0, sizeof(buf));
//              continue;
//      }

	memset(domain, 0, sizeof(domain));
	memset(ltype, 0, sizeof(ltype));
	memset(item, 0, sizeof(item));
	memset(value, 0, sizeof(value));

	i = sscanf(buf, "%s%s%s%s", domain, ltype, item, value);


	for (j = 0; j < strlen(ltype); j++)
		ltype[j] = tolower(ltype[j]);
	for (j = 0; j < strlen(item); j++)
		item[j] = tolower(item[j]);
	for (j = 0; j < strlen(value); j++)
		value[j] = tolower(value[j]);

	if (i == 4) {		/* a complete line */
		if (strcmp(uname, domain) == 0)	/* this user have a limit */
			xospm_quotalimit(pamh, LIMITS_DEF_USER,
					 ltype, item, value, ctrl, pl);
		else if (domain[0] == '@') {
			if (ctrl & USE_DEBUG) {
				DEBUG
				    ("PAM:checking if %s is in group %s\n",
				     uname, domain + 1);
			}
			if (pam_modutil_user_in_group_nam_nam
			    (pamh, uname, domain + 1))
				xospm_quotalimit(pamh,
						 LIMITS_DEF_GROUP,
						 ltype, item,
						 value, ctrl, pl);
		} else if (domain[0] == '%') {
			if (ctrl & USE_DEBUG) {
				DEBUG
				    ("PAM:checking if %s is in group %s\n",
				     uname, domain + 1);
			}
			if (strcmp(domain, "%") == 0)
				xospm_quotalimit(pamh,
						 LIMITS_DEF_ALL,
						 ltype, item,
						 value, ctrl, pl);
			else if (pam_modutil_user_in_group_nam_nam
				 (pamh, uname, domain + 1)) {
				strcpy(pl->login_group, domain + 1);
				xospm_quotalimit(pamh,
						 LIMITS_DEF_ALLGROUP,
						 ltype, item,
						 value, ctrl, pl);
			}
		} else if (strcmp(domain, "*") == 0)
			xospm_quotalimit(pamh, LIMITS_DEF_DEFAULT,
					 ltype, item, value, ctrl, pl);
	} else if (i == 2 && ltype[0] == '-') {
		/* Probably a no-limit line */
		if (strcmp(uname, domain) == 0) {
			if (ctrl & USE_DEBUG)
				DEBUG("PAM:no limits for '%s'", uname);

			return PAM_IGNORE;
		} else if (domain[0] == '@'
			   &&
			   pam_modutil_user_in_group_nam_nam(pamh,
							     uname,
							     domain + 1)) {
			if (ctrl & USE_DEBUG) {
				DEBUG
				    ("PAM:no limits for '%s' in group '%s'\n",
				     uname, domain + 1);
			}
			return PAM_IGNORE;
		}
	} else {
		DEBUG("PAM:invalid line '%s' - skipped", buf);
	}

	return PAM_SUCCESS;;
}
