/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/** @file xos_reslimit.c
 * @brief poly management. Currently, it is built only for resouce quota 
 *       enforcement.
 * 
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <limits.h>

#include <utmp.h>
#ifndef UT_USER
#define UT_USER ut_user
#endif

#include "xos_utils.h"
#include "xos_reslimit.h"
#include "xos_dbaux.h"


/** @brief Define type of seting limitation
 */
static const char *default_limitname[] = {
	"USER",
	"GROUP",
	"ALLGROUP",
	"ALL",
	"DEFAULT",
	"NONE",
	NULL
};




int xos_reslimit_ismember(char *user, char *group)
{
	GPASSWD *gpwd = gpwd_new();
	GGROUPS *ggrp = ggrp_new();
	char uid_str[LINE_LENGTH];
	int found = 0, len = 0;
	FILE *fp;
	char buf[LINE_LENGTH], *p = NULL, *pp = NULL;

	if (!user || !group)
		return -1;

	if (bdb_getdbuname_pwd(user, gpwd)
	    && bdb_getdbgname_grp(group, ggrp)) {
		sprintf(uid_str, "%d", gpwd->l_idtoken.g_mappeduid);
		len = strlen(uid_str);
		if ((p = strstr(ggrp->l_grptoken.g_members, uid_str))) {
			pp = ggrp->l_grptoken.g_members;
			if ((pp == p || *(p - 1) == ',')
			    && (*(p + len) == ',' || *(p + len) == '\0'))
				found = 1;
		}

	}

	if (!found) {
		fp = fopen("/etc/group", "r");
		if (fp == NULL) {
			DEBUG("Can not open the /etc/group\n");
			return -1;
		}

		memset(buf, 0, sizeof(buf));
		while (fgets(buf, LINE_LENGTH, fp) != NULL) {
			if (!strncmp(buf, group, strlen(group))) {
				p = strstr(buf, user);
				if (p == NULL)
					continue;
				len = strlen(user);
				pp = buf;
				if ((*(p - 1) == ':' || *(p - 1) == ',')
				    && (*(p + len) == ','
					|| *(p + len) == '\0'
					|| *(p + len) == '\n')) {
					found = 1;
					break;
				}
			}
		}
		fclose(fp);
	}

	if (found)
		return 0;
	else
		return -1;
}


/** Check login number of account
 */

int xos_reslimit_checklogin(const char *name, int limit, XLIMIT * pl)
{
	struct utmp *ut;
	int count;


	if (limit < 0)
		return 0;	/* no limits imposed */
	if (limit == 0) {	/* maximum 0 logins ? */
		DEBUG("No logins allowed for '%s'", name);
		return LOGIN_ERR;
	}

	setutent();


	count = 0;

	while ((ut = getutent())) {
#ifdef USER_PROCESS
		if (ut->ut_type != USER_PROCESS)
			continue;
#endif
		if (ut->UT_USER[0] == '\0')
			continue;

		if (!pl->flag_numsyslogins) {
			if (((pl->login_limit_def == LIMITS_DEF_USER)
			     || (pl->login_limit_def == LIMITS_DEF_GROUP)
			     || (pl->login_limit_def ==
				 LIMITS_DEF_DEFAULT))
			    && strncmp(name, ut->UT_USER,
				       sizeof(ut->UT_USER)) != 0)
				continue;

			if ((pl->login_limit_def == LIMITS_DEF_ALLGROUP)
			    && !xos_reslimit_ismember(ut->UT_USER,
						      pl->login_group))
				continue;

		}
		if (++count > limit)
			break;

	}
	endutent();

	if (count > limit) {
		if (name) {
			DEBUG("Too many logins (max %d) for %s", limit,
			      name);
		} else {
			DEBUG("Too many system logins (max %d)", limit);
		}
		return LOGIN_ERR;
	}
	return 0;
}


/** Initiate the policy manager
 * @param pl limit paramters defined in configure file.
 * @return PAM_SUCCESS if success, or PAM_SYS_ERR.
 */
int xos_reslimit_init(XLIMIT * pl)
{
	int i;
	int retval = 0;

	if (!pl)
		return -1;

	memset(pl, 0, sizeof(*pl));

	for (i = 0; i < RLIM_NLIMITS; i++) {
		int r = getrlimit(i, &pl->limits[i].limit);
		if (r == -1) {
			pl->limits[i].supported = 0;
			if (errno != EINVAL)
				retval = -1;

		} else {
			pl->limits[i].supported = 1;
			pl->limits[i].src_soft = LIMITS_DEF_NONE;
			pl->limits[i].src_hard = LIMITS_DEF_NONE;
		}
	}

	errno = 0;
	pl->priority = getpriority(PRIO_PROCESS, 0);
	if (pl->priority == -1 && errno != 0)
		retval = -1;
	pl->login_limit = -2;
	pl->login_limit_def = LIMITS_DEF_NONE;

	return retval;
}

/** Limit resouce quota by setrlimit()
 * 
 */
void xos_reslimit_quota(int source,
			const char *lim_type, const char *lim_item,
			const char *lim_value, XLIMIT * pl)
{
	int limit_item;
	int limit_type = 0;
	int int_value = 0;
	rlim_t rlimit_value = 0;
	char *endptr;
	const char *value_orig = lim_value;

	if (source < 0 || source > 6 || !lim_type || !lim_item
	    || !lim_value || !pl)
		return;

	DEBUG("%s: processing %s %s %s for %s",
	      __FUNCTION__, lim_type, lim_item, lim_value,
	      default_limitname[source]);

	if (strcmp(lim_item, "cpu") == 0)
		limit_item = RLIMIT_CPU;
	else if (strcmp(lim_item, "fsize") == 0)
		limit_item = RLIMIT_FSIZE;
	else if (strcmp(lim_item, "data") == 0)
		limit_item = RLIMIT_DATA;
	else if (strcmp(lim_item, "stack") == 0)
		limit_item = RLIMIT_STACK;
	else if (strcmp(lim_item, "core") == 0)
		limit_item = RLIMIT_CORE;
	else if (strcmp(lim_item, "mem") == 0)
		limit_item = RLIMIT_RSS;
	else if (strcmp(lim_item, "nproc") == 0)
		limit_item = RLIMIT_NPROC;
	else if (strcmp(lim_item, "nofile") == 0)
		limit_item = RLIMIT_NOFILE;
	else if (strcmp(lim_item, "memlock") == 0)
		limit_item = RLIMIT_MEMLOCK;
	else if (strcmp(lim_item, "as") == 0)
		limit_item = RLIMIT_AS;
#ifdef RLIMIT_LOCKS
	else if (strcmp(lim_item, "locks") == 0)
		limit_item = RLIMIT_LOCKS;
#endif
#ifdef RLIMIT_SIGPENDING
	else if (strcmp(lim_item, "sigpending") == 0)
		limit_item = RLIMIT_SIGPENDING;
#endif
#ifdef RLIMIT_MSGQUEUE
	else if (strcmp(lim_item, "msgqueue") == 0)
		limit_item = RLIMIT_MSGQUEUE;
#endif
#ifdef RLIMIT_NICE
	else if (strcmp(lim_item, "nice") == 0)
		limit_item = RLIMIT_NICE;
#endif
#ifdef RLIMIT_RTPRIO
	else if (strcmp(lim_item, "rtprio") == 0)
		limit_item = RLIMIT_RTPRIO;
#endif
	else if (strcmp(lim_item, "maxlogins") == 0) {
		limit_item = LIMIT_LOGIN;
		pl->flag_numsyslogins = 0;
	} else if (strcmp(lim_item, "maxsyslogins") == 0) {
		limit_item = LIMIT_NUMSYSLOGINS;
		pl->flag_numsyslogins = 1;
	} else if (strcmp(lim_item, "priority") == 0) {
		limit_item = LIMIT_PRI;
	} else if (strcmp(lim_item, "disk") == 0) {
		DEBUG("No support limit item '%s' currently", lim_item);
		return;
	} else {
		DEBUG("unknown limit item '%s'", lim_item);
		return;
	}

	if (strcmp(lim_type, "soft") == 0)
		limit_type = LIMIT_SOFT;
	else if (strcmp(lim_type, "hard") == 0)
		limit_type = LIMIT_HARD;
	else if (strcmp(lim_type, "-") == 0)
		limit_type = LIMIT_SOFT | LIMIT_HARD;
	else if (limit_item != LIMIT_LOGIN
		 && limit_item != LIMIT_NUMSYSLOGINS) {
		DEBUG("unknown limit type '%s'", lim_type);
		return;
	}
	if (limit_item != LIMIT_PRI
#ifdef RLIMIT_NICE
	    && limit_item != RLIMIT_NICE
#endif
	    && (strcmp(lim_value, "-1") == 0
		|| strcmp(lim_value, "-") == 0
		|| strcmp(lim_value, "unlimited") == 0
		|| strcmp(lim_value, "infinity") == 0)) {
		int_value = -1;
		rlimit_value = RLIM_INFINITY;
	} else if (limit_item == LIMIT_PRI || limit_item == LIMIT_LOGIN ||
#ifdef RLIMIT_NICE
		   limit_item == RLIMIT_NICE ||
#endif
		   limit_item == LIMIT_NUMSYSLOGINS) {
		long temp;
		temp = strtol(lim_value, &endptr, 10);
		temp = temp < INT_MAX ? temp : INT_MAX;
		int_value = temp > INT_MIN ? temp : INT_MIN;
		if (int_value == 0 && value_orig == endptr) {
			DEBUG
			    ("PAM:wrong limit value '%s' for limit type '%s'\n",
			     lim_value, lim_type);
			return;
		}
	} else {
#ifdef __USE_FILE_OFFSET64
		rlimit_value = strtoull(lim_value, &endptr, 10);
#else
		rlimit_value = strtoul(lim_value, &endptr, 10);
#endif
		if (rlimit_value == 0 && value_orig == endptr) {
			DEBUG
			    ("PAM:wrong limit value '%s' for limit type '%s'\n",
			     lim_value, lim_type);
			return;
		}
	}

	/* one more special case when limiting logins */
	if ((source == LIMITS_DEF_ALL || source == LIMITS_DEF_ALLGROUP)
	    && (limit_item != LIMIT_LOGIN)) {
		DEBUG("PAM:'%%' domain valid for maxlogins type only\n");
		return;
	}

	switch (limit_item) {
	case RLIMIT_CPU:
		if (rlimit_value != RLIM_INFINITY)
			rlimit_value *= 60;
		break;
	case RLIMIT_FSIZE:
	case RLIMIT_DATA:
	case RLIMIT_STACK:
	case RLIMIT_CORE:
	case RLIMIT_RSS:
	case RLIMIT_MEMLOCK:
	case RLIMIT_AS:
		if (rlimit_value != RLIM_INFINITY)
			rlimit_value *= 1024;
		break;
#ifdef RLIMIT_NICE
	case RLIMIT_NICE:
		if (int_value > 19)
			int_value = 19;
		rlimit_value = 19 - int_value;
#endif
		break;
	}

	if ((limit_item != LIMIT_LOGIN)
	    && (limit_item != LIMIT_NUMSYSLOGINS)
	    && (limit_item != LIMIT_PRI)) {
		if (limit_type & LIMIT_SOFT) {
			if (pl->limits[limit_item].src_soft < source)
				return;
			else {
				pl->limits[limit_item].limit.rlim_cur =
				    rlimit_value;
				pl->limits[limit_item].src_soft = source;
			}
		}
		if (limit_type & LIMIT_HARD) {
			if (pl->limits[limit_item].src_hard < source)
				return;
			else {
				pl->limits[limit_item].limit.rlim_max =
				    rlimit_value;
				pl->limits[limit_item].src_hard = source;
			}
		}
	} else {
		/* recent kernels support negative priority limits (=raise priority) */

		if (limit_item == LIMIT_PRI)
			pl->priority = int_value;
		else {
			if (pl->login_limit_def < source)
				return;
			else {
				pl->login_limit = int_value;
				pl->login_limit_def = source;
			}
		}
	}
	return;


}


/** setup the quota limitation by setrlimit
 * @param pamh handler of PAM module
 * @param uname user to be limited
 * @param uid uid of current user
 * @param ctrl control code
 * @param pl limit paramters defined in configure file.
 * @return PAM_SUCCESS if success, or PAM_SYS_ERR.
 */
int xos_reslimit_setup(const char *uname, uid_t uid, XLIMIT * pl)
{
	int i;
	int status;
	int retval = LIMITED_OK;

	DEBUG("Setup quota for [user = '%s', uid = %d]", uname, uid);

	for (i = 0, status = LIMITED_OK; i < RLIM_NLIMITS; i++) {
		if (!pl->limits[i].supported)
			/* skip it if its not known to the system */
			continue;

		if (pl->limits[i].src_soft == LIMITS_DEF_NONE &&
		    pl->limits[i].src_hard == LIMITS_DEF_NONE)
			/* skip it if its not initialized */
			continue;

		if (pl->limits[i].limit.rlim_cur >
		    pl->limits[i].limit.rlim_max)
			pl->limits[i].limit.rlim_cur =
			    pl->limits[i].limit.rlim_max;
		status |= setrlimit(i, &pl->limits[i].limit);
	}

	if (status)
		retval = LIMIT_ERR;


	status = setpriority(PRIO_PROCESS, 0, pl->priority);
	if (status != 0)
		retval = LIMIT_ERR;


	if (uid == 0) {
		DEBUG("skip login limit check for uid=0");
	} else if (pl->login_limit > 0) {
		if (xos_reslimit_checklogin
		    (uname, pl->login_limit, pl) == LOGIN_ERR)
			retval |= LOGIN_ERR;

	} else if (pl->login_limit == 0)
		retval |= LOGIN_ERR;


	return retval;
}
