/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

#include <sys/types.h>

#include <openssl/bn.h>

#include <string.h>
#include <stdarg.h>

#include "xmalloc.h"
#include "buffer.h"
#include "xos_utils.h"

/*
 *  integers operation.
 */
u_int64_t get_u64(const void *vp)
{
	const u_char *p = (const u_char *) vp;
	u_int64_t v;

	v = (u_int64_t) p[0] << 56;
	v |= (u_int64_t) p[1] << 48;
	v |= (u_int64_t) p[2] << 40;
	v |= (u_int64_t) p[3] << 32;
	v |= (u_int64_t) p[4] << 24;
	v |= (u_int64_t) p[5] << 16;
	v |= (u_int64_t) p[6] << 8;
	v |= (u_int64_t) p[7];

	return (v);
}

u_int32_t get_u32(const void *vp)
{
	const u_char *p = (const u_char *) vp;
	u_int32_t v;

	v = (u_int32_t) p[0] << 24;
	v |= (u_int32_t) p[1] << 16;
	v |= (u_int32_t) p[2] << 8;
	v |= (u_int32_t) p[3];

	return (v);
}

u_int16_t get_u16(const void *vp)
{
	const u_char *p = (const u_char *) vp;
	u_int16_t v;

	v = (u_int16_t) p[0] << 8;
	v |= (u_int16_t) p[1];

	return (v);
}

void put_u64(void *vp, u_int64_t v)
{
	u_char *p = (u_char *) vp;

	p[0] = (u_char) (v >> 56) & 0xff;
	p[1] = (u_char) (v >> 48) & 0xff;
	p[2] = (u_char) (v >> 40) & 0xff;
	p[3] = (u_char) (v >> 32) & 0xff;
	p[4] = (u_char) (v >> 24) & 0xff;
	p[5] = (u_char) (v >> 16) & 0xff;
	p[6] = (u_char) (v >> 8) & 0xff;
	p[7] = (u_char) v & 0xff;
}

void put_u32(void *vp, u_int32_t v)
{
	u_char *p = (u_char *) vp;

	p[0] = (u_char) (v >> 24) & 0xff;
	p[1] = (u_char) (v >> 16) & 0xff;
	p[2] = (u_char) (v >> 8) & 0xff;
	p[3] = (u_char) v & 0xff;
}


void put_u16(void *vp, u_int16_t v)
{
	u_char *p = (u_char *) vp;

	p[0] = (u_char) (v >> 8) & 0xff;
	p[1] = (u_char) v & 0xff;
}

/*
 * Returns integers from the buffer (msb first).
 */

int buffer_get_short_ret(u_short * ret, Buffer * buffer)
{
	u_char buf[2];

	if (buffer_get_ret(buffer, (char *) buf, 2) == -1)
		return (-1);
	*ret = get_u16(buf);
	return (0);
}

u_short buffer_get_short(Buffer * buffer)
{
	u_short ret;

	if (buffer_get_short_ret(&ret, buffer) == -1)
		fatal("buffer_get_short: buffer error");

	return (ret);
}

int buffer_get_int_ret(u_int * ret, Buffer * buffer)
{
	u_char buf[4];

	if (buffer_get_ret(buffer, (char *) buf, 4) == -1)
		return (-1);
	*ret = get_u32(buf);
	return (0);
}

u_int buffer_get_int(Buffer * buffer)
{
	u_int ret;

	if (buffer_get_int_ret(&ret, buffer) == -1)
		fatal("buffer_get_int: buffer error");

	return (ret);
}

int buffer_get_int64_ret(u_int64_t * ret, Buffer * buffer)
{
	u_char buf[8];

	if (buffer_get_ret(buffer, (char *) buf, 8) == -1)
		return (-1);
	*ret = get_u64(buf);
	return (0);
}

u_int64_t buffer_get_int64(Buffer * buffer)
{
	u_int64_t ret;

	if (buffer_get_int64_ret(&ret, buffer) == -1)
		fatal("buffer_get_int: buffer error");

	return (ret);
}

/*
 * Stores integers in the buffer, msb first.
 */
void buffer_put_short(Buffer * buffer, u_short value)
{
	char buf[2];

	put_u16(buf, value);
	buffer_append(buffer, buf, 2);
}

void buffer_put_int(Buffer * buffer, u_int value)
{
	char buf[4];

	put_u32(buf, value);
	buffer_append(buffer, buf, 4);
}

void buffer_put_int64(Buffer * buffer, u_int64_t value)
{
	char buf[8];

	put_u64(buf, value);
	buffer_append(buffer, buf, 8);
}

/*
 * Returns an arbitrary binary string from the buffer.  The string cannot
 * be longer than 256k.  The returned value points to memory allocated
 * with xmalloc; it is the responsibility of the calling function to free
 * the data.  If length_ptr is non-NULL, the length of the returned data
 * will be stored there.  A null character will be automatically appended
 * to the returned string, and is not counted in length.
 */
void *buffer_get_string_ret(Buffer * buffer, u_int * length_ptr)
{
	u_char *value;
	u_int len;

	/* Get the length. */
	len = buffer_get_int(buffer);
	if (len > 256 * 1024) {
		DEBUG("buffer_get_string_ret: bad string length %u", len);
		return (NULL);
	}
	/* Allocate space for the string.  Add one byte for a null character. */
	value = xmalloc(len + 1);
	/* Get the string. */
	if (buffer_get_ret(buffer, value, len) == -1) {
		DEBUG("buffer_get_string_ret: buffer_get failed");
		xfree(value);
		return (NULL);
	}
	/* Append a null character to make processing easier. */
	value[len] = 0;
	/* Optionally return the length of the string. */
	if (length_ptr)
		*length_ptr = len;
	return (value);
}

void *buffer_get_string(Buffer * buffer, u_int * length_ptr)
{
	void *ret;

	if ((ret = buffer_get_string_ret(buffer, length_ptr)) == NULL)
		fatal("buffer_get_string: buffer error");
	return (ret);
}

/*
 * Stores and arbitrary binary string in the buffer.
 */
void buffer_put_string(Buffer * buffer, const void *buf, u_int len)
{
	buffer_put_int(buffer, len);
	buffer_append(buffer, buf, len);
}
void buffer_put_cstring(Buffer * buffer, const char *s)
{
	if (s == NULL)
		fatal("buffer_put_cstring: s == NULL");
	buffer_put_string(buffer, s, strlen(s));
}

/*
 * Returns a character from the buffer (0 - 255).
 */
int buffer_get_char_ret(char *ret, Buffer * buffer)
{
	if (buffer_get_ret(buffer, ret, 1) == -1) {
		DEBUG("buffer_get_char_ret: buffer_get_ret failed");
		return (-1);
	}
	return (0);
}

int buffer_get_char(Buffer * buffer)
{
	char ch;

	if (buffer_get_char_ret(&ch, buffer) == -1)
		fatal("buffer_get_char: buffer error");
	return (u_char) ch;
}

/*
 * Stores a character in the buffer.
 */
void buffer_put_char(Buffer * buffer, int value)
{
	char ch = value;

	buffer_append(buffer, &ch, 1);
}
