/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/conf.h>
#include <openssl/err.h>
#include <openssl/asn1.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/objects.h>
#include <openssl/pem.h>
#include <openssl/bn.h>
#include <openssl/lhash.h>
#include <openssl/rsa.h>

#include "xos_interfaces.h"


int rand_serial(BIGNUM * b, ASN1_INTEGER * ai)
{
	BIGNUM *btmp;
	int ret = 0;
	if (b)
		btmp = b;
	else
		btmp = BN_new();

	if (!btmp)
		return 0;

	if (!BN_pseudo_rand(btmp, 64, 0, 0))
		goto error;
	if (ai && !BN_to_ASN1_INTEGER(btmp, ai))
		goto error;

	ret = 1;

      error:

	if (!b)
		BN_free(btmp);

	return ret;
}


static int genrsa_cb(int p, int n, BN_GENCB * cb)
{
	char c = '*';

	if (p == 0)
		c = '.';
	if (p == 1)
		c = '+';
	if (p == 2)
		c = '*';
	if (p == 3)
		c = '\n';
	BIO_write(cb->arg, &c, 1);
	(void) BIO_flush(cb->arg);
#ifdef LINT
	p = n;
#endif
	return 1;
}

int create_proxy(char *certfile, char *keyfile, int days, char *attrs,
		 char *proxyfile)
{
	X509 *usercert = NULL, *proxy = NULL;
	int ret = -1, modulus = 512;;
	ASN1_INTEGER *sno = NULL;
	char *attributes = NULL;
	BIO *out = NULL, *bio_err = NULL;

	//long tout = 60*60*12*days;
	unsigned long f4 = RSA_F4;
	EVP_PKEY *proxykey = EVP_PKEY_new(), *signkey = EVP_PKEY_new();
	LHASH *hash = lh_new(NULL, NULL);
	X509V3_CTX ctx;
	X509_EXTENSION *ku_ext, *pci_ext, *mku_ext;
	X509_NAME *sname = NULL;
	RSA *rsa = RSA_new();
	BN_GENCB cb;
	BIGNUM *bn = BN_new();
	EVP_PKEY *userkey = NULL;
	const EVP_MD *digest;

	XOS_CREDS *creds = NULL;

	/*initialize */
	if (bio_err == NULL)
		bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);


	ERR_load_crypto_strings();

	out = BIO_new(BIO_s_file());

	if (out == NULL) {
		ERR_print_errors(bio_err);
		goto end;
	}

	if (proxyfile == NULL) {
		BIO_set_fp(out, stdout, BIO_NOCLOSE);
	} else {
		if (BIO_write_filename(out, proxyfile) <= 0) {
			perror(proxyfile);
			goto end;
		}
	}

	/* Create a new X509 Certificate to be the proxy */
	proxy = X509_new();

	creds = xos_creds_new();

	/* Load the certificate */
	if (certfile) {
		if (xos_cert_loadfromfile(creds, certfile) == XOS_ERROR) {
			printf("load certificate fail\n");
			goto end;
		}
		usercert = creds->certificate;
		if (usercert == NULL)
			goto end;
		while (X509_get_ext_count(usercert) > 0)
			X509_delete_ext(usercert, 0);
	}

	/* Load the private key */
	if (keyfile) {
		if (xos_cert_prvkeyfromfile
		    (creds, keyfile, NULL, "Your passwds:") == XOS_ERROR) {
			printf("load private key fail\n");
			goto end;
		}
		userkey = creds->private_key;
	}

	/* Generate a new RSA private key */
	BN_GENCB_set(&cb, genrsa_cb, bio_err);
	BIO_printf(bio_err,
		   "Generating RSA private key, %d bit long modulus\n",
		   modulus);
	if (!BN_set_word(bn, f4) ||
	    !RSA_generate_key_ex(rsa, modulus, bn, &cb) ||
	    !EVP_PKEY_assign_RSA(proxykey, rsa)) {
		BN_free(bn);
		RSA_free(rsa);
		goto end;
	}

	if (attrs == NULL)
		attributes = strdup("/OTHER");
	else
		attributes = strdup(attrs);


	/* Create the fancy proxy extensions */
	X509V3_set_conf_lhash(&ctx, hash);
	ku_ext = X509V3_EXT_conf(hash, &ctx, "keyUsage",
				 "Digital Signature, Key Encipherment, Data Encipherment");
	X509_EXTENSION_set_critical(ku_ext, 1);

	pci_ext = X509V3_EXT_conf(hash, &ctx, "proxyCertInfo",
				  "critical, language:Inherit all");
	X509_EXTENSION_set_critical(pci_ext, 1);

	mku_ext = X509V3_EXT_conf(hash, &ctx, "nsComment", attrs);
	X509_EXTENSION_set_critical(mku_ext, 0);

	/* Get a serial number, creating one if necessary */
	if (sno) {
		if (!X509_set_serialNumber(proxy, sno))
			goto end;
	} else {
		if (!rand_serial(NULL, X509_get_serialNumber(proxy)))
			goto end;
	}

	/* Assign the bits to the proxy */
	if (!X509_set_pubkey(proxy, proxykey))
		goto end;
	if (!X509_set_version(proxy, 2))
		goto end;
	if (X509_gmtime_adj(X509_get_notBefore(proxy), (long) 0) == NULL)
		goto end;
	if (X509_gmtime_adj
	    (X509_get_notAfter(proxy), (long) 3600 * 24 * days) == NULL)
		goto end;
	sname = X509_NAME_dup(X509_get_subject_name(usercert));
	if (!X509_NAME_add_entry_by_txt(sname, "commonName", MBSTRING_ASC,
					(unsigned char *)
					i2s_ASN1_INTEGER(NULL,
							 X509_get_serialNumber
							 (proxy)), -1, -1,
					0))
		goto end;
	if (!X509_set_issuer_name(proxy, X509_get_subject_name(usercert)))
		goto end;
	if (!X509_set_subject_name(proxy, sname))
		goto end;
	if (!X509_add_ext(proxy, ku_ext, 1))
		goto end;
	if (!X509_add_ext(proxy, pci_ext, 1))
		goto end;
	if (!X509_add_ext(proxy, mku_ext, 1))
		goto end;

	/* Assign the user supplied private key to the signing key */
	EVP_PKEY_assign_RSA(signkey, EVP_PKEY_get1_RSA(userkey));

	if (EVP_PKEY_type(signkey->type) == EVP_PKEY_DSA)
		digest = EVP_dss1();
	else if (EVP_PKEY_type(signkey->type) == EVP_PKEY_RSA)
		digest = EVP_sha1();
	else
		digest = EVP_md5();

	/* sign the proxy */
	if (!X509_sign(proxy, signkey, digest)) {
		ERR_print_errors(bio_err);
		RSA_free(rsa);
		EVP_PKEY_free(proxykey);
		goto end;
	}


	/* Print out the proxy */
	if (!PEM_write_bio_X509(out, proxy)) {
		BIO_printf(bio_err, "Unable to write proxy certificate\n");
		ERR_print_errors(bio_err);
	}
	if (!PEM_write_bio_RSAPrivateKey
	    (out, rsa, NULL, NULL, 0, NULL, NULL)) {
		BIO_printf(bio_err, "Unable to write private key\n");
		ERR_print_errors(bio_err);
	}
	if (!PEM_write_bio_X509(out, usercert)) {
		BIO_printf(bio_err, "Unable to write user certificate\n");
		ERR_print_errors(bio_err);
	}

	if (creds->private_key == NULL) {
		BIO_printf(bio_err,
			   "Malformed proxy credentials (No private key)");
		ERR_print_errors(bio_err);
	}



	if (PEM_write_bio_PrivateKey(out, creds->private_key, NULL,
				     NULL, 0, NULL, NULL) == 0) {
		BIO_printf(bio_err, "Error packing private key");

		ERR_print_errors(bio_err);
	}

	if (creds->certificate_chain != NULL) {
		int cert_chain_index;
		for (cert_chain_index = 0;
		     cert_chain_index < sk_num(creds->certificate_chain);
		     cert_chain_index++) {
			X509 *cert;

			cert = (X509 *) sk_value(creds->certificate_chain,
						 cert_chain_index);

			if (PEM_write_bio_X509(out, cert) == SSL_ERROR) {
				BIO_printf(bio_err,
					   "Error packing certificate chain");

				ERR_print_errors(bio_err);
			}
		}
	}

	RSA_free(rsa);


	ret = 0;

      end:

	if (creds)
		xos_creds_free(creds);

	OBJ_cleanup();
	BIO_free_all(out);
	ASN1_INTEGER_free(sno);

	return ret;
}
