/*
 * Copyright (c) 2008 Institute of Computing Technology of Chinese Academy of Sciences and Telefónica Investigación y Desarrollo SAU.
 * All rights reserved.
 *
 * This file is part of Linux-XOS-VOsupport, which is part of XtreemOS,
 * a Linux-based Grid Operating System, see <http://www.xtreemos.eu> for more details.
 * The XtreemOS project has been developed with the financial support of the European
 * Commission's IST program under contract #FP6-033576.
 *
 * Linux-XOS-VOsupport is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 2 of the License, or (at your option) any later version.
 *
 * Linux-XOS-VOsupport is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with LinuxMD-XOS.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * */

/*
 * xos_sslutils.c
 * SSL utility lib for XtreemOS
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include <assert.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>

#include "xos_utils.h"
#include "xos_sslutils.h"
#include "xos_key.h"
#include "buffer.h"

#define IS_DEBUG


/*
 ******************************************************************************************
 * ssl_init()
 *
 * Do any needed initialization for these routines.
 * Should be called first. Can be called multiple times.
 */
static
void ssl_init()
{
	static int ssl_inited = 0;

	if (ssl_inited == 0) {
		ssl_inited = 1;

#ifdef _WIN32
		srand(time(NULL));
#else
		srandom(time(NULL));
		srand(time(NULL));
#endif

		/* Initialize the ssleay libraries */
		SSL_load_error_strings();
		SSLeay_add_ssl_algorithms();
	}
}

/*
 ******************************************************************************************
 * ssl_credentials_new()
 * 
 */
SSL_CREDENTIALS *ssl_credentials_new()
{
	SSL_CREDENTIALS *creds = NULL;

	ssl_init();

	creds = malloc(sizeof(*creds));

	if (creds == NULL) {
		perror("errno");
		goto error;
	}

	creds->certificate = NULL;
	creds->private_key = NULL;
	creds->certificate_chain = NULL;

      error:
	return creds;
}

/*
 ******************************************************************************************
 */

int xosssl_loadcertfrombuffer(char *buffer, SSL_CREDENTIALS * cerds)
{
	return SSL_ERROR;
}

/*
 ******************************************************************************************
 */
int xosssl_loadcertfromfile(SSL_CREDENTIALS * creds, const char *path)
{
	FILE *cert_file = NULL;
	X509 *cert = NULL;
	int return_status = SSL_ERROR;
	STACK *cert_chain = NULL;

	assert(creds != NULL);
	assert(path != NULL);

	ssl_init();

	cert_file = fopen(path, "r");

	if (cert_file == NULL) {
		DEBUG("%s:Error opening certificate file %s:",
		      __FUNCTION__, path);
		perror("errno");
		goto error;
	}

	if ((cert =
	     PEM_read_X509(cert_file, NULL, PEM_NO_CALLBACK)) == NULL) {
		DEBUG("%s:Error reading certificate %s:", __FUNCTION__,
		      path);
		perror("errno");
		goto error;
	}

	if (creds->certificate != NULL) {
		X509_free(creds->certificate);
	}

	creds->certificate = cert;

	/* Ok, now read the certificate chain */

	/* Create empty stack */
	cert_chain = sk_new_null();

	while (1) {
		cert = NULL;

		if ((cert =
		     PEM_read_X509(cert_file, NULL,
				   PEM_NO_CALLBACK)) == NULL) {
			/*
			 * If we just can't find a start line then we've reached EOF.
			 */
			if (ERR_GET_REASON(ERR_peek_error()) ==
			    PEM_R_NO_START_LINE) {
				/* Just EOF, clear error and break out of loop */
				ERR_clear_error();
				break;
			}

			/* Actual error */
			DEBUG("Error parsing certificate chain\n");
			//ssl_error_to_verror();
			perror("error");
			goto error;
		}

		/* Add to chain */
		if (sk_insert(cert_chain, (char *) cert,
			      sk_num(cert_chain)) == SSL_ERROR) {
			DEBUG("%s:Error parsing certificate chain\n",
			      __FUNCTION__);
			//ssl_error_to_verror();
			perror("error");
			goto error;
		}
	}			/* while(1) */

	creds->certificate_chain = cert_chain;

	/* Success */
	return_status = SSL_SUCCESS;

      error:
	if (cert_file != NULL) {
		fclose(cert_file);
	}

	return return_status;
}

/*
 ******************************************************************************************
 */
int xosssl_getsubjectfromcert(X509 * cert, char *buffer, int buflen)
{
	X509_NAME *subject;

	if (cert == NULL || buffer == NULL || buflen <= 0) {
		DEBUG("%s: parameters error!\n", __FUNCTION__);
		return SSL_ERROR;
	}


	subject = X509_NAME_dup(X509_get_subject_name(cert));
	if (subject == NULL) {
		DEBUG("%s:Failed to get the subject from cert",
		      __FUNCTION__);
		return SSL_ERROR;
	}

	X509_NAME_oneline(subject, buffer, buflen);
	X509_NAME_free(subject);

	//DEBUG("%s,buffer is :%s\n",__FUNCTION__, buffer);

	return SSL_SUCCESS;


}

/*
 ******************************************************************************************
 */
Key *pemkey_transfor(EVP_PKEY * pk)
{
	Key *prv = NULL;

	if (pk == NULL)
		return NULL;

	if (pk->type == EVP_PKEY_RSA) {
		prv = key_new(KEY_UNSPEC);
		prv->rsa = EVP_PKEY_get1_RSA(pk);
		prv->type = KEY_RSA;

	} else if (pk->type == EVP_PKEY_DSA) {
		prv = key_new(KEY_UNSPEC);
		prv->dsa = EVP_PKEY_get1_DSA(pk);
		prv->type = KEY_DSA;
	}

	return prv;
}

Key *xosssl_getpublickeyfromcert(X509 * cert)
{
	EVP_PKEY *key = NULL;
	Key *pubkey = NULL;

	key = X509_get_pubkey(cert);
	pubkey = pemkey_transfor(key);
	//pubkey = key;
	if (pubkey == NULL) {
		DEBUG("Invalid pem public key");
		return NULL;
	}

	if (key != NULL) {
		EVP_PKEY_free(key);
	}

	return pubkey;
}



/*
 ******************************************************************************************
 */

const char *xosssl_passphrase = NULL;
/*
 * xos_passphrass_callback()
 *
 * Callback from PEM_read_PrivateKey() in ssl_load_user_key()
 * to return the passphrase stored in xosssl_passphrase.
 */
static
int xos_passphrass_callback(char *buffer, int buffer_len,
			    int verify /* Ignored */ ,
			    void *u)
{
	/* SSL libs supply these, make sure they are reasonable */
	if (buffer == NULL || buffer_len <= 0)
		return 0;

	if (xosssl_passphrase == NULL) {
		strcpy(buffer, "");
	} else {
		strncpy(buffer, xosssl_passphrase, buffer_len);
		buffer[buffer_len - 1] = '\0';
	}

	return strlen(buffer);
}

/*
 * bio_from_buffer()
 *
 * Given a buffer of length buffer_len, return a memory bio with the
 * contents of the buffer.
 *
 * Returns pointer to bio on success, NULL on error.
 */
static
BIO *bio_from_buffer(const unsigned char *buffer, int buffer_len)
{
	BIO *bio = NULL;

	if (buffer == NULL) {
		return NULL;
	}

	bio = BIO_new(BIO_s_mem());

	if (bio == NULL) {
		DEBUG("Failed creating memory BIO\n");
		goto error;
	}

	if (BIO_write(bio, (unsigned char *) buffer, buffer_len) ==
	    SSL_ERROR) {
		DEBUG("Failed writing buffer to BIO\n");
		BIO_free(bio);
		bio = NULL;
		goto error;
	}

      error:
	return bio;
}

int
xosssl_bufferfromfile(const char *path, unsigned char **pbuffer,
		      int *pbuffer_len)
{
	int fd = -1;
	int open_flags;
	int return_status = -1;
	struct stat statbuf;
	unsigned char *buffer = NULL, *b = NULL;
	int buffer_len;
	int remaining;
	int rval;

	assert(path != NULL);
	assert(pbuffer != NULL);

	open_flags = O_RDONLY;

	fd = open(path, open_flags);

	if (fd == -1) {
		DEBUG("Failure opening file \"%s\"", path);
		perror("errno");
		goto error;
	}

	if (fstat(fd, &statbuf) == -1) {
		DEBUG("Failure stating file \"%s\"", path);
		perror("errno");
		goto error;
	}

	buffer_len = statbuf.st_size;

	b = buffer = malloc(buffer_len + 1);
	if (buffer == NULL) {
		DEBUG("malloc() failed");
		perror("errno");
		goto error;
	}

	remaining = buffer_len;
	while (remaining) {
		rval = read(fd, b, remaining);
		if (rval == -1) {
			DEBUG("Error reading file \"%s\"", path);
			perror("errno");
			goto error;
		}
		remaining -= rval;
		b += rval;
	}
	buffer[buffer_len++] = '\0';

	/* Succcess */
	*pbuffer = buffer;
	if (pbuffer_len)
		*pbuffer_len = buffer_len;
	return_status = 0;

      error:
	if (fd != -1) {
		close(fd);
	}

	if (return_status == -1) {
		if (buffer != NULL)
			free(buffer);
	}

	return return_status;
}

/*
 * xosssl_certchainfree()
 *
 * Free the given certificate chain and all it contents.
 */
static
void xosssl_certchainfree(STACK * cert_chain)
{
	if (cert_chain != NULL)
		sk_pop_free(cert_chain, (void (*)(void *)) X509_free);

}

/*
 * xosssl_credsfreecontents()
 *
 * Free all the contents of the given credentials without freeing
 * the credentials structure itself.
 */
void xosssl_credsfreecontents(SSL_CREDENTIALS * creds)
{
	if (creds != NULL) {
		if (creds->certificate != NULL)
			X509_free(creds->certificate);

		if (creds->private_key != NULL)
			EVP_PKEY_free(creds->private_key);

		if (creds->certificate_chain != NULL)
			xosssl_certchainfree(creds->certificate_chain);

	}
}

int
xosssl_proxyfrompem(SSL_CREDENTIALS * creds, const unsigned char *buffer,
		    int buffer_len, const char *pass_phrase)
{
	BIO *bio = NULL;
	X509 *cert = NULL;
	EVP_PKEY *key = NULL;
	STACK *cert_chain = NULL;
	int return_status = SSL_ERROR;

	assert(creds != NULL);
	assert(buffer != NULL);

	ssl_init();

	/* 
	 * Put pass phrase where the callback function can find it.
	 */
	xosssl_passphrase = pass_phrase;

	bio = bio_from_buffer(buffer, buffer_len);

	if (bio == NULL) {
		goto error;
	}

	/*
	 * Proxy file contains proxy certificate followed by proxy
	 * private key, followed by the certificate chain.
	 */

	/* Read proxy certificate */
	if (PEM_read_bio_X509(bio, &cert, PEM_NO_CALLBACK) == NULL) {
		DEBUG("Error parsing proxy certificate");
		//ssl_error_to_verror();
		goto error;
	}

	/* Read proxy private key */
	if (PEM_read_bio_PrivateKey(bio, &(key),
				    PEM_CALLBACK(xos_passphrass_callback))
	    == NULL) {
		unsigned long error, reason;

		error = ERR_peek_error();
		reason = ERR_GET_REASON(error);

		/* If this is a bad password, return a better error message */
		if (ERR_GET_REASON(error) == EVP_R_BAD_DECRYPT ||
		    reason == EVP_R_NO_SIGN_FUNCTION_CONFIGURED) {
			DEBUG("Bad password");
		} else {
			DEBUG("Error parsing private key");
			//ssl_error_to_verror();
		}

		goto error;
	}

	/* Ok, now read the certificate chain */

	/* Create empty stack */
	cert_chain = sk_new_null();

	while (1) {
		X509 *certificate = NULL;

		if (PEM_read_bio_X509(bio, &certificate, PEM_NO_CALLBACK)
		    == NULL) {
			/*
			 * If we just can't find a start line then we've reached EOF.
			 */
			if (ERR_GET_REASON(ERR_peek_error()) ==
			    PEM_R_NO_START_LINE) {
				/* Just EOF, clear error and break out of loop */
				ERR_clear_error();
				break;
			}

			/* Actual error */
			DEBUG
			    ("Error parsing certificate chain from proxy");
			//ssl_error_to_verror();
			goto error;
		}

		/* Add to chain */
		if (sk_insert(cert_chain, (char *) certificate,
			      sk_num(cert_chain)) == SSL_ERROR) {
			DEBUG
			    ("Error parsing certificate chain from proxy");
			//ssl_error_to_verror();
			goto error;
		}
	}			/* while(1) */

	/*
	 * Ok, everything has been successfully read, now store it into
	 * creds, removing any existing contents.
	 */
	xosssl_credsfreecontents(creds);

	creds->private_key = key;
	creds->certificate = cert;
	creds->certificate_chain = cert_chain;

	/* Success */
	return_status = SSL_SUCCESS;

      error:
	if (return_status == SSL_ERROR) {
		/*
		 * On error, clean up any key, cert or chain. On success
		 * we don't want to do this as they are part of the creds.
		 */

		if (cert != NULL)
			X509_free(cert);

		if (key != NULL)
			EVP_PKEY_free(key);


		if (cert_chain)
			xosssl_certchainfree(cert_chain);

	}

	if (bio != NULL)
		BIO_free(bio);


	return return_status;
}

int xosssl_loadproxyfromfile(SSL_CREDENTIALS * creds, const char *path,
			     const char *pass_phrase)
{
	unsigned char *buffer = NULL;
	int buffer_len;
	int return_status = SSL_ERROR;

	assert(creds != NULL);
	assert(path != NULL);

	ssl_init();

	/* Read the whole contents of the given file */
	if (xosssl_bufferfromfile(path, &buffer, &buffer_len) == -1) {
		DEBUG("%s,Error in read file to buffer\n", __FUNCTION__);
		goto error;
	}

	if (xosssl_proxyfrompem(creds, buffer, buffer_len, pass_phrase) ==
	    SSL_ERROR) {
		DEBUG("%s:Error reading proxy from %s", __FUNCTION__,
		      path);
		goto error;
	}

	/* Success */
	return_status = SSL_SUCCESS;

      error:
	if (buffer != NULL) {
		free(buffer);
	}

	return return_status;
}

/*
 ******************************************************************************************
 */
int xosssl_getprivatekeyfromfile(SSL_CREDENTIALS * creds,
				 const char *path,
				 const char *pass_phrase,
				 const char *pass_phrase_prompt)
{
	FILE *key_file = NULL;
	EVP_PKEY *key = NULL;
	int return_status = SSL_ERROR;

	assert(creds != NULL);
	assert(path != NULL);

	ssl_init();

	/* 
	 * Put pass phrase where the callback function can find it.
	 */
	xosssl_passphrase = pass_phrase;

	if (pass_phrase_prompt)
		EVP_set_pw_prompt((char *) pass_phrase_prompt);

	key_file = fopen(path, "r");

	if (key_file == NULL) {
		DEBUG("Error opening key file %s\n", path);
		goto error;
	}
	if (PEM_read_PrivateKey(key_file, &(key), (pass_phrase_prompt) ?
				NULL : xos_passphrass_callback,
				NULL) == NULL) {

		goto error;
	}

	if (creds->private_key != NULL) {
		EVP_PKEY_free(creds->private_key);
	}

	creds->private_key = key;

	/* Success */
	return_status = SSL_SUCCESS;

      error:
	if (key_file != NULL) {
		fclose(key_file);
	}
	xosssl_passphrase = NULL;

	return return_status;

}

/*
 ******************************************************************************************
 * Verify signature
 * Need buffer.h and xos_key.h to support
 */

int xosssl_signatureverify(u_char * sig, u_char * data)	// (Buffer *b)
{
	/* Verify steps:
	 *
	 * 1. get certificate from data
	 * 2. get public key from certificate
	 * 3. call xos_key_verify() in key.c
	 *      ret = xos_key_verify(pulickey, sig, strlen(sig), data, strlen(data));
	 * Return: Success, if ret = 1; or ret = 0;
	 */

	/* If we take the vega-openssh as example:
	 *  We must follow the data format, and codes will be prototype:

	 int ret;
	 Buffer b;
	 buffer_init(&b);
	 buffer_put_string(&b, session_id2, session_id2_len);

	 // reconstruct packet 
	 buffer_put_char(&b, SSH2_MSG_USERAUTH_REQUEST);
	 buffer_put_cstring(&b, agorauser);
	 buffer_put_cstring(&b, service);
	 buffer_put_cstring(&b, VEGA_AUTH_NAME);
	 buffer_put_cstring(&b, subject);
	 buffer_put_char(&b, have_sig);
	 buffer_put_cstring(&b, pkalg);
	 buffer_put_string(&b, certificate, certlen);  

	 ret = xos_key_verify(key, sig, slen, buffer_ptr(&b),buffer_len(&b)))
	 if(ret != SSL_SUCCESS)
	 {
	 DEBUG("%s: verify sigature fail\n", __FUNCTION__);
	 return SSL_ERROR;
	 }
	 return SSL_SUCCESS;

	 * Todo: the sigature procedure must compromise with client-end
	 */

	return SSL_ERROR;
}


/*
 ******************************************************************************************
 */

int xosssl_gettimes(const char *path, time_t * not_before,
		    time_t * not_after)
{
	return SSL_ERROR;
}

/*
 ******************************************************************************************
 */
int xosssl_encryptWithSymKey(unsigned char *dataToEncrypt,
			     int dataToEncryptLen,
			     unsigned char *encryptedData,
			     int *encryptedDataLen, SymKey * k)
{

	ssl_init();

	EVP_CIPHER_CTX encryptCtx;
	EVP_CIPHER_CTX_init(&encryptCtx);

	int tmpLen;
	*encryptedDataLen = 0;

	if (!EVP_CipherInit_ex(&encryptCtx, EVP_bf_cbc(), NULL, k->key, k->iv, 1 /* encrypt */ ))	//ENC_ALGO
	{
		DEBUG("Unable to initialize encryption context!\n");
		//ERR_print_errors_fp(stderr);
		return SSL_ERROR;
	}

	if (!EVP_CipherUpdate
	    (&encryptCtx, encryptedData, &tmpLen, dataToEncrypt,
	     dataToEncryptLen)) {

		DEBUG("Unable to encrypt!\n");
		//ERR_print_errors_fp(stderr);
		return SSL_ERROR;
	}
	*encryptedDataLen += tmpLen;

	// Write any remaining parts of the encrypted data to the out buffer
	EVP_CipherFinal_ex(&encryptCtx,
			   (encryptedData + *encryptedDataLen), &tmpLen);

	*encryptedDataLen += tmpLen;

	EVP_CIPHER_CTX_cleanup(&encryptCtx);


	return SSL_SUCCESS;
}


/*
 ******************************************************************************************
 * Decrypt the data with the specified symmetric key.
 */
int xosssl_decryptWithSymKey(unsigned char *dataToDecrypt,
			     int dataToDecryptLen,
			     unsigned char *decryptedData,
			     int *decryptedDataLen, SymKey * k)
{

	ssl_init();
	EVP_CIPHER_CTX decryptCtx;
	EVP_CIPHER_CTX_init(&decryptCtx);
	if (!EVP_CipherInit_ex(&decryptCtx, EVP_bf_cbc(), NULL, k->key, k->iv, 0 /* decrypt */ ))	//ENC_ALGO
	{

		DEBUG("Unable to initialize decryption context!\n");
		//ERR_print_errors_fp(stderr);
		return SSL_ERROR;
	}

	int tmpLen;
	*decryptedDataLen = 0;

	if (!EVP_CipherUpdate
	    (&decryptCtx, decryptedData, &tmpLen, dataToDecrypt,
	     dataToDecryptLen)) {

		DEBUG("Unable to decrypt!\n");
		//ERR_print_errors_fp(stderr);
		return SSL_ERROR;
	}

	*decryptedDataLen += tmpLen;
	EVP_CipherFinal_ex(&decryptCtx,
			   (decryptedData + *decryptedDataLen), &tmpLen);
	*decryptedDataLen += tmpLen;
	EVP_CIPHER_CTX_cleanup(&decryptCtx);

	return SSL_SUCCESS;

}

/*
 ******************************************************************************************
 */
static int
bio_to_buffer(BIO * bio, unsigned char **pbuffer, int *pbuffer_len)
{
	char *buffer = NULL;
	int buffer_len;
	int return_status = SSL_ERROR;

	assert(bio != NULL);

	buffer_len = BIO_pending(bio);

	buffer = malloc(buffer_len);

	if (buffer == NULL) {
		DEBUG("Failed dumping BIO to buffer (malloc() failed)");
		goto error;
	}

	if (BIO_read(bio, buffer, buffer_len) == SSL_ERROR) {
		DEBUG("Failed dumping BIO to buffer (BIO_read() failed)");
		goto error;
	}

	/* Success */
	*pbuffer = (unsigned char *) buffer;
	*pbuffer_len = buffer_len;
	return_status = SSL_SUCCESS;

      error:
	if (return_status == SSL_ERROR) {
		if (buffer != NULL) {
			free(buffer);
		}
	}

	return return_status;
}

int
xosssl_proxytopem(SSL_CREDENTIALS * creds,
		  unsigned char **pbuffer,
		  int *pbuffer_len, const char *pass_phrase)
{
	BIO *bio = NULL;
	const EVP_CIPHER *cipher;
	int pass_phrase_len;
	int cert_chain_index;
	int return_status = SSL_ERROR;

	assert(creds != NULL);
	assert(pbuffer != NULL);
	assert(pbuffer_len != NULL);

	ssl_init();

	bio = BIO_new(BIO_s_mem());

	if (bio == NULL) {
		goto error;
	}

	/*
	 * Write out proxy certificate, followed by proxy private key and
	 * then followed by the cert chain.
	 */
	if (creds->certificate == NULL) {
		goto error;
	}

	if (PEM_write_bio_X509(bio, creds->certificate) == SSL_ERROR) {
		DEBUG("Error packing proxy certificate");
		goto error;
	}

	if (creds->private_key == NULL) {
		DEBUG("Malformed proxy credentials (No private key)");
		goto error;
	}

	if (pass_phrase == NULL) {
		/* No encryption */
		cipher = NULL;
		pass_phrase_len = 0;
	} else {
		/* Encrypt with pass phrase */
		/* XXX This is my best guess at a cipher */
		cipher = EVP_des_ede3_cbc();
		pass_phrase_len = strlen(pass_phrase);
	}

	if (PEM_write_bio_PrivateKey(bio, creds->private_key, cipher,
				     (unsigned char *) pass_phrase,
				     pass_phrase_len,
				     PEM_NO_CALLBACK) == SSL_ERROR) {
		DEBUG("Error packing private key");
		goto error;
	}

	if (creds->certificate_chain != NULL) {

		for (cert_chain_index = 0;
		     cert_chain_index < sk_num(creds->certificate_chain);
		     cert_chain_index++) {
			X509 *cert;

			cert = (X509 *) sk_value(creds->certificate_chain,
						 cert_chain_index);

			if (PEM_write_bio_X509(bio, cert) == SSL_ERROR) {
				DEBUG("Error packing certificate chain");
				goto error;
			}
		}
	}

	/* OK, bio is filled, now dump to buffer */
	if (bio_to_buffer(bio, pbuffer, pbuffer_len) == SSL_ERROR) {
		goto error;
	}

	/* Success */
	return_status = SSL_SUCCESS;

      error:
	if (bio != NULL) {
		BIO_free(bio);
	}

	return return_status;
}

int xosssl_proxydestroy(const char *proxyfile)
{
	FILE *fp;
	long offset, i;
	char zero = '\0';
	struct stat s;
	int return_status = SSL_ERROR;

	assert(proxyfile != NULL);

	fp = fopen(proxyfile, "r+");
	if (!fp) {
		DEBUG("fopen(%s): %s\n", proxyfile, strerror(errno));
		goto error;
	}
	/* Don't get fooled into zeroing out the wrong file via tricks
	   with links and the like. */
	if (fstat(fileno(fp), &s) < 0) {
		DEBUG("fstat(%s): %s\n", proxyfile, strerror(errno));
		goto error;
	}
	if (S_ISDIR(s.st_mode)) {
		DEBUG("proxy file %s is a directory!\n", proxyfile);
		goto error;
	}
	if (!S_ISREG(s.st_mode)) {
		DEBUG("proxy file %s is not a regular file!\n", proxyfile);
		goto error;
	}
	if (s.st_nlink != 1) {
		DEBUG("proxy file %s has links!\n", proxyfile);
		goto error;
	}
	if (fseek(fp, 0L, SEEK_END) < 0) {
		DEBUG("fseek(%s): %s\n", proxyfile, strerror(errno));
		goto error;
	}
	offset = ftell(fp);
	if (offset < 0) {
		DEBUG("ftell(%s): %s\n", proxyfile, strerror(errno));
		goto error;
	}
	if (fseek(fp, 0L, SEEK_SET) < 0) {
		DEBUG("fseek(%s): %s\n", proxyfile, strerror(errno));
		goto error;
	}
	for (i = 0; i < offset; i++) {
		if (fwrite(&zero, 1, 1, fp) != 1) {
			DEBUG("fwrite(%s): %s\n", proxyfile,
			      strerror(errno));
			goto error;
		}
	}

	return_status = SSL_SUCCESS;

      error:
	if (fp)
		fclose(fp);
	if (unlink(proxyfile) < 0) {	/* always try to unlink it, even on error */
		DEBUG("unlink: %s\n", strerror(errno));
		return SSL_ERROR;
	}

	return return_status;
}

int
xosssl_proxytofile(SSL_CREDENTIALS * proxy_creds,
		   const char *path, const char *pass_phrase)
{
	int fd = -1;
	int open_flags;
	int return_status = SSL_ERROR;
	unsigned char *buffer = NULL;
	int buffer_len;
	mode_t file_mode = 0;

	assert(proxy_creds != NULL);
	assert(path != NULL);

	ssl_init();

	/*
	 * Use open to open the file so we can make sure it doesn't already
	 * exist.
	 */
	open_flags = O_CREAT | O_EXCL | O_WRONLY;
	file_mode = S_IRUSR | S_IWUSR;	/* 0600 */

	fd = open(path, open_flags, file_mode);

	if (fd == -1) {
		DEBUG("Error creating %s", path);
		goto error;
	}

	/*
	 * Dump proxy to buffer
	 */
	if (xosssl_proxytopem(proxy_creds, &buffer,
			      &buffer_len, pass_phrase) == SSL_ERROR) {
		goto error;
	}

	if (write(fd, buffer, buffer_len) == -1) {
		DEBUG("Error writing proxy to %s", path);
		goto error;
	}

	/* Success */
	return_status = SSL_SUCCESS;

      error:
	if (buffer != NULL) {
		free(buffer);
	}

	if (fd != -1) {
		close(fd);

		if (return_status == SSL_ERROR) {
			/* Remove any file we created */
			xosssl_proxydestroy(path);
		}
	}

	return return_status;
}

/* added for CDA cert testing */

int xosssl_certfrombuffer(unsigned char *buffer, int buffer_length,
			  SSL_CREDENTIALS * creds)
{
	X509 *cert = NULL;
	
BIO * bio;

	if (!buffer || !creds)
		return SSL_ERROR;

	bio = BIO_new_mem_buf(buffer, -1);
	
PEM_read_bio_X509(bio, &cert, 0, NULL);
	
if (cert == NULL)
		return SSL_ERROR;
	
if (creds->certificate != NULL)
		X509_free(creds->certificate);
	
creds->certificate = cert;
	return SSL_SUCCESS;
}

